/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 *  Authors: Iain Holmes <iain@prettypeople.org>
 *
 *  Copyright 2002 Iain Holmes
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of version 2 of the GNU General Public 
 *  License as published by the Free Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <errno.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <math.h>

#include <glib.h>
#include <glib-object.h>
#include <glib/gi18n.h>

#include <marlin/marlin-block.h>

#undef LOCK_DEBUG
/* #define LOCK_DEBUG 1 */

#include <marlin/marlin-channel.h>
#include <marlin/marlin-utils.h>

/* Ugly forward declarations */
static void lockless_unlink_range (MarlinChannel *channel,
				   guint64 start_frame,
				   guint64 finish_frame,
				   MarlinBlock **blocks,
				   MarlinUndoContext *ctxt);
static void lockless_link_blocks (MarlinChannel *channel,
				  guint64 insert_frame,
				  guint64 num_frames,
				  MarlinBlock *block_list,
				  MarlinUndoContext *ctxt);


GQuark
marlin_channel_error_quark (void)
{
	static GQuark quark = 0;
	if (quark == 0) {
		quark = g_quark_from_static_string ("marlin-channel-error-quark");
	}

	return quark;
}

/**
 * marlin_channel_new:
 * sample:
 * filename:
 * error:
 * 
 * Creates a new channel for @sample, using @filename as the basename for its
 * temp files.
 *
 * Returns: Newly created MarlinChannel or NULL on error.
 */
MarlinChannel *
marlin_channel_new (const char *filename,
		    GError **error)
{
	MarlinChannel *channel;
	char *basename, *name;
	
	channel = g_new0 (MarlinChannel, 1);
	channel->lock = marlin_read_write_lock_new ();

	basename = g_path_get_basename (filename);
	name = g_strdup_printf ("%s/%s.XXXXXX",
				marlin_get_tmp_dir (), 
				basename);
	g_free (basename);
	
	channel->frame_file = marlin_file_new (name, error);
	g_free (name);

	if (channel->frame_file == NULL) {
		marlin_channel_free (channel);
		return NULL;
	}

	basename = g_path_get_basename (channel->frame_file->filename);

	name = g_strdup_printf ("%s/%s.XXXXXX",
				marlin_get_tmp_dir (), 
				basename);
	g_free (basename);
		
	channel->peak_file = marlin_file_new (name, error);
	g_free (name);

	if (channel->peak_file == NULL) {
		marlin_channel_free (channel);
		return NULL;
	}

	channel->pages = g_queue_new ();
	return channel;
}

/**
 * marlin_channel_free:
 * channel:
 *
 * Frees all resources that @channel was using.
 */
void
marlin_channel_free (MarlinChannel *channel)
{
	MarlinBlock *block;
	GList *l;
	
	g_return_if_fail (channel != NULL);

	WRITE_LOCK (channel->lock);

	if (channel->first) {
		marlin_block_free_list (channel->first);
	}

	if (channel->frame_file) {
		marlin_file_unref (channel->frame_file);
	}

	if (channel->peak_file) {
		marlin_file_unref (channel->peak_file);
	}

	g_queue_free (channel->pages);

	WRITE_UNLOCK (channel->lock);
	marlin_read_write_lock_destroy (channel->lock);
	
	g_free (channel);
}

/**
 * marlin_channel_create_block:
 * @channel:
 *
 * Creates a new block in this channel.
 *
 * Returns: A newly created MarlinBlock.
 */
MarlinBlock *
marlin_channel_create_block (MarlinChannel *channel)
{
	return marlin_block_new (channel, channel->frame_file, 
				 channel->peak_file);
}

/* Implements an LRU paging system */
/* Max number of blocks that can be mapped for each channel
   Each page is about 2meg of data */
#define MAX_PAGE_COUNT 10 

/* define this to turn off the paging system to simplify testing
   of the block/channel manipulation code. */
/* #define NO_PAGE_SYSTEM */

/**
 * marlin_channel_map_block:
 * @channel:
 * @block:
 *
 * Adds @block to the LRU cache for @channel or if it is already in
 * the cache it is moved to the start of the queue. If there isn't room
 * in the cache then the block at the end of the queue is removed and 
 * unmapped.
 *
 * The channel must be write locked.
 */
void
marlin_channel_map_block (MarlinChannel *channel,
			  MarlinBlock *block)
{
#ifndef NO_PAGE_SYSTEM
	/* Remove the block from the page system */
	g_queue_remove (channel->pages, block);

	if (g_queue_get_length (channel->pages) == MAX_PAGE_COUNT) {
		MarlinBlock *b;

		b = g_queue_pop_tail (channel->pages);
		marlin_block_unmap (b);
	}
	
	g_queue_push_head (channel->pages, block);
#endif
	marlin_block_map (block, NULL);
}

/**
 * marlin_channel_unmap_block:
 * @channel:
 * @block:
 *
 * Removes @block from the LRU cache.
 *
 * The channel must be write locked.
 */
void
marlin_channel_unmap_block (MarlinChannel *channel,
			    MarlinBlock *block)
{
#ifndef NO_PAGE_SYSTEM
	g_queue_remove (channel->pages, block);
#endif
	marlin_block_unmap (block);
}

/* Channel lock must be held before calling this function */
static MarlinBlock *
lockless_get_for_frame (MarlinBlock *block,
			guint64 frame)
{
	/* Yes, this is even slower */
	for (; block; block = marlin_block_next (block)) {
		if (frame >= block->start && frame <= block->end) {
			return block;
		}
	}

	return NULL;
}

/**
 * marlin_channel_get_block_for_frame:
 * channel:
 * frame:
 *
 * Gets the block from @channel that @frame is found in.
 *
 * Returns: MarlinBlock or NULL on failure.
 */
MarlinBlock *
marlin_channel_get_block_for_frame (MarlinChannel *channel,
				    guint64 frame)
{
	MarlinBlock *block;
	g_return_val_if_fail (channel != NULL, NULL);

	READ_LOCK (channel->lock);

	if (frame == channel->frames) {
		MarlinBlock *b;

		b = channel->last;
		READ_UNLOCK (channel->lock);
		
		return b;
	}

	if (frame == (guint64) 0) {
		MarlinBlock *b;

		b = channel->first;
		READ_UNLOCK (channel->lock);

		return b;
	}
	
	block = lockless_get_for_frame (channel->first, frame);

	READ_UNLOCK (channel->lock);
		
	return block;
}

/*
 * splits the channel at split_frame
 * 0 |--------------| 20
 * Split at 10
 * 0 |----| 9 10 |------| 20
 */
static void
lockless_split_block (MarlinChannel *channel,
		      guint64 split_frame)
{
	MarlinBlock *first, *second;

	g_return_if_fail (split_frame <= channel->frames);

	if (split_frame == channel->frames || split_frame == 0) {
		return;
	}

	first = lockless_get_for_frame (channel->first, split_frame);
	g_return_if_fail (first != NULL);
	
	if (split_frame != first->start) {

		second = marlin_block_split (first, split_frame);

		/* Insert second into the list after first */
		if (first->next == NULL) {
			first->next = second;
			second->previous = first;
			second->next = NULL;
			
			/* If first->next was NULL,
			   then first was the last block
			   Set the channel last point to second */
			channel->last = second;
		} else {
			second->next = first->next;
			second->next->previous = second;
			second->previous = first;
			first->next = second;
		}
	}
}

/**
 * marlin_channel_split_block:
 * @channel:
 * @split_frame:
 *
 * Splits the block in @channel located at @split_frame into 2 separate blocks.
 */
void
marlin_channel_split_block (MarlinChannel *channel,
			    guint64 split_frame)
{
	g_return_if_fail (channel != NULL);
	
	WRITE_LOCK (channel->lock);
	lockless_split_block (channel, split_frame);
	WRITE_UNLOCK (channel->lock);
}

struct _unlink_closure {
	MarlinChannel *channel;
	MarlinBlock *original;
	guint64 start_frame, end_frame;
};

/* FIXME: Do we need to copy all the block lists around? */
static void
unlink_undo (gpointer data)
{
	struct _unlink_closure *c = data;
	MarlinBlock *prev_block;
	MarlinBlock *next_block;
	MarlinBlock *first_block;
	MarlinBlock *last_block;

	WRITE_LOCK (c->channel->lock);

	if (c->channel->first == NULL) {
		c->channel->first = marlin_block_copy_list (c->original, NULL);
		c->channel->last = marlin_block_last (c->channel->first);
	} else {
		prev_block = lockless_get_for_frame (c->channel->first, c->start_frame - 1);
		first_block = marlin_block_copy_list (c->original, NULL);
		last_block = marlin_block_last (first_block);

		if (prev_block == NULL) {
			next_block = c->channel->first;
			c->channel->first = first_block;
			first_block->previous = NULL;
		} else {
			next_block = prev_block->next;
			prev_block->next = first_block;
			first_block->previous = prev_block;
		}

		last_block->next = next_block;
		if (next_block) {
			next_block->previous = last_block;
		}
	}

	c->channel->frames = marlin_block_recalculate_ranges (c->channel->first);

	WRITE_UNLOCK (c->channel->lock);
}

static void
unlink_redo (gpointer data)
{
	struct _unlink_closure *c = data;
	MarlinBlock *b;
	
	WRITE_LOCK (c->channel->lock);
	lockless_unlink_range (c->channel, c->start_frame, c->end_frame,
			       &b, NULL);
	WRITE_UNLOCK (c->channel->lock);

	marlin_block_free_list (b);
}

static void
unlink_destroy (gpointer data)
{
	struct _unlink_closure *c = data;

	marlin_block_free_list (c->original);
	g_free (c);
}

/* Removes data between start_frame and finish_frame inclusive
   by unlinking the blocks that contain it, and then returns
   the lists of removed blocks and peaks in &blocks and &peaks. */
static void
lockless_unlink_range (MarlinChannel *channel,
		       guint64 start_frame,
		       guint64 finish_frame,
		       MarlinBlock **blocks,
		       MarlinUndoContext *ctxt)
{
	MarlinBlock *first_block, *last_block;

	g_return_if_fail (start_frame <= channel->frames);
	g_return_if_fail (finish_frame <= channel->frames);
	g_return_if_fail (finish_frame >= start_frame);

	/* Split on the start and finish frames */
	if (finish_frame != channel->frames) {
		lockless_split_block (channel, finish_frame + 1);
	}

	lockless_split_block (channel, start_frame);

	first_block = lockless_get_for_frame (channel->first, start_frame);
	g_assert (first_block != NULL);

	last_block = lockless_get_for_frame (channel->first, finish_frame);
	g_assert (last_block != NULL);

	/* Unlink these blocks */
	if (channel->first == first_block) {
		channel->first = marlin_block_next (last_block);
	}

	if (channel->last == last_block) {
		channel->last = marlin_block_previous (first_block);
	}

	/* Need to lock the blocks */
	if (first_block->previous != NULL) {
		first_block->previous->next = last_block->next;
	}
	
	if (last_block->next != NULL) {
		last_block->next->previous = first_block->previous;
	}
	
	last_block->next = NULL;

	/* Need to unlock the blocks */
	channel->frames = marlin_block_recalculate_ranges (channel->first);

	if (blocks) {
		*blocks = first_block;
	} else {
		/* FIXME: Destroy the blocks here? */
	}

	if (ctxt) {
		MarlinUndoable *u;
		struct _unlink_closure *c;

		c = g_new (struct _unlink_closure, 1);
		c->channel = channel;
		c->original = marlin_block_copy_list (first_block, NULL);
		c->start_frame = start_frame;
		c->end_frame = finish_frame;

		u = marlin_undoable_new (unlink_undo,
					 unlink_redo,
					 unlink_destroy,
					 c);
		marlin_undo_context_add (ctxt, u);
	}
}

struct _link_closure {
	MarlinChannel *channel;
	MarlinBlock *blocks, *l_block;
	guint64 insert_frame, num_frames;
};

static void
link_undo (gpointer data)
{
	MarlinChannel *channel;
	struct _link_closure *c = data;

	channel = c->channel;
	WRITE_LOCK (channel->lock);

	if (channel->last == c->l_block) {
		channel->last = c->blocks->previous;
	}

	if (channel->first == c->blocks) {
		channel->first = c->l_block->next;
		
		c->l_block->next = NULL;
		channel->first->previous = NULL;
	} else {
		MarlinBlock *n, *p;

		n = c->l_block->next;
		p = c->blocks->previous;

		n->previous = p;
		p->next = n;

		c->l_block->next = NULL;
		c->blocks->previous = NULL;
	}
	channel->frames = marlin_block_recalculate_ranges (channel->first);
	WRITE_UNLOCK (channel->lock);
}

static void
link_redo (gpointer data)
{
	struct _link_closure *c = data;

	WRITE_LOCK (c->channel->lock);
	lockless_link_blocks (c->channel,
			      c->insert_frame, c->num_frames,
			      c->blocks, NULL);
	WRITE_UNLOCK (c->channel->lock);
}

static void
link_destroy (gpointer data)
{
	struct _link_closure *c = data;
	g_free (c);
}

static void
lockless_link_blocks (MarlinChannel *channel,
		      guint64 insert_frame,
		      guint64 num_frames,
		      MarlinBlock *block_list,
		      MarlinUndoContext *ctxt)
{
	MarlinBlock *f_block, *s_block, *l_block;

	g_return_if_fail (insert_frame <= channel->frames);

	l_block = marlin_block_last (block_list);

	if (ctxt) {
		MarlinUndoable *u;
		struct _link_closure *c;

		c = g_new (struct _link_closure, 1);
		c->channel = channel;

		c->insert_frame = insert_frame;
		c->num_frames = num_frames;

		c->blocks = block_list;
		c->l_block = l_block;

		u = marlin_undoable_new (link_undo,
					 link_redo,
					 link_destroy,
					 c);
		marlin_undo_context_add (ctxt, u);
	}

	if (insert_frame == 0) {
		/* If we're inserting at the start, we don't really need
		   to do anything special */

		l_block->next = channel->first;
		if (channel->first == NULL) {
			channel->last = l_block;
		} else {
			channel->first->previous = l_block;
		}
		channel->first = block_list;
	} else if (insert_frame == channel->frames - 1) {
		channel->last->next = block_list;
		block_list->previous = channel->last;
		channel->last = l_block;
	} else {
		/* Split on insert frame */
		lockless_split_block (channel, insert_frame);

		f_block = lockless_get_for_frame (channel->first, insert_frame - 1);
		g_assert (f_block != NULL);
		
		s_block = f_block->next;
		g_assert (s_block != NULL);
		
		/* Link in the blocks */
		f_block->next = block_list;
		block_list->previous = f_block;
		
		s_block->previous = l_block;
		l_block->next = s_block;
	}

	channel->frames = marlin_block_recalculate_ranges (channel->first);
}

/**
 * marlin_channel_link_blocks:
 * @channel:
 * @insert_frame:
 * @num_frames:
 * @block_list:
 * @ctxt:
 */
void
marlin_channel_link_blocks (MarlinChannel *channel,
			    guint64 insert_frame,
			    guint64 num_frames,
			    MarlinBlock *block_list,
			    MarlinUndoContext *ctxt)
{
	WRITE_LOCK (channel->lock);

	lockless_link_blocks (channel, insert_frame, num_frames,
			      block_list, ctxt);

	WRITE_UNLOCK (channel->lock);
}
 
/* Deletes between start_frame and finish_frame inclusive */
static void
lockless_delete_range (MarlinChannel *channel,
		       guint64 start_frame,
		       guint64 finish_frame,
		       MarlinUndoContext *ctxt)
{
	MarlinBlock *blocks;

	lockless_unlink_range (channel, start_frame, finish_frame, 
			       &blocks, ctxt);
	marlin_block_free_list (blocks);
}

struct _insert_data_closure {
	MarlinChannel *channel;
	MarlinBlock *blocks;
	guint64 insert_frame;
	guint64 num_frames;
};

static void
insert_data_undo (gpointer data)
{
	struct _insert_data_closure *c = data;

	WRITE_LOCK (c->channel->lock);
	lockless_unlink_range (c->channel, c->insert_frame,
			       c->insert_frame + (c->num_frames - 1),
			       NULL, NULL);
	WRITE_UNLOCK (c->channel->lock);
}

static void
insert_data_redo (gpointer data)
{
	struct _insert_data_closure *c = data;
	MarlinBlock *prev_block;
	MarlinBlock *next_block;
	MarlinBlock *first_block;
	MarlinBlock *last_block;
	
	WRITE_LOCK (c->channel->lock);

	if (c->channel->first == NULL) {
		c->channel->first = marlin_block_copy_list (c->blocks, NULL);
		c->channel->last = marlin_block_last (c->channel->first);
	} else {
		prev_block = lockless_get_for_frame (c->channel->first, c->insert_frame - 1);
		first_block = marlin_block_copy_list (c->blocks, NULL);
		last_block = marlin_block_last (first_block);

		if (prev_block == NULL) {
			next_block = c->channel->first;
			c->channel->first = first_block;
			first_block->previous = NULL;
		} else {
			next_block = prev_block->next;
			prev_block->next = first_block;
			first_block->previous = prev_block;
		}

		last_block->next = next_block;
		if (next_block) {
			next_block->previous = last_block;
		}
	}

	c->channel->frames = marlin_block_recalculate_ranges (c->channel->first);

	WRITE_UNLOCK (c->channel->lock);	
}

static void
insert_data_destroy (gpointer data)
{
	struct _insert_data_closure *c = data;

	marlin_block_free_list (c->blocks);
	g_free (c);
}

/* Splits the block on the insert_frame, so insert_frame will be first in
   new block. Then it gets the block for insert_frame - 1, and then inserts 
   data */
static gboolean
lockless_insert_data (MarlinChannel *channel,
		      float *data,
		      guint64 num_frames,
		      guint64 insert_frame,
		      MarlinUndoContext *ctxt,
		      GError **error)
{
	MarlinBlock *prev, *next;
	MarlinBlock *block, *b;
	MarlinBlock *ps = NULL, *pe = NULL;
	struct _insert_data_closure *c = NULL;
	MarlinUndoable *u;
	guint64 frames;
	gboolean ret;
	
	g_return_val_if_fail (insert_frame <= channel->frames, FALSE);

	frames = channel->frames;
	block = marlin_channel_create_block (channel);
	ret = marlin_block_set_data (block, data, num_frames, error);

	if (ret == FALSE) {
		return FALSE;
	}

	if (ctxt) {
		MarlinUndoable *u;

		c = g_new (struct _insert_data_closure, 1);
		c->channel = channel;
		c->num_frames = num_frames;
		c->insert_frame = insert_frame;
		c->blocks = marlin_block_copy_list (block, NULL);

		u = marlin_undoable_new (insert_data_undo,
					 insert_data_redo,
					 insert_data_destroy,
					 c);

		marlin_undo_context_add (ctxt, u);

	}

	if (channel->first == NULL) {
		channel->first = block;
		channel->last = block;
	} else {
		if (insert_frame != frames && insert_frame != 0) {
			lockless_split_block (channel, insert_frame);
		}

		if (insert_frame != 0) {
			prev = lockless_get_for_frame (channel->first, insert_frame - 1);
			g_assert (prev != NULL);
			
			next = prev->next;
			
			WRITE_LOCK (prev->lock);
			
			prev->next = block;
			block->previous = prev;
			WRITE_UNLOCK (prev->lock);
		} else {
			next = channel->first;
			channel->first = block;
		}

		block->next = next;
		if (next != NULL) {
			WRITE_LOCK (next->lock);
			next->previous = block;
			WRITE_UNLOCK (next->lock);
		} else {
			channel->last = block;
		}
	}

	channel->frames = marlin_block_recalculate_ranges (channel->first);

	return TRUE;
}

/**
 * marlin_channel_insert_data:
 * channel:
 * data:
 * num_frames:
 * insert_frame:
 * error:
 *
 * Inserts @num_frames pointed to by @data into @channel starting at
 * @insert_frame.
 *
 * Returns: TRUE on success, FALSE on failure with details in @error.
 */
gboolean
marlin_channel_insert_data (MarlinChannel *channel,
			    float *data,
			    guint64 num_frames,
			    guint64 insert_frame,
			    MarlinUndoContext *ctxt,
			    GError **error)
{
	gboolean ret;

	WRITE_LOCK (channel->lock);

	ret = lockless_insert_data (channel, data, num_frames, insert_frame, ctxt, error);

	WRITE_UNLOCK (channel->lock);

	return ret;
}

/**
 * marlin_channel_delete_range:
 * channel:
 * start_frame:
 * finish_frame:
 *
 * Deletes the data between @start_frame and @finish_frame from @channel.
 */
void
marlin_channel_delete_range (MarlinChannel *channel,
			     guint64 start_frame,
			     guint64 finish_frame,
			     MarlinUndoContext *ctxt)
{
	g_return_if_fail (channel != NULL);

	WRITE_LOCK (channel->lock);
	lockless_delete_range (channel, start_frame, finish_frame, ctxt);
	WRITE_UNLOCK (channel->lock);
}

static gboolean
lockless_insert_silence (MarlinChannel *channel,
			 MarlinOperation *operation,
			 guint64 insert_frame,
			 guint64 num_frames,
			 MarlinUndoContext *ctxt,
			 GError **error)
{
	float *data;
	guint64 length;
	
	g_return_val_if_fail (insert_frame <= channel->frames, FALSE);
	
	if (num_frames == 0) {
		return TRUE;
	}
	
	/* Make a space for the zer0 */
	data = g_new0 (float, MARLIN_BLOCK_SIZE);
	
	/* Make blocks out of the silence */
	length = num_frames;
	
	while (length > (guint64) 0) {
		guint64 len_to_copy;
		gboolean ret;
		
		len_to_copy = MIN (MARLIN_BLOCK_SIZE, length);
	
		ret = lockless_insert_data (channel, data, 
					    len_to_copy, insert_frame, 
					    ctxt, error);
		if (ret == FALSE) {
			g_free (data);
			return FALSE;
		}

		insert_frame += len_to_copy;
		length -= len_to_copy;
	}

	g_free (data);

	return TRUE;
}

/**
 * marlin_channel_insert_silence:
 * channel:
 * operation:
 * insert_frame:
 * num_frames:
 * ctxt:
 * error:
 *
 * Inserts @num_frames of silence into @channel at @insert_frame.
 *
 * Returns: TRUE on success, FALSE on failure with details in @error.
 */
gboolean
marlin_channel_insert_silence (MarlinChannel *channel,
			       MarlinOperation *operation,
			       guint64 insert_frame,
			       guint64 num_frames,
			       MarlinUndoContext *ctxt,
			       GError **error)
{
	gboolean ret;

	g_return_val_if_fail (channel != NULL, FALSE);

	WRITE_LOCK (channel->lock);
	ret = lockless_insert_silence (channel, operation, insert_frame, num_frames, ctxt, error);
	WRITE_UNLOCK (channel->lock);

	return ret;
}

/**
 * marlin_channel_clear_range:
 * channel:
 * operation:
 * start_frame:
 * end_frame:
 * error:
 *
 * Clears the @channel between @start_frame and @end_frame
 * In reality it deletes the range and then adds silence
 *
 * Returns: TRUE on success, FALSE on failure with details in @error.
 */
gboolean
marlin_channel_clear_range (MarlinChannel *channel,
			    MarlinOperation *operation,
			    guint64 start_frame,
			    guint64 finish_frame,
			    MarlinUndoContext *ctxt,
			    GError **error)
{
	gboolean ret;

	g_return_val_if_fail (channel != NULL, FALSE);

	WRITE_LOCK (channel->lock);

	lockless_delete_range (channel, start_frame, finish_frame, ctxt);
	ret = lockless_insert_silence (channel, operation, start_frame, 
				       (finish_frame - start_frame) + 1,
				       ctxt, error);

	WRITE_UNLOCK (channel->lock);

	return ret;
}

/**
 * marlin_channel_insert:
 * dest:
 * src:
 * start_frame:
 * error:
 *
 * Inserts the data in @src into @dest at @start_frame.
 *
 * Returns: TRUE on success, FALSE on failure with details in @error.
 */
void
marlin_channel_insert (MarlinChannel *dest,
		       MarlinChannel *src,
		       guint64 start_frame,
		       MarlinUndoContext *ctxt)
{
	MarlinBlock *blocks, *b;
	guint64 count;

	READ_LOCK (src->lock);
	WRITE_LOCK (dest->lock);

	blocks = marlin_block_copy_list (src->first, &count);

	/* Move the blocks to the dest channel */
	marlin_block_move_blocks_list (blocks, dest);

	lockless_link_blocks (dest, start_frame, count, blocks, ctxt);

	WRITE_UNLOCK (dest->lock);
	READ_UNLOCK (src->lock);
}

/**
 * marlin_channel_crop_range:
 * channel:
 * start_frame:
 * finish_frame:
 *
 * Crops the area defined by @start_frame and @finish_frame (inclusive)
 * from @channel.
 */
void
marlin_channel_crop_range (MarlinChannel *channel,
			   guint64 start_frame,
			   guint64 finish_frame,
			   MarlinUndoContext *ctxt) 
{
	MarlinBlock *first_block, *last_block;
	MarlinBlock *current;

	g_return_if_fail (start_frame <= channel->frames);
	g_return_if_fail (finish_frame <= channel->frames);
	g_return_if_fail (finish_frame >= start_frame);

	WRITE_LOCK (channel->lock);

	if (finish_frame + 1 < channel->frames) {
		lockless_delete_range (channel, finish_frame + 1, channel->frames - 1, ctxt);
	}

	if (start_frame >= (guint64) 1) {
		lockless_delete_range (channel, (guint64) 0, start_frame - 1, ctxt);
	}
	
	WRITE_UNLOCK (channel->lock);
}

/**
 * marlin_channel_copy_data:
 * src_channel:
 * dest_channel:
 * start_frame:
 * end_frame:
 * error:
 *
 * Copy the data between @start_frame and @end_frame (inclusive) from 
 * @src_channel into @dest_channel.
 *
 * Returns: TRUE on success, FALSE on failure with details in @error.
 */
gboolean
marlin_channel_copy_data (MarlinChannel *src_channel,
			  MarlinChannel *dest_channel,
			  guint64 start_frame,
			  guint64 end_frame,
			  GError **error)
{
	MarlinBlock *src_block, *dest_block = NULL, *db_end, *block;
	guint64 frames_needed;
	float *buf;
	off_t offset;
	
	g_return_val_if_fail (src_channel != dest_channel, FALSE);

	if (end_frame >= src_channel->frames) {
		end_frame = src_channel->frames - 1;
	}

	frames_needed = (end_frame - start_frame) + 1;

	READ_LOCK (src_channel->lock);
	WRITE_LOCK (dest_channel->lock);
	
	src_block = lockless_get_for_frame (src_channel->first, start_frame);
	g_assert (src_block);

	buf = g_new0 (float, MARLIN_BLOCK_SIZE);

	READ_LOCK (src_block->lock);
	
	while (frames_needed > (guint64) 0) {
		guint64 length_got;

		length_got = (guint64) 0;

		while (length_got < MARLIN_BLOCK_SIZE &&
		       frames_needed > (guint64) 0) {
			guint64 buf_needed, length, pos_in_block;
			guint64 length_needed, left_in_block;
			
			float *data;

			if (start_frame > src_block->end) {
				/* We need a new block */
				READ_UNLOCK (src_block->lock);
				
				src_block = marlin_block_next (src_block);
				g_assert (src_block != NULL);
				
				READ_LOCK (src_block->lock);
				start_frame = src_block->start;
			}
			
			data = marlin_block_get_frame_data (src_block);

			pos_in_block = start_frame - src_block->start;
			left_in_block = (src_block->end - start_frame) + 1;
			
			buf_needed = MARLIN_BLOCK_SIZE - length_got;
			
			/* length_needed is the MIN of what is needed to fill
			   buffer and what is needed to fill the whole thing */
			length_needed = MIN (buf_needed, frames_needed);

			/* Length is the MIN of the amount needed and the
			   amount we can copy */
			length = MIN (length_needed, left_in_block);
			
			/* Now we copy length into the buf */
			memcpy (buf + length_got, data + pos_in_block,
				length * sizeof (float));

			start_frame += length;
			length_got += length;
			frames_needed -= length;

			/* Check if we've filled frames_needed */
			if (frames_needed == 0) {
				break;
			}
		}

		block = marlin_channel_create_block (dest_channel);
		marlin_block_set_data (block, buf, length_got, NULL);

		if (dest_block == NULL) {
			dest_block = block;
			db_end = block;
		} else {
			marlin_block_append (db_end, block);
			db_end = block;
		}

		memset (buf, 0, MARLIN_BLOCK_SIZE * sizeof (float));
	}

	READ_UNLOCK (src_block->lock);

	dest_channel->first = dest_block;
	dest_channel->last = db_end;
	dest_channel->frames = marlin_block_recalculate_ranges (dest_channel->first);

	g_free (buf);

	READ_UNLOCK (src_channel->lock);
	WRITE_UNLOCK (dest_channel->lock);

	return TRUE;
}

/**
 * marlin_channel_mix:
 * @dest:
 * @src:
 * @src_db:
 * @dest_db:
 * @start_frame:
 * @end_frame:
 * @clip:
 * @operation:
 * @ctxt:
 * @error:
 *
 * Takes the data from @src and mixes it into @dest at @start_frame.
 * If @clip is TRUE, the data is only mixed up until @end_frame,
 * otherwise all data in @src is mixed in.
 *
 * Returns: TRUE on success, FALSE on failure with details in @error.
 */
gboolean
marlin_channel_mix (MarlinChannel *dest,
		    MarlinChannel *src,
		    double src_db,
		    double dest_db,
		    guint64 start_frame,
		    guint64 end_frame,
		    gboolean clip,
		    MarlinOperation *operation,
		    MarlinUndoContext *ctxt,
		    GError **error)
{
	MarlinBlock *s_block, *d_block, *d;
	float s_ratio, d_ratio;
	guint64 dest_length, range, limit;
	guint64 pos_in_src, pos_in_dest, current, position;
	float *s_data, *d_data, *buf;
	gboolean ret = TRUE;

	g_return_val_if_fail (dest != NULL, FALSE);
	g_return_val_if_fail (src != NULL, FALSE);

	READ_LOCK (src->lock);
	WRITE_LOCK (dest->lock);

	/* If clip == TRUE, then we're only going to end_frame
	   otherwise we go to the end of the sample */
	limit = clip ? end_frame : MIN (dest->frames - 1, start_frame + src->frames - 1);
	dest_length = (limit - start_frame) + 1;
	range = clip ? dest_length : src->frames;

	/* Unlink all the blocks we're going to mix. */
	lockless_unlink_range (dest, start_frame, limit, &d, ctxt);
	d_block = d;

	if (src_db == MARLIN_INFINITE_DB) {
		s_ratio = 0;
	} else {
		s_ratio = pow (10.0, src_db / 20.0);
	}

	if (dest_db == MARLIN_INFINITE_DB) {
		d_ratio = 0;
	} else {
		d_ratio = pow (10.0, dest_db / 20.0);
	}

	s_block = src->first;
	pos_in_src = (guint64) 0;

	READ_LOCK (s_block->lock);

	s_data = marlin_block_get_frame_data (s_block);

	buf = g_new (float, MARLIN_BLOCK_SIZE);
	current = start_frame;
	while (range > 0) {
		guint64 amount = MIN (MARLIN_BLOCK_SIZE, range);
		
		range -= amount;

		if (d_block != NULL) {
			WRITE_LOCK (d_block->lock);
			
			pos_in_dest = start_frame - d_block->start;
			d_data = marlin_block_get_frame_data (d_block);
		} else {
			pos_in_dest = 0;
			d_data = NULL;
		}

		for (position = 0; position < amount; position++) {
			float d;

			if (d_block && pos_in_dest + d_block->start > d_block->end) {
				WRITE_UNLOCK (d_block->lock);
				
				/* Get next block */
				d_block = marlin_block_next (d_block);

				/* This is not allowed if we're clipping */
				if (clip == TRUE) {
					g_assert (d_block != NULL);
					
					WRITE_LOCK (d_block->lock);
				
					pos_in_dest = 0;
					d_data = marlin_block_get_frame_data (d_block);
				} else {
					d_data = NULL;
				}
			}
		
			if (pos_in_src + s_block->start > s_block->end) {
				READ_UNLOCK (s_block->lock);
				
				/* Get next block */
				s_block = marlin_block_next (s_block);
				g_assert (s_block != NULL);
			
				READ_LOCK (s_block->lock);
				
				pos_in_src = 0;
				s_data = marlin_block_get_frame_data (s_block);
			}
		
			/* Combine the frames */
			if (d_data == NULL && clip == FALSE) {
				d = 0.0;
			} else {
				d = d_data[pos_in_dest] * d_ratio;
			}

			buf[position] = d + (s_data[pos_in_src] * s_ratio);

			pos_in_src++;
			pos_in_dest++;
		}
		
		if (d_block != NULL) {
			WRITE_UNLOCK (d_block->lock);
		}

		ret = lockless_insert_data (dest, buf, amount, current, ctxt, error);
		if (ret == FALSE) {
			goto err;
		}

		current += amount;
	}

 err:
	marlin_block_free_list (d);

	READ_UNLOCK (s_block->lock);
	g_free (buf);

	WRITE_UNLOCK (dest->lock);
	READ_UNLOCK (src->lock);

	return ret;
}

static gboolean
lockless_channel_multiply (MarlinChannel *channel,
			   guint64 start,
			   guint64 finish,
			   MarlinFadeFunc ratio_func,
			   gpointer closure,
			   MarlinUndoContext *ctxt,
			   GError **error)
{
	MarlinBlock *block;
	guint64 range, pos_in_block, position;
	float *data, *buf;
	guint64 current;
	gboolean ret;
	
	range = finish - start;

	buf = g_new (float, MARLIN_BLOCK_SIZE);
	current = start;
	while (range > 0) {
		guint64 amount = MIN (MARLIN_BLOCK_SIZE, range);
		
		range -= amount;

		block = lockless_get_for_frame (channel->first, current);

		WRITE_LOCK (block->lock);
		
		pos_in_block = current - block->start;
		data = marlin_block_get_frame_data (block);
		
		for (position = 0; position < amount; position++) {
			if (pos_in_block + block->start > block->end) {
				WRITE_UNLOCK (block->lock);
				
				block = marlin_block_next (block);
				g_assert (block != NULL);
				
				WRITE_LOCK (block->lock);
				
				pos_in_block = (guint64) 0;
				data = marlin_block_get_frame_data (block);
			}
			
			/* Apply the ratio */
			buf[position] = data[pos_in_block] * ratio_func (position, closure);
			pos_in_block++;
		}
		
		WRITE_UNLOCK (block->lock);
		
		/* Now we need to delete the old block that we've just
		   adjusted and then insert the new block */
		lockless_delete_range (channel, current, (current + amount) - 1, ctxt);
		ret = lockless_insert_data (channel, buf, amount, current, ctxt, error);
		if (ret == FALSE) {
			g_free (buf);
			return FALSE;
		}

		current += amount;
	}

	g_free (buf);
	return TRUE;
}

static float
invert_func (guint64 position,
	     gpointer data)
{
	return -1.0;
}

/**
 * marlin_channel_invert:
 * channel:
 * start:
 * finish:
 * error:
 *
 * Inverts the data between @start and @finish of @channel.
 *
 * Returns: TRUE on success, FALSE on failure with details in @error.
 */
gboolean
marlin_channel_invert (MarlinChannel *channel,
		       guint64 start,
		       guint64 finish,
		       MarlinUndoContext *ctxt,
		       GError **error)
{
	gboolean ret;
	g_return_val_if_fail (channel != NULL, FALSE);

	WRITE_LOCK (channel->lock);

	ret = lockless_channel_multiply (channel, start, finish, invert_func, NULL, ctxt, error);

	WRITE_UNLOCK (channel->lock);

	return ret;
}

struct _volume_closure {
	float ratio;
};

static float
volume_func (guint64 position,
	     gpointer data)
{
	struct _volume_closure *vc = data;
	return vc->ratio;
}

/**
 * marlin_channel_adjust_volume:
 * channel:
 * db: The new volume in decibels
 * start:
 * finish:
 * error:
 *
 * Adjusts the volume of the data between @start and @finish of @channel
 * by @db decibels.
 *
 * Returns: TRUE on success, FALSE on failure with details in @error.
 */
gboolean
marlin_channel_adjust_volume (MarlinChannel *channel,
			      float db,
			      guint64 start,
			      guint64 finish,
			      MarlinUndoContext *ctxt,
			      GError **error)
{
	struct _volume_closure *vc;
	gboolean ret;

	g_return_val_if_fail (channel != NULL, FALSE);

	vc = g_new (struct _volume_closure, 1);

	if (db == MARLIN_INFINITE_DB) {
		vc->ratio = 0.0;
	} else {
		vc->ratio = pow (10.0, db / 10.0);
	}

	WRITE_LOCK (channel->lock);

	ret = lockless_channel_multiply (channel, start, finish, volume_func, vc, ctxt, error);

	WRITE_UNLOCK (channel->lock);

	g_free (vc);
	return ret;
}

/**
 * marlin_channel_fade:
 * @channel: The channel.
 * @start: Start frame.
 * @finish: Finish frame.
 * @fade_func: The function to get the fade level from.
 * @closure: The data to be passed to the fade_func.
 * @operation:
 * @ctxt:
 * @error:
 *
 * Change the level of @channel between @start and @finish (inclusive)
 * using @fade_func as the function to generate new values.
 *
 * Returns: TRUE on success, FALSE on failure with details in @error.
 */
gboolean
marlin_channel_fade (MarlinChannel *channel,
		     guint64 start,
		     guint64 finish,
		     MarlinFadeFunc fade_func,
		     gpointer closure,
		     MarlinOperation *operation,
		     MarlinUndoContext *ctxt,
		     GError **error)
{
	gboolean ret;

	g_return_val_if_fail (channel != NULL, FALSE);

	WRITE_LOCK (channel->lock);

	ret = lockless_channel_multiply (channel, start, finish, fade_func, closure, ctxt, error);

	WRITE_UNLOCK (channel->lock);

	return ret;
}

/* FIXME: Is this function needed anymore? */
/**
 * marlin_channel_is_ready:
 * @channel: The channel.
 *
 * Checks if @channel is mmapped and peaks are generated.
 *
 * Returns: TRUE if @channel is ready, FALSE otherwise.
 */
gboolean
marlin_channel_is_ready (MarlinChannel *channel)
{
	return (channel->first != NULL);
}

/**
 * marlin_channel_next_zero:
 * @channel: The channel.
 * @position:
 *
 */
guint64
marlin_channel_next_zero (MarlinChannel *channel,
			  guint64 position)
{
	MarlinBlock *block;
	guint64 pos_in_block, orig;
	float *data;
	float d, dm, dp;

	g_return_val_if_fail (channel != NULL, position);
	g_return_val_if_fail (position <= channel->frames, position);

	READ_LOCK (channel->lock);

	block = lockless_get_for_frame (channel->first, position);
	g_assert (block);

	/* Store position so we can return it if we can't find any more */
	orig = position;

 start:
	READ_LOCK (block->lock);
	data = marlin_block_get_frame_data (block);

	pos_in_block = position - block->start;

	d = data[pos_in_block];

	if (position != channel->frames - 1) {
		if (position + 1 > block->end) {
			MarlinBlock *tb = marlin_block_next (block);
			float *tdata;
			
			READ_LOCK (tb->lock);
			
			tdata = marlin_block_get_frame_data (tb);
			
			dp = tdata[(position + 1) - tb->start];
			
			READ_UNLOCK (tb->lock);
		} else {
			dp = data[pos_in_block + 1];
		}
	} else {
		dp = 0.0;
	}

	if (position != 0) {
		if (position - 1 < block->start) {
			MarlinBlock *tb = marlin_block_previous (block);
			float *tdata;
			
			READ_LOCK (tb->lock);
			
			tdata = marlin_block_get_frame_data (tb);
			
			dm = tdata[(position - 1) - tb->start];
			
			READ_UNLOCK (tb->lock);
		} else {
			dm = data[pos_in_block - 1];
		}
	} else {
		dm = 0.0;
	}

	READ_UNLOCK (block->lock);

	if (d == 0.0 && dm != 0.0 && dp != 0.0) {
		/* Found zer0 crossing 
		   Don't do anything else */
	} else {
		if (position != channel->frames - 1) {
			position++;
			
			if (position - block->start > block->end) {
				block = marlin_block_next (block);
			}
			
			goto start;
		} else {
			position = orig;
		}
	}

	READ_UNLOCK (channel->lock);

	return position;
}

/**
 * marlin_channel_previous_zero:
 * @channel:
 * @position:
 *
 */
guint64
marlin_channel_previous_zero (MarlinChannel *channel,
			      guint64 position)
{
	MarlinBlock *block;
	guint64 pos_in_block, orig;
	float *data;
	float d, dm, dp;

	g_return_val_if_fail (channel != NULL, position);
	g_return_val_if_fail (position <= channel->frames, position);

	READ_LOCK (channel->lock);

	block = lockless_get_for_frame (channel->first, position);
	g_assert (block);

	orig = position;

 start:
	READ_LOCK (block->lock);
	data = marlin_block_get_frame_data (block);

	pos_in_block = position - block->start;

	d = data[pos_in_block];
	if (position != channel->frames - 1) {
		if (position + 1 > block->end) {
			MarlinBlock *tb = marlin_block_next (block);
			float *tdata;
			
			READ_LOCK (tb->lock);
			
			tdata = marlin_block_get_frame_data (tb);
			
			dp = tdata[(position + 1) - tb->start];
			
			READ_UNLOCK (tb->lock);
		} else {
			dp = data[pos_in_block + 1];
		}
	} else {
		dp = 0.0;
	}

	if (position != 0) {
		if (position - 1 < block->start) {
			MarlinBlock *tb = marlin_block_previous (block);
			float *tdata;
			
			READ_LOCK (tb->lock);
			
			tdata = marlin_block_get_frame_data (tb);
			
			dm = tdata[(position - 1) - tb->start];
			
			READ_UNLOCK (tb->lock);
		} else {
			dm = data[pos_in_block - 1];
		}
	} else {
		dm = 0.0;
	}

	if (d == 0.0 && dm != 0.0 && dp != 0.0) {
		/* Found zer0 crossing */
		READ_UNLOCK (block->lock);
	} else {
		if (position != 0) {
			position--;
			
			if (position - block->start > block->end) {
				READ_UNLOCK (block->lock);
				
				block = marlin_block_next (block);
			}
			
			goto start;
		} else {
			position = orig;
		}
	}

	READ_UNLOCK (channel->lock);

	return position;
}

/**
 * marlin_channel_reverse:
 * @channel:
 * @start:
 * @finish:
 * @operation:
 * @ctxt:
 * @error:
 *
 */
gboolean
marlin_channel_reverse_range (MarlinChannel *channel,
			      guint64 start,
			      guint64 finish,
			      MarlinOperation *operation,
			      MarlinUndoContext *ctxt,
			      GError **error)
{
	MarlinBlock *block, *reversed = NULL, *rb;
	guint64 range, pos_in_block, position;
	float *data, *buf;
	guint64 current;
	gboolean ret;
	
	range = (finish - start) + 1;
	
	buf = g_new (float, MARLIN_BLOCK_SIZE);
	current = finish;

	WRITE_LOCK (channel->lock);

	while (range > 0) {
		MarlinBlock *tblock;
		guint64 amount = MIN (MARLIN_BLOCK_SIZE, range);
		
		range -= amount;

		block = lockless_get_for_frame (channel->first, current);

		WRITE_LOCK (block->lock);
		
		pos_in_block = (current - block->start) + 1;
		data = marlin_block_get_frame_data (block);
		
		for (position = 0; position < amount; ++position) {
			if (pos_in_block < 0) {
				WRITE_UNLOCK (block->lock);
				
				block = marlin_block_previous (block);
				g_assert (block != NULL);
				
				WRITE_LOCK (block->lock);
				
				pos_in_block = (guint64) (block->end - block->start) + 1;
				data = marlin_block_get_frame_data (block);
			}
			
			buf[position] = data[pos_in_block];
			--pos_in_block;
		}
	
		/* Got a newly reversed block,
		   write it into the channel, and link it onto the temp list */
		tblock = marlin_channel_create_block (channel);
		ret = marlin_block_set_data (tblock, buf, amount, error);

		if (!ret) {
			marlin_block_free (tblock);
			g_free (buf);

			WRITE_UNLOCK (block->lock);
			WRITE_UNLOCK (channel->lock);

			return FALSE;
		}

		if (reversed == NULL) {
			rb = reversed = tblock;
		} else {
			marlin_block_append (rb, tblock);
			rb = tblock;
		}

		WRITE_UNLOCK (block->lock);
		
		current -= amount;
	}

	/* Now we need to delete the old block that we've just
	   adjusted and then insert the new block */
	lockless_unlink_range (channel, start, finish, NULL, ctxt);
	lockless_link_blocks (channel, start, (finish - start) + 1, 
			      reversed, ctxt);
	
	g_free (buf);

	WRITE_UNLOCK (channel->lock);
	return TRUE;
}
