/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 *  Authors: Iain Holmes <iain@prettypeople.org>
 *
 *  Copyright 2003 Iain Holmes
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of version 2 of the GNU General Public 
 *  License as published by the Free Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <math.h>

#include <glib/gi18n.h>

#include <gtk/gtkwindow.h>
#include <gtk/gtklabel.h>

#include <marlin/marlin-cross-fader.h>
#include <marlin/marlin-cursors.h>
#include <marlin/marlin-types.h>
#include <marlin/marlin-utils.h>

enum {
	PROP_0,
	PROP_SRC,
	PROP_DEST,
	PROP_START_FRAME,
	PROP_LENGTH,
	PROP_SRC_FADE,
	PROP_DEST_FADE,
	PROP_SNAP_TO_GRID,
	PROP_SCALE,
};

enum _box {
	SRC_IN,
	SRC_OUT,
	DEST_IN,
	DEST_OUT,
	NONE
};

struct _MarlinCrossFaderPrivate {
	MarlinSample *src, *dest; /* You mix src into the dest. */
	char *src_name, *dest_name;

	MarlinSampleFade *src_fader, *dest_fader; /* The faders for each
						     thing to be mixed */
	guint64 start_frame; /* The starting frame in dest */
	guint64 length; /* length of the whole crossfade in frames */

	GdkGC *src_gc, *dest_gc;
	GdkColor src_colour, dest_colour; /* src - red
					     dest - blue */

	int xofs; /* The x offset of the actual XFade */
	int yofs; /* The y offset */
	int frames_per_pixel;
	int iw; /* The ideal width that we want out XFade to be */

	enum _box current_box; /* Are we in a box? */
	
	gboolean in_drag; /* Are we dragging? */
	enum _box dragging_box; /* What box are we dragging? */
	
	GtkWidget *tooltip, *tip_label;
	char *cur_tip;

	PangoLayout *layout;

	GdkPixmap *background;
	GdkGC *non_gr_exp_gc;

	gboolean snap_to_grid;
	int x_grid, y_grid;

	MarlinScale scale;
};

static GtkWidgetClass *parent_class = NULL;

#define XFADE_HEIGHT 200
#define XFADE_WIDTH 432

static void
finalize (GObject *object)
{
	MarlinCrossFader *xfade;

	xfade = MARLIN_CROSS_FADER (object);

	if (xfade->priv == NULL) {
		return;
	}

	g_free (xfade->priv->src_fader);
	g_free (xfade->priv->dest_fader);

	if (xfade->priv->src != NULL) {
		g_object_unref (G_OBJECT (xfade->priv->src));
	}

	if (xfade->priv->dest != NULL) {
		g_object_unref (G_OBJECT (xfade->priv->dest));
	}

	if (xfade->priv->background) {
		g_object_unref (xfade->priv->background);
	}

	if (xfade->priv->non_gr_exp_gc) {
		g_object_unref (xfade->priv->non_gr_exp_gc);
	}

	g_object_unref (G_OBJECT (xfade->priv->layout));

	g_free (xfade->priv);
	xfade->priv = NULL;

	G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
snap_coords_to_grid (MarlinCrossFader *xfade,
		     int *x,
		     int *y)
{
	MarlinCrossFaderPrivate *priv = xfade->priv;
	guint64 fx;
	int fy;

	fx = (*x * priv->frames_per_pixel) + (priv->x_grid / 2);
	*x = ((fx / priv->x_grid) * priv->x_grid) / priv->frames_per_pixel;

	fy = *y + 10;
	*y = (fy / priv->y_grid) * priv->y_grid;
}

static void
move_box_to (MarlinCrossFader *xfade,
	     enum _box box,
	     int x,
	     int y)
{
	MarlinCrossFaderPrivate *priv;
	GtkWidget *widget = GTK_WIDGET (xfade);
	int height, tx;
	MarlinSampleFade *f;

	priv = xfade->priv;

	if (priv->snap_to_grid) {
		snap_coords_to_grid (xfade, &x, &y);
	}

	height = XFADE_HEIGHT;

	switch (box) {
	case NONE:
		return;

	case SRC_IN:
	case SRC_OUT:
		f = priv->src_fader;
		break;

	case DEST_IN:
	case DEST_OUT:
		f = priv->dest_fader;
		break;

	default:
		g_assert_not_reached ();
		break;
	}
	
	switch (box) {
	case SRC_IN:
	case DEST_IN:

		tx = (f->fade_end - priv->start_frame) / priv->frames_per_pixel;
		if (x >= tx) {
			x = tx - 1;
		}

		f->in_level = MAX (0.0, (float) (height - y) / (float) height);
		f->fade_start = priv->start_frame + (x * priv->frames_per_pixel);
		break;

	case SRC_OUT:
	case DEST_OUT:

		tx = (f->fade_start - priv->start_frame) / priv->frames_per_pixel;
		if (x <= tx) {
			x = tx + 1;
		}
		
		f->out_level = MAX (0.0, (float) (height - y) / (float) height);
		f->fade_end = MIN (priv->start_frame + (x * priv->frames_per_pixel), 
				   priv->start_frame + priv->length);
		break;

	default:
		g_assert_not_reached ();
	}
}

static void
snap_boxes_to_grid (MarlinCrossFader *xfade)
{
	GtkWidget *widget = GTK_WIDGET (xfade);
	MarlinCrossFaderPrivate *priv = xfade->priv;
	int height;

	height = XFADE_HEIGHT;

	move_box_to (xfade, SRC_IN, 
		     (priv->src_fader->fade_start - priv->start_frame) / priv->frames_per_pixel,
		     height - (priv->src_fader->in_level * (float) height));
	move_box_to (xfade, SRC_OUT,
		     (priv->src_fader->fade_end - priv->start_frame) / priv->frames_per_pixel,
		     height - (priv->src_fader->out_level * (float) height));
	move_box_to (xfade, DEST_IN,
		     (priv->dest_fader->fade_start - priv->start_frame) / priv->frames_per_pixel,
		     height - (priv->dest_fader->in_level * (float) height));
	move_box_to (xfade, DEST_OUT,
		     (priv->dest_fader->fade_end - priv->start_frame) / priv->frames_per_pixel,
		     height - (priv->dest_fader->out_level * (float) height));

	if (GTK_WIDGET_DRAWABLE (widget)) {
		GdkRectangle rect;

		rect.x = 0;
		rect.y = 0;
		rect.width = widget->allocation.width;
		rect.height = widget->allocation.height;

		gdk_window_invalidate_rect (widget->window, &rect, FALSE);
	}		
}

static void
set_property (GObject *object,
	      guint prop_id,
	      const GValue *value,
	      GParamSpec *pspec)
{
	MarlinCrossFader *xfade;
	MarlinCrossFaderPrivate *priv;
	MarlinSample *sample;
	MarlinSampleFade *f;
	GtkWidget *widget = GTK_WIDGET (object);
	
	xfade = MARLIN_CROSS_FADER (object);
	priv = xfade->priv;
	
	switch (prop_id) {
	case PROP_SRC:
		sample = g_value_get_object (value);
		if (priv->src == sample) {
			return;
		} else {
			if (priv->src) {
				g_object_unref (G_OBJECT (priv->src));
			}

			priv->src = sample;
			g_object_ref (G_OBJECT (sample));
		}

		g_object_get (G_OBJECT (sample),
			      "total_frames", &priv->length,
			      NULL);

		priv->dest_fader->fade_end = priv->start_frame + priv->length;
		priv->src_fader->fade_end = priv->start_frame + priv->length;

		break;

	case PROP_DEST:
		sample = g_value_get_object (value);
		if (priv->dest == sample) {
			return;
		} else {
			if (priv->dest) {
				g_object_unref (G_OBJECT (priv->dest));
			}

			priv->dest = sample;
			g_object_ref (G_OBJECT (sample));
		}

		break;

	case PROP_START_FRAME:
		priv->start_frame = g_value_get_uint64 (value);

		priv->dest_fader->fade_start = priv->start_frame;
		priv->dest_fader->fade_end = priv->start_frame + priv->length;
		priv->src_fader->fade_start = priv->start_frame;
		priv->src_fader->fade_end = priv->start_frame + priv->length;
		break;

	case PROP_LENGTH:
		break;

	case PROP_SRC_FADE:
		f = g_value_get_pointer (value);
		
		priv->src_fader->in_level = f->in_level;
		priv->src_fader->out_level = f->out_level;
		priv->src_fader->fade_start = f->fade_start;
		priv->src_fader->fade_end = f->fade_end;

		if (GTK_WIDGET_DRAWABLE (widget)) {
			GdkRectangle rect;

			rect.x = priv->xofs;
			rect.y = priv->yofs;
			rect.width = priv->iw;
			rect.height = XFADE_HEIGHT;

			gdk_window_invalidate_rect (widget->window, &rect, FALSE);
		}
		
		break;

	case PROP_DEST_FADE:
		f = g_value_get_pointer (value);
		
		priv->dest_fader->in_level = f->in_level;
		priv->dest_fader->out_level = f->out_level;
		priv->dest_fader->fade_start = f->fade_start;
		priv->dest_fader->fade_end = f->fade_end;
		
		if (GTK_WIDGET_DRAWABLE (widget)) {
			GdkRectangle rect;

			rect.x = priv->xofs;
			rect.y = priv->yofs;
			rect.width = priv->iw;
			rect.height = XFADE_HEIGHT;

			gdk_window_invalidate_rect (widget->window, &rect, FALSE);
		}		

		break;

	case PROP_SNAP_TO_GRID:
		priv->snap_to_grid = g_value_get_boolean (value);
		if (priv->snap_to_grid) {
			snap_boxes_to_grid (xfade);
		}

		break;

	case PROP_SCALE:
		priv->scale = g_value_get_enum (value);
		break;

	default:
		break;
	}
}

static void
get_property (GObject *object,
	      guint prop_id,
	      GValue *value,
	      GParamSpec *pspec)
{
	MarlinCrossFader *xfade;
	MarlinCrossFaderPrivate *priv;
	
	xfade = MARLIN_CROSS_FADER (object);
	priv = xfade->priv;
	
	switch (prop_id) {

	case PROP_SRC:
		g_value_set_object (value, priv->src);
		break;

	case PROP_DEST:
		g_value_set_object (value, priv->dest);
		break;

	case PROP_START_FRAME:
		g_value_set_uint64 (value, priv->start_frame);
		break;

	case PROP_LENGTH:
		g_value_set_uint64 (value, priv->length);
		break;

	case PROP_SRC_FADE:
		/* The src fader needs to be adjusted back to the clipboard */
		priv->src_fader->fade_start -= priv->start_frame;
		priv->src_fader->fade_end -= priv->start_frame;
		g_value_set_pointer (value, priv->src_fader);
		break;

	case PROP_DEST_FADE:
		g_value_set_pointer (value, priv->dest_fader);
		break;

	case PROP_SNAP_TO_GRID:
		g_value_set_boolean (value, priv->snap_to_grid);
		break;

	case PROP_SCALE:
		g_value_set_enum (value, priv->scale);
		break;

	default:
		break;
	}
}

static void
redraw_background (MarlinCrossFader *xfade)
{
	MarlinCrossFaderPrivate *priv = xfade->priv;
	GtkWidget *widget = GTK_WIDGET (xfade);
	int i;
	GdkRectangle rect;
	guint64 division;
	
	rect.x = 0;
	rect.y = 0;
	rect.width = priv->iw;
	rect.height = (widget->allocation.height - priv->yofs);
	
	gtk_paint_flat_box (widget->style, priv->background,
			    GTK_STATE_NORMAL, GTK_SHADOW_NONE,
			    &rect, widget, "marlin-xfade-background",
			    0, 0, priv->iw, rect.height);

	priv->y_grid = 20;

	for (i = 0; i <= 100; i += 10) {
		GdkGC *gc = (i == 0 || i == 100) ? widget->style->black_gc :
			widget->style->dark_gc[GTK_STATE_NORMAL];
		
		gdk_draw_line (priv->background, gc,
			       0, i * 2,
			       priv->iw, i * 2);
	}

	division = priv->length / (guint64) 10;
	for (i = 0; i <= 10; i++) {
		GdkGC *gc = (i == 0 || i == 10) ? widget->style->black_gc :
			widget->style->dark_gc[GTK_STATE_NORMAL];

		gdk_draw_line (priv->background, gc,
			       (i * division) / priv->frames_per_pixel, 0,
			       (i * division) / priv->frames_per_pixel, XFADE_HEIGHT);
	};  

	priv->x_grid = division;
}

static void
make_pixmap (GtkWidget *widget)
{
	MarlinCrossFader *xfade = MARLIN_CROSS_FADER (widget);
	MarlinCrossFaderPrivate *priv = xfade->priv;
	int width, height, aw, ah;

	aw = priv->iw;
	ah = widget->allocation.height - priv->yofs;
	
	if (priv->background) {
		gdk_drawable_get_size (priv->background, &width, &height);
		if (width == aw &&
		    height == ah) {
			return;
		}
	
		g_object_unref (priv->background);
	}

	priv->background = gdk_pixmap_new (widget->window,
					   aw, ah, -1);

	if (priv->non_gr_exp_gc == NULL) {
		priv->non_gr_exp_gc = gdk_gc_new (widget->window);

		gdk_gc_set_exposures (priv->non_gr_exp_gc, FALSE);
	}

	redraw_background (xfade);
}

static void
size_allocate (GtkWidget *widget,
	       GtkAllocation *allocation)
{
	MarlinCrossFader *xfade = MARLIN_CROSS_FADER (widget);
	double fpp, ppp;
	int aw, tw;
	PangoRectangle log_rect;
	
	if (GTK_WIDGET_REALIZED (widget)) {
		GdkRectangle rect;

		gdk_window_move_resize (widget->window,
					allocation->x, allocation->y,
					allocation->width, allocation->height);

		make_pixmap (widget);
		
		rect.x = 0;
		rect.y = 0;
		rect.width = allocation->width;
		rect.height = allocation->height;

		gdk_window_invalidate_rect (widget->window, &rect, FALSE);
	}

	parent_class->size_allocate (widget, allocation);
}

static void
size_request (GtkWidget *widget,
	      GtkRequisition *req)
{
	MarlinCrossFader *xfade = MARLIN_CROSS_FADER (widget);
	PangoRectangle log_rect;
	double fpp;
	int h;
	
	pango_layout_get_extents (xfade->priv->layout, NULL, &log_rect);
	xfade->priv->yofs = PANGO_PIXELS (log_rect.height) + 4;

	/* iw is the ideal width:
	   width will not always be an even number of frames per pixel
	   so when we multiply up, iw will be the width that is needed
	   for an even number. We then centre iw inside width */
	fpp = (double) xfade->priv->length / (double) XFADE_WIDTH;
	xfade->priv->frames_per_pixel = (int) fpp;

	pango_layout_set_text (xfade->priv->layout, _("100%"), strlen (_("100%")));
	pango_layout_get_extents (xfade->priv->layout, NULL, &log_rect);
	xfade->priv->xofs = PANGO_PIXELS (log_rect.width) + 4;

	xfade->priv->iw = xfade->priv->length / xfade->priv->frames_per_pixel + 1;

	req->width = xfade->priv->iw + xfade->priv->xofs;

	pango_layout_set_text (xfade->priv->layout, _("0%"), strlen (_("0%")));
	pango_layout_get_size (xfade->priv->layout, NULL, &h);
	req->height = xfade->priv->yofs + XFADE_HEIGHT + (PANGO_PIXELS (h) / 2);
}

static void
realize (GtkWidget *widget)
{
	MarlinCrossFader *xfade = MARLIN_CROSS_FADER (widget);
	
	/* Add events */
	gtk_widget_add_events (widget,
			       GDK_BUTTON_PRESS_MASK |
			       GDK_BUTTON_RELEASE_MASK |
			       GDK_POINTER_MOTION_MASK);

	GTK_WIDGET_CLASS (parent_class)->realize (widget);

	/* Now we have our window...
	   create our GCs */
	xfade->priv->src_gc = gdk_gc_new (widget->window);
	gdk_gc_set_foreground (xfade->priv->src_gc, &xfade->priv->src_colour);

	xfade->priv->dest_gc = gdk_gc_new (widget->window);
	gdk_gc_set_foreground (xfade->priv->dest_gc, &xfade->priv->dest_colour);

	gdk_window_set_back_pixmap (widget->window, NULL, FALSE);
	make_pixmap (widget);
}

static void
unrealize (GtkWidget *widget)
{
	MarlinCrossFader *xfade = MARLIN_CROSS_FADER (widget);

	g_object_unref (G_OBJECT (xfade->priv->src_gc));
	g_object_unref (G_OBJECT (xfade->priv->dest_gc));

	if (xfade->priv->background) {
		g_object_unref (xfade->priv->background);
	}
	
	if (xfade->priv->non_gr_exp_gc) {
		g_object_unref (xfade->priv->non_gr_exp_gc);
	}
		
	xfade->priv->background = NULL;
	xfade->priv->non_gr_exp_gc = NULL;
}

#define HALF_BOX_WIDTH 6
#define BOX_WIDTH 12
static void
draw_box (GtkWidget *widget,
	  GdkGC *gc,
	  GtkStateType state_type,
	  int centre_x,
	  int centre_y)
{
	int x1, y1;

	x1 = centre_x - HALF_BOX_WIDTH;
	y1 = centre_y - HALF_BOX_WIDTH;

	if (state_type != GTK_STATE_PRELIGHT &&
	    state_type != GTK_STATE_ACTIVE) {
		/* Draw white centre */
		gdk_gc_set_foreground (gc, &widget->style->white);
		gdk_draw_rectangle (widget->window, gc, TRUE, x1, y1, BOX_WIDTH, BOX_WIDTH);
	}
	
	/* Draw black outline */
	gdk_gc_set_foreground (gc, &widget->style->black);
	gdk_draw_rectangle (widget->window, gc, FALSE, x1, y1, BOX_WIDTH - 1, BOX_WIDTH - 1);
}

static void
draw_fader_line (MarlinCrossFader *xfade,
		 MarlinSampleFade *f,
		 GdkGC *gc,
		 const char *name)
{
	MarlinCrossFaderPrivate *priv = xfade->priv;
	GtkWidget *widget = (GtkWidget *) xfade;
	int y1, y2, x1, x2, ty, tx;
	int height, width, t_width, t_height;
	GdkPoint points[4];
	GdkRectangle clip_rect;
	GdkGC *bw_gc;
	
	width = widget->allocation.width;
	height = XFADE_HEIGHT;

	y1 = height - (f->in_level * (float) height);
	y2 = height - (f->out_level * (float) height);
	
	x1 = (guint) (f->fade_start - priv->start_frame) / priv->frames_per_pixel;
	x2 = (guint) (f->fade_end - priv->start_frame) / priv->frames_per_pixel;
	points[0].x = priv->xofs;
	points[0].y = priv->yofs + y1;
	points[1].x = priv->xofs + x1;
	points[1].y = priv->yofs + y1;
	points[2].x = priv->xofs + x2;
	points[2].y = priv->yofs + y2;
	points[3].x = priv->xofs + priv->iw - 1;
	points[3].y = priv->yofs + y2;

	gdk_draw_lines (widget->window, gc, points, 4);

	/* Draw the grab handles */
	clip_rect.x = priv->xofs;
	clip_rect.y = priv->yofs;
	clip_rect.width = priv->iw;
	clip_rect.height = height;

	/* FIXME: Cache this GC? */
	bw_gc = gdk_gc_new (widget->window);

	gdk_gc_set_clip_rectangle (bw_gc, &clip_rect);
	draw_box (widget, bw_gc, GTK_STATE_NORMAL, 
		  priv->xofs + x1, priv->yofs + y1);
	draw_box (widget, bw_gc, GTK_STATE_NORMAL, 
		  priv->xofs + x2, priv->yofs + y2);

	/* Draw the key */
	pango_layout_set_text (priv->layout, name, strlen (name));
	
	pango_layout_get_size (priv->layout, &t_width, &t_height);
	if (y1 - PANGO_PIXELS (t_height) < 0) {
		ty = y1 + 2;
	} else {
		ty = y1 - 2 - PANGO_PIXELS (t_height);
	}

	if (x1 + PANGO_PIXELS (t_width) > priv->iw + priv->xofs) {
		tx = x1 - PANGO_PIXELS (t_width) - 2;
	} else {
		tx = x1;
	}

	gdk_draw_layout (widget->window, gc,
			 priv->xofs + tx, priv->yofs + ty,
			 priv->layout);

	g_object_unref (G_OBJECT (bw_gc));
}

#define TEXT_X_OFFSET 2
#define TEXT_Y_OFFSET 2
static void
_marlin_cross_fader_paint (MarlinCrossFader *xfade,
			   GdkRectangle *area,
			   GtkStateType state_type)
{
	MarlinCrossFaderPrivate *priv;
	GtkWidget *widget;
	int width, height;
	GdkRectangle fa, inter;
	
	priv = xfade->priv;
	widget = GTK_WIDGET (xfade);
	
	width = widget->allocation.width;
	height = widget->allocation.height - priv->yofs;

	fa.x = priv->xofs;
	fa.y = priv->yofs;
	fa.width = priv->iw;
	fa.height = height;

	if (gdk_rectangle_intersect (area, &fa, &inter)) {
		PangoRectangle rect;
		PangoRenderer *renderer;
		PangoMatrix matrix = PANGO_MATRIX_INIT;
		PangoContext *context;
		int t_height, t_width;
		char *text;

		/* Draw the text */

		/* Rotate the Level label */
		context = gtk_widget_get_pango_context (widget);
		pango_matrix_rotate (&matrix, 90.0);
		pango_context_set_matrix (context, &matrix);
		pango_layout_context_changed (priv->layout);

		pango_layout_set_text (priv->layout, _("Level"), strlen (_("Level")));
		pango_layout_get_size (priv->layout, &t_width, &t_height);
		gdk_draw_layout (widget->window,
				 widget->style->text_gc[state_type],
				 priv->xofs - t_height / PANGO_SCALE, 
				 priv->yofs + (height - (t_width / PANGO_SCALE)) / 2,
				 priv->layout);

		/* Reset the matrix */
		pango_context_set_matrix (context, NULL);

		if (priv->scale == MARLIN_SCALE_LINEAR) {
			pango_layout_set_text (priv->layout, _("100%"), -1);
		} else {
			pango_layout_set_text (priv->layout, _("0 dB"), -1);
		}
		pango_layout_get_size (priv->layout, NULL, &t_height);
		gdk_draw_layout (widget->window,
				 widget->style->text_gc[state_type],
				 TEXT_X_OFFSET, priv->yofs - ((t_height / PANGO_SCALE) / 2),
				 priv->layout);

		if (priv->scale == MARLIN_SCALE_LINEAR) {
			pango_layout_set_text (priv->layout, _("0%"), -1);
		} else {
			pango_layout_set_text (priv->layout, 
					       _("-" MARLIN_INFINITY_DB_TEXT), -1);
		}
		pango_layout_get_extents (priv->layout, NULL, &rect);
		gdk_draw_layout (widget->window,
				 widget->style->text_gc[state_type],
				 priv->xofs - TEXT_X_OFFSET - PANGO_PIXELS (rect.width),
				 (priv->yofs + XFADE_HEIGHT) - ((PANGO_PIXELS (rect.height) / 2) + TEXT_Y_OFFSET),
				 priv->layout);

		/* Draw the frame labels */
		text = g_strdup_printf ("%llu", priv->start_frame);
		pango_layout_set_text (priv->layout, text, strlen (text));
		g_free (text);

		pango_layout_get_size (priv->layout, NULL, &t_height);
		gdk_draw_layout (widget->window,
				 widget->style->text_gc[state_type],
				 priv->xofs + TEXT_X_OFFSET,
				 priv->yofs - (t_height / PANGO_SCALE),
				 priv->layout);

		pango_layout_set_text (priv->layout, _("Frames"),
				       strlen (_("Frames")));
		pango_layout_get_size (priv->layout, &t_width, &t_height);

		gdk_draw_layout (widget->window,
				 widget->style->text_gc[state_type],
				 priv->xofs + (priv->iw - PANGO_PIXELS (t_width)) / 2,
				 priv->yofs - PANGO_PIXELS (t_height),
				 priv->layout);

		text = g_strdup_printf ("%llu", priv->start_frame + priv->length);
		pango_layout_set_text (priv->layout, text, strlen (text));
		g_free (text);

		pango_layout_get_size (priv->layout, &t_width, &t_height);
		gdk_draw_layout (widget->window,
				 widget->style->text_gc[state_type],
				 priv->xofs + (priv->iw - PANGO_PIXELS (t_width)),
				 priv->yofs - PANGO_PIXELS (t_height),
				 priv->layout);
		
		gdk_gc_set_clip_rectangle (widget->style->black_gc, &inter);
		gdk_gc_set_clip_rectangle (priv->dest_gc, &inter);
		gdk_gc_set_clip_rectangle (priv->src_gc, &inter);

		/* Draw the background */
		gdk_draw_drawable (widget->window,
				   priv->non_gr_exp_gc,
				   priv->background,
				   inter.x - priv->xofs, inter.y - priv->yofs,
				   inter.x, inter.y,
				   inter.width, inter.height);

		/* Draw the destination line */
		g_object_get (G_OBJECT (priv->dest),
			      "name", &text,
			      NULL);
		draw_fader_line (xfade, priv->dest_fader, priv->dest_gc, text);
		g_free (text);
		
		/* And draw one for the source too */
		g_object_get (G_OBJECT (priv->src),
			      "name", &text,
			      NULL);
		draw_fader_line (xfade, priv->src_fader, priv->src_gc, text);
		g_free (text);

		gdk_gc_set_clip_rectangle (widget->style->black_gc, NULL);
		gdk_gc_set_clip_rectangle (priv->dest_gc, NULL);
		gdk_gc_set_clip_rectangle (priv->src_gc, NULL);
	}
}

static gboolean
expose_event (GtkWidget *widget,
	      GdkEventExpose *event)
{
	if (GTK_WIDGET_DRAWABLE (widget)) {
		MarlinCrossFader *xfade = MARLIN_CROSS_FADER (widget);

		_marlin_cross_fader_paint (xfade, &event->area,
					   GTK_WIDGET_STATE (widget));
	}

	return FALSE;
}

static gboolean
in_box (int box_x,
	int box_y,
	int x,
	int y)
{
	if (x >= box_x - HALF_BOX_WIDTH &&
	    x <= box_x + HALF_BOX_WIDTH &&
	    y >= box_y - HALF_BOX_WIDTH &&
	    y <= box_y + HALF_BOX_WIDTH) {
		return TRUE;
	}

	return FALSE;
}

static int
in_fader_box (MarlinCrossFader *fader,
	      int x,
	      int y)
{
	MarlinCrossFaderPrivate *priv = fader->priv;
	MarlinSampleFade *f;
	int height;

	height = XFADE_HEIGHT;

	f = priv->src_fader;
	if (in_box ((f->fade_start - priv->start_frame) / priv->frames_per_pixel,
		    height - (f->in_level * (float) height),
		    x, y)) {
		return SRC_IN;
	}

	if (in_box ((f->fade_end - priv->start_frame) / priv->frames_per_pixel,
		    height - (f->out_level * (float) height),
		    x, y)) {
		return SRC_OUT;
	}

	f = priv->dest_fader;
	if (in_box ((f->fade_start - priv->start_frame) / priv->frames_per_pixel,
		    height - (f->in_level * (float) height),
		    x, y)) {
		return DEST_IN;
	}

	if (in_box ((f->fade_end - priv->start_frame) / priv->frames_per_pixel,
		    height - (f->out_level * (float) height),
		    x, y)) {
		return DEST_OUT;
	}


	return NONE;
}

static gboolean
tooltip_paint_window (GtkWidget *widget,
		      GdkEventExpose *event,
		      MarlinCrossFader *xfade)
{
	gtk_paint_flat_box (widget->style, widget->window,
			    GTK_STATE_NORMAL, GTK_SHADOW_OUT,
			    NULL, widget, "tooltip",
			    0, 0,
			    widget->allocation.width,
			    widget->allocation.height);
	return FALSE;
}

static char *
position_to_string (MarlinCrossFader *xfade,
		    enum _box box)
{
	char *str;
	MarlinSampleFade *f;
	double level;
	
	switch (box) {
	case NONE:
		return NULL;

	case SRC_IN:
	case SRC_OUT:
		f = xfade->priv->src_fader;
		break;

	case DEST_IN:
	case DEST_OUT:
		f = xfade->priv->dest_fader;
		break;

	default:
		g_assert_not_reached ();
		break;
	}
	
	switch (box) {
	case SRC_IN:
	case DEST_IN:
		level = f->in_level * 100.0f;
		if (xfade->priv->scale == MARLIN_SCALE_LOG) {
			level = marlin_percent_to_db (level);
			if (level <= MARLIN_INFINITE_DB) {
				str = g_strdup_printf (_("Position %llu\nLevel: -%s"),
						       f->fade_start, 
						       MARLIN_INFINITY_DB_TEXT);
			} else {
				str = g_strdup_printf (_("Position %llu\nLevel: %.1f dB"),
						       f->fade_start, level);
			}
		} else {
			
			str = g_strdup_printf (_("Position %llu\nLevel: %.2f%%"),
					       f->fade_start, level);
		}

		break;

	case SRC_OUT:
	case DEST_OUT:
		level = f->out_level * 100.0f;
		if (xfade->priv->scale == MARLIN_SCALE_LOG) {
			level = marlin_percent_to_db (level);
			if (level <= MARLIN_INFINITE_DB) {
				str = g_strdup_printf (_("Position %llu\nLevel: -%s"),
						       f->fade_end, 
						       MARLIN_INFINITY_DB_TEXT);
			} else {
				str = g_strdup_printf (_("Position %llu\nLevel: %.1f dB"),
						       f->fade_end, level);
			}
		} else {
			str = g_strdup_printf (_("Position %llu\nLevel: %.2f%%"),
					       f->fade_end, level);
		}

		break;

	default:
		g_assert_not_reached ();
	}

	return str;
}

#define TOOLTIP_OFFSET 4
static gboolean
button_press_event (GtkWidget *widget,
		    GdkEventButton *event)
{
	MarlinCrossFader *xfade = MARLIN_CROSS_FADER (widget);
	MarlinCrossFaderPrivate *priv = xfade->priv;
	int real_x, real_y;
	enum _box box;

	real_x = event->x - priv->xofs;
	real_y = event->y - priv->yofs;
	
	switch (event->button) {
	case 1:
		if ((box = in_fader_box (xfade, real_x, real_y)) != NONE) {
			GdkCursor *hand = marlin_cursor_get (widget, HAND_CLOSED);
			int ox, oy;
			char *tip;
			
			gdk_window_get_origin (widget->window, &ox, &oy);
			priv->in_drag = TRUE;
			priv->dragging_box = priv->current_box;
			
			gdk_pointer_grab (widget->window, FALSE,
					  GDK_BUTTON_RELEASE_MASK |
					  GDK_BUTTON1_MOTION_MASK, NULL,
					  hand, event->time);
			gdk_cursor_unref (hand);

			/* Show tooltip */
			priv->tooltip = gtk_window_new (GTK_WINDOW_POPUP);
 			gtk_widget_set_app_paintable (priv->tooltip, TRUE);
			gtk_window_set_resizable (GTK_WINDOW (priv->tooltip),
						  FALSE);
			gtk_widget_set_name (priv->tooltip, "marlin-xfader-tooltip");
			gtk_container_set_border_width (GTK_CONTAINER (priv->tooltip), 4);
			gtk_window_move (GTK_WINDOW (priv->tooltip), ox + event->x,
					 oy + real_y + TOOLTIP_OFFSET);

			g_signal_connect (priv->tooltip, "expose_event",
					  G_CALLBACK (tooltip_paint_window),
					  xfade);

			tip = position_to_string (xfade, box);
			priv->tip_label = gtk_label_new (tip);
			gtk_label_set_line_wrap (GTK_LABEL (priv->tip_label), TRUE);
			gtk_misc_set_alignment (GTK_MISC (priv->tip_label), 0.5, 0.5);
			if (priv->cur_tip != NULL) {
				g_free (priv->cur_tip);
			}
			
			priv->cur_tip = tip;
			/* Don't need to free tip here */
			
			gtk_widget_show (priv->tip_label);
			gtk_container_add (GTK_CONTAINER (priv->tooltip),
					   priv->tip_label);

			g_signal_connect (priv->tooltip,
					  "destroy",
					  G_CALLBACK (gtk_widget_destroyed),
					  &priv->tooltip);
			
			gtk_widget_show (priv->tooltip);
		}
		break;

	default:
		break;
	}

	return TRUE;
}

static gboolean
button_release_event (GtkWidget *widget,
		      GdkEventButton *event)
{
	MarlinCrossFader *xfade = MARLIN_CROSS_FADER (widget);
	MarlinCrossFaderPrivate *priv = xfade->priv;

	if (priv->in_drag) {
		gdk_pointer_ungrab (event->time);
		priv->in_drag = FALSE;
		priv->dragging_box = NONE;
		g_free (priv->cur_tip);
		priv->cur_tip = NULL;
		gtk_widget_destroy (priv->tooltip);
	}
}

static gboolean
motion_notify_event (GtkWidget *widget,
		     GdkEventMotion *event)
{
	MarlinCrossFader *xfade = MARLIN_CROSS_FADER (widget);
	MarlinCrossFaderPrivate *priv = xfade->priv;
	int real_x, real_y;
	enum _box new_box;
	
	real_x = event->x - priv->xofs;
	real_y = event->y - priv->yofs;
	new_box = in_fader_box (xfade, real_x, real_y);

	if (real_x < 0) {
		real_x = 0;
	} else if (real_x > priv->iw) {
		real_x = priv->iw;
	}

	if (real_y < 0) {
		real_y = 0;
	} else if (real_y > widget->allocation.height) {
		real_y = widget->allocation.height;
	}
	
	if (new_box != NONE) {
		GdkCursor *hand = marlin_cursor_get (widget, HAND_OPEN);

		gdk_window_set_cursor (widget->window, hand);
		gdk_cursor_unref (hand);
	} else {
		gdk_window_set_cursor (widget->window, NULL);
	}

	priv->current_box = new_box;

	if (priv->in_drag) {
		GdkRectangle rect;
		
		move_box_to (xfade, priv->dragging_box, real_x, real_y);

		/* Move the tooltip too */
		if (priv->tooltip != NULL) {
			int ox, oy;
			char *tip;

			tip = position_to_string (xfade, priv->dragging_box);
			if (strcmp (priv->cur_tip, tip) != 0) {
				gtk_label_set_text (GTK_LABEL (priv->tip_label), tip);
				g_free (priv->cur_tip);
				priv->cur_tip = tip;
				/* Don't need to free tip here */
			} else {
				/* Only free tip if it isn't used */
				g_free (tip);
			}
			
			gdk_window_get_origin (widget->window, &ox, &oy);
			gtk_window_move (GTK_WINDOW (priv->tooltip),
					 ox + real_x + priv->xofs,
					 oy + real_y + TOOLTIP_OFFSET);
		}
		
		rect.x = priv->xofs;
		rect.y = 0;
		rect.width = priv->iw;
		rect.height = widget->allocation.height;

		gdk_window_invalidate_rect (widget->window, &rect, FALSE);
	}
}

static void
class_init (MarlinCrossFaderClass *klass)
{
	GObjectClass *object_class;
	GtkWidgetClass *widget_class;

	object_class = G_OBJECT_CLASS (klass);
	widget_class = GTK_WIDGET_CLASS (klass);

	object_class->finalize = finalize;
	object_class->set_property = set_property;
	object_class->get_property = get_property;

	widget_class->size_allocate = size_allocate;
	widget_class->size_request = size_request;
	widget_class->realize = realize;
	widget_class->unrealize = unrealize;
	widget_class->expose_event = expose_event;
	widget_class->button_press_event = button_press_event;
	widget_class->button_release_event = button_release_event;
	widget_class->motion_notify_event = motion_notify_event;
	
	parent_class = g_type_class_peek_parent (klass);

	g_object_class_install_property (object_class,
					 PROP_SRC,
					 g_param_spec_object ("src",
							      "", "",
							      MARLIN_SAMPLE_TYPE,
							      G_PARAM_READWRITE));
	g_object_class_install_property (object_class,
					 PROP_DEST,
					 g_param_spec_object ("dest",
							      "", "",
							      MARLIN_SAMPLE_TYPE,
							      G_PARAM_READWRITE));
	g_object_class_install_property (object_class,
					 PROP_START_FRAME,
					 g_param_spec_uint64 ("start_frame",
							      "", "",
							      0, G_MAXUINT64,
							      0,
							      G_PARAM_READWRITE));
	g_object_class_install_property (object_class,
					 PROP_LENGTH,
					 g_param_spec_uint64 ("length",
							      "", "",
							      0, G_MAXUINT64,
							      0,
							      G_PARAM_READWRITE));
	g_object_class_install_property (object_class,
					 PROP_SRC_FADE,
					 g_param_spec_pointer ("src_fade",
							       "", "",
							       G_PARAM_READWRITE));
	g_object_class_install_property (object_class,
					 PROP_DEST_FADE,
					 g_param_spec_pointer ("dest_fade",
							       "", "",
							       G_PARAM_READWRITE));
	g_object_class_install_property (object_class,
					 PROP_SNAP_TO_GRID,
					 g_param_spec_boolean ("snap_to_grid",
							       "", "",
							       FALSE,
							       G_PARAM_READWRITE));
	g_object_class_install_property (object_class,
					 PROP_SCALE,
					 g_param_spec_enum ("scale",
							    "", "",
							    MARLIN_TYPE_SCALE,
							    MARLIN_SCALE_LOG,
							    G_PARAM_READWRITE));
}

static void
init (MarlinCrossFader *xfade)
{
	GdkColormap *cmap;
	
	xfade->priv = g_new0 (MarlinCrossFaderPrivate, 1);

	xfade->priv->start_frame = 0;
	xfade->priv->length = 0;
	xfade->priv->snap_to_grid = FALSE;
	
	/* Init the faders to basic x-fade */
	xfade->priv->src_fader = g_new (MarlinSampleFade, 1);

	xfade->priv->src_fader->in_level = 0.0;
	xfade->priv->src_fader->out_level = 1.0;
	xfade->priv->src_fader->fade_start = 0;
	xfade->priv->src_fader->fade_end = 0; /* Need to set this when we
						 know how long the sample is */

	xfade->priv->dest_fader = g_new (MarlinSampleFade, 1);

	xfade->priv->dest_fader->in_level = 1.0;
	xfade->priv->dest_fader->out_level = 0.0;
	xfade->priv->dest_fader->fade_start = 0;
	xfade->priv->dest_fader->fade_end = 0; /* Need to set this when we
						  know how long the sample is */
	
	xfade->priv->src_colour.red = 65535;
	xfade->priv->dest_colour.blue = 65535;

	cmap = gdk_colormap_get_system ();
	gdk_colormap_alloc_color (cmap, &xfade->priv->src_colour, FALSE, TRUE);
	gdk_colormap_alloc_color (cmap, &xfade->priv->dest_colour, FALSE, TRUE);

	xfade->priv->layout = gtk_widget_create_pango_layout (GTK_WIDGET (xfade),
							      _("Level"));
	xfade->priv->scale = MARLIN_SCALE_LOG;
}

GType
marlin_cross_fader_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		GTypeInfo info = {
			sizeof (MarlinCrossFaderClass), NULL, NULL,
			(GClassInitFunc) class_init, NULL, NULL,
			sizeof (MarlinCrossFader), 0, (GInstanceInitFunc) init
		};

		type = g_type_register_static (GTK_TYPE_DRAWING_AREA,
					       "MarlinCrossFader",
					       &info, 0);
	}

	return type;
}
