/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 *  Authors: Iain Holmes <iain@prettypeople.org>
 *
 *  Copyright 2003 Iain Holmes
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of version 2 of the GNU General Public 
 *  License as published by the Free Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *
 */

#include <glib/gi18n.h>

#include <gtk/gtkhbox.h>
#include <gtk/gtkwindow.h>
#include <gtk/gtkstock.h>
#include <gtk/gtktreeselection.h>
#include <gtk/gtktreeview.h>
#include <gtk/gtkvbox.h>
#include <gtk/gtklabel.h>

#include <gst/media-info/media-info.h>
#include <gst/gconf/gconf.h>

#include <marlin/marlin-stock.h>
#include <marlin/marlin-file-chooser.h>
#include <marlin/marlin-x-utils.h>
#include <marlin/marlin-utils.h>
#include <marlin/marlin-a11y-helper.h>

struct _ChooserOpenData {
	GtkWidget *chooser;
	
	GtkWidget *options_frame;
	GtkWidget *info_frame;

	/* Info widgets */
	GtkWidget *length;
	GtkWidget *sample_rate;
	GtkWidget *name;
	GtkWidget *channels;
	GtkWidget *mimetype;
	
	/* Transport controls */
	GtkWidget *play;
	GtkWidget *stop;

	/* For the metadata */
	GstMediaInfo *info;
	GstMediaInfoStream *stream;
	guint32 info_id;
	
	/* Filesel hooks */
	GtkTreeSelection *selection;
	guint32 change_id;
	
	/* Preview */
	GstElement *player;
	GstElement *src;
	GstElement *decoder;
	GstElement *sink;
	guint32 play_id;
};

static void
free_open_data (GtkFileChooser *fc,
		struct _ChooserOpenData *od)
{
	if (od->info_id > 0) {
		g_source_remove (od->info_id);
	}

	if (od->info != NULL) {
		g_object_unref (G_OBJECT (od->info));
	}

	if (od->stream != NULL) {
		g_free (od->stream);
	}

	if (od->play_id > 0) {
		gst_element_set_state (od->player, GST_STATE_READY);
		g_source_remove (od->play_id);
		g_object_unref (G_OBJECT (od->player));
	}
	
	g_free (od);
}

static void
clear_info (struct _ChooserOpenData *od)
{
	gtk_label_set_text (GTK_LABEL (od->length), _("Unknown"));
	gtk_label_set_text (GTK_LABEL (od->sample_rate), _("Unknown"));
	gtk_label_set_text (GTK_LABEL (od->channels), _("Unknown"));
	gtk_label_set_text (GTK_LABEL (od->name), _("Unknown"));
	gtk_label_set_text (GTK_LABEL (od->mimetype), _("Unknown"));
}

static void
set_info (struct _ChooserOpenData *od,
	  GstMediaInfoStream *stream)
{
	char *str_val;
	int int_val;
	GstMediaInfoTrack *track;
	
	str_val = marlin_ms_to_pretty_time ((stream->length_time / GST_SECOND) * 1000);
	gtk_label_set_text (GTK_LABEL (od->length), str_val);
	g_free (str_val);

	if (stream->tracks == NULL) {
		/* Erk? */
		return;
	}

	track = stream->tracks->data;
	
	int_val = get_int_from_caps (track->format, "channels");
	str_val = g_strdup_printf ("%d (%s)", int_val, int_val == 1 ? _("mono") : _("stereo"));
	gtk_label_set_text (GTK_LABEL (od->channels), str_val);
	g_free (str_val);

	int_val = get_int_from_caps (track->format, "rate");
	str_val = g_strdup_printf ("%d hz", int_val);
	gtk_label_set_text (GTK_LABEL (od->sample_rate), str_val);
	g_free (str_val);

	str_val = get_string_from_caps (track->metadata, "title");
	gtk_label_set_text (GTK_LABEL (od->name), str_val);
	g_free (str_val);

	gtk_label_set_text (GTK_LABEL (od->mimetype), stream->mime);
}
	
static gboolean
get_media_info (gpointer data)
{
	struct _ChooserOpenData *od = data;
	GError *error = NULL;

	if (gst_media_info_read_idler (od->info, &od->stream, &error) &&
	    od->stream == NULL) {
		/* Keep idling */
		return TRUE;
	}

	if (od->stream) {
		set_info (od, od->stream);
	} else {
		clear_info (od);
	}

	return FALSE;
}

static void
get_metadata (struct _ChooserOpenData *od,
	      const char *filename)
{
	GtkFileChooser *chooser = GTK_FILE_CHOOSER (od->chooser);
	GError *error = NULL;
	
	if (od->info_id > 0) {
		g_source_remove (od->info_id);
	}

	if (filename == NULL || g_file_test (filename, G_FILE_TEST_IS_DIR)) {
		gtk_file_chooser_set_preview_widget_active (chooser, FALSE);
		gtk_widget_set_sensitive (od->play, FALSE);
		clear_info (od);
		return;
	}

	gtk_widget_set_sensitive (od->play, TRUE);

	if (od->play_id > 0) {
		gst_element_set_state (od->player, GST_STATE_READY);
		g_source_remove (od->play_id);
		od->play_id = 0;

		gtk_widget_show (od->play);
		gtk_widget_hide (od->stop);
	}
	
	if (od->info != NULL) {
		g_object_unref (G_OBJECT (od->info));
	}

	if (od->stream != NULL) {
		g_free (od->stream);
		od->stream = NULL;
	}
	
	od->info = g_object_new (GST_MEDIA_INFO_TYPE, NULL);
	gst_media_info_read_with_idler (od->info, filename, GST_MEDIA_INFO_ALL, &error);
	
	od->info_id = g_idle_add (get_media_info, od);
	gtk_file_chooser_set_preview_widget_active (chooser, TRUE);
}

static void
selection_changed (GtkFileChooser *chooser,
		   struct _ChooserOpenData *od)
{
	const char *filename = gtk_file_chooser_get_preview_filename (chooser);

	get_metadata (od, filename);
}

static gboolean
play_iterate (gpointer data)
{
	struct _ChooserOpenData *od = data;

	return gst_bin_iterate (GST_BIN (od->player));
}

static void
eos_reached (GstElement *element,
	     struct _ChooserOpenData *od)
{
	gtk_widget_show (od->play);
	gtk_widget_hide (od->stop);
}

static void
start_playing (GtkWidget *button,
	       struct _ChooserOpenData *od)
{
	const char *filename;
	char *mimetype;

	/* Tear it down */
	if (od->player != NULL) {
		g_object_unref (G_OBJECT (od->player));
		od->player = NULL;
	}
	
	/* Build it up */
 	filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (od->chooser));
	mimetype = gnome_vfs_get_mime_type (filename);
	
	if (mimetype == NULL) {
		g_warning ("Mimetype is NULL");
		return;
	}
	
	od->decoder = get_decoder_for_mime (mimetype, "decoder");
	g_free (mimetype);
	
	if (od->decoder == NULL) {
		g_warning ("Decoder is NULL");
		return;
	}

	od->player = gst_pipeline_new ("previewer");

	od->src = gst_element_factory_make ("filesrc", "src");
	od->sink = gst_gconf_get_default_audio_sink ();
	g_signal_connect (od->sink, "eos",
			  G_CALLBACK (eos_reached), od);

	gst_bin_add_many (GST_BIN (od->player), od->src,
			  od->sink, od->decoder, NULL);

	gst_element_link (od->src, od->decoder);
	gst_element_link (od->decoder, od->sink);
	
	g_object_set (G_OBJECT (od->src),
		      "location", filename,
		      NULL);
	
	gst_element_set_state (od->player, GST_STATE_PLAYING);
	od->play_id = g_idle_add (play_iterate, od);

	gtk_widget_hide (od->play);
	gtk_widget_show (od->stop);
}

static void
stop_playing (GtkWidget *button,
	      struct _ChooserOpenData *od)
{
	gst_element_set_state (od->player, GST_STATE_READY);
	g_source_remove (od->play_id);
	od->play_id = 0;
	
	gtk_widget_hide (od->stop);
	gtk_widget_show (od->play);
}

static void
build_info_contents (struct _ChooserOpenData *od)
{
	GtkWidget *table, *label, *vbox, *spacer, *hbox;

	vbox = gtk_vbox_new (FALSE, 6);
	gtk_container_set_border_width (GTK_CONTAINER (vbox), 6);
	gtk_file_chooser_set_preview_widget (GTK_FILE_CHOOSER (od->chooser), vbox);

	label = marlin_make_title_label (_("File Information"));
	gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, FALSE, 0);

	hbox = gtk_hbox_new (FALSE, 6);
	gtk_box_pack_start (GTK_BOX (vbox), hbox, TRUE, TRUE, 0);

	spacer = gtk_label_new ("    ");
	gtk_box_pack_start (GTK_BOX (hbox), spacer, FALSE, FALSE, 0);
	
	table = marlin_make_table (5, 2, FALSE);
	gtk_box_pack_start (GTK_BOX (hbox), table, TRUE, TRUE, 0);

	label = marlin_make_title_label (_("Name:"));
	PACK (table, label, 0, 0, GTK_FILL);

	od->name = marlin_make_info_label (_("Unknown"));
	gtk_label_set_line_wrap (GTK_LABEL (od->name), TRUE);
	PACK (table, od->name, 1, 0, GTK_FILL);

	marlin_add_paired_relations (label, ATK_RELATION_LABEL_FOR,
				     od->name, ATK_RELATION_LABELLED_BY);
	
	label = marlin_make_title_label (_("Mime Type:"));
	PACK (table, label, 0, 1, GTK_FILL);

	od->mimetype = marlin_make_info_label (_("Unknown"));
	PACK (table, od->mimetype, 1, 1, GTK_FILL);

	marlin_add_paired_relations (label, ATK_RELATION_LABEL_FOR,
				     od->mimetype, ATK_RELATION_LABELLED_BY);
	
	label = marlin_make_title_label (_("Length:"));
	PACK (table, label, 0, 2, GTK_FILL);

	od->length = marlin_make_info_label (_("Unknown"));
	PACK (table, od->length, 1, 2, GTK_FILL);

	marlin_add_paired_relations (label, ATK_RELATION_LABEL_FOR,
				     od->length, ATK_RELATION_LABELLED_BY);
	
	label = marlin_make_title_label (_("Sample Rate:"));
	PACK (table, label, 0, 3, GTK_FILL);

	od->sample_rate = marlin_make_info_label (_("Unknown"));
	PACK (table, od->sample_rate, 1, 3, GTK_FILL);

	marlin_add_paired_relations (label, ATK_RELATION_LABEL_FOR,
				     od->sample_rate, ATK_RELATION_LABELLED_BY);
	
	label = marlin_make_title_label (_("Channels:"));
	PACK (table, label, 0, 4, GTK_FILL);

	od->channels = marlin_make_info_label (_("Unknown"));
	PACK (table, od->channels, 1, 4, GTK_FILL);

	marlin_add_paired_relations (label, ATK_RELATION_LABEL_FOR,
				     od->channels, ATK_RELATION_LABELLED_BY);
	
	hbox = gtk_hbox_new (FALSE, 6);
	gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);

	od->play = marlin_make_button (_("Play"), MARLIN_STOCK_PLAY);
	gtk_widget_set_sensitive (od->play, FALSE);
	g_signal_connect (G_OBJECT (od->play), "clicked",
			  G_CALLBACK (start_playing), od);
	gtk_box_pack_start (GTK_BOX (hbox), od->play, FALSE, FALSE, 0);

	od->stop = marlin_make_button (_("Stop"), MARLIN_STOCK_STOP);
	g_signal_connect (G_OBJECT (od->stop), "clicked",
			  G_CALLBACK (stop_playing), od);
	gtk_box_pack_start (GTK_BOX (hbox), od->stop, FALSE, FALSE, 0);
	
	gtk_widget_show_all (vbox);
	gtk_widget_hide (od->stop);
}

#if 0
static void
caps_print (GstCaps *caps)
{
	if (caps == NULL) return;

	if (TRUE) {
		GstProps *props = caps->properties;
		GList *walk;
		
		if (props == NULL) {
			g_print ("    none\n");
			return;
		}
		walk = props->properties;
		
		while (walk) {
			GstPropsEntry *entry = (GstPropsEntry *) walk->data;
			const gchar *name;
			gchar *str_val;
			gint int_val;
			GstPropsType type;
			
			name = gst_props_entry_get_name (entry);
			type = gst_props_entry_get_props_type (entry);
			switch (type) {
			case GST_PROPS_STRING_TYPE:
				gst_props_entry_get_string (entry, &str_val);
				g_print ("      %s='%s'\n", name, str_val);
				break;
			case GST_PROPS_INT_TYPE:
				gst_props_entry_get_int (entry, &int_val);
				g_print ("      %s=%d\n", name, int_val);
				break;
			default:
				break;
			}
			walk = g_list_next (walk);
		}
	}
	else {
		g_print (" unkown caps type\n");
	}
}

static void
info_print (GstMediaInfoStream *stream)
{
	int i;
	GList *p;
	GstMediaInfoTrack *track;
	
	g_print ("- mime type: %s\n", stream->mime);
	g_print ("- length: %.3f seconds\n",
		 (gdouble) stream->length_time / GST_SECOND);
	g_print ("- bitrate: %.3f kbps\n", stream->bitrate / 1000.0);
	g_print ("- number of tracks: %ld\n", stream->length_tracks);
	p = stream->tracks;
	if (p == NULL) {
		g_print ("- no track information, probably an error\n");
		return;
	}
	
	for (i = 0; i < stream->length_tracks; ++i) {
		g_print ("- track %d\n", i);
		track = (GstMediaInfoTrack *) p->data;
		g_print ("  - metadata:\n");
		caps_print (track->metadata);
		g_print ("  - streaminfo:\n");
		caps_print (track->streaminfo);
		g_print ("  - format:\n");
		caps_print (track->format);
		p = p->next;
	}
}
#endif

GtkWidget *
marlin_file_open_dialog_new (GtkWindow *parent_window)
{
	struct _ChooserOpenData *od;
	GtkWidget *widget;
	GList *mime_types;
	GtkFileFilter *filter;
	const char *filename;

	od = g_new0 (struct _ChooserOpenData, 1);
	
	widget = g_object_new (GTK_TYPE_FILE_CHOOSER_DIALOG,
			       "action", GTK_FILE_CHOOSER_ACTION_OPEN,
			       "title", _("Open File"),
			       "local-only", FALSE,
			       NULL);
	od->chooser = widget;

	build_info_contents (od);

	/* Check if we can display the first selected file */
	filename = gtk_file_chooser_get_preview_filename (GTK_FILE_CHOOSER (widget));
	get_metadata (od, filename);
	
	g_signal_connect (G_OBJECT (od->chooser), "destroy",
			  G_CALLBACK (free_open_data), od);
	g_signal_connect (G_OBJECT (od->chooser), "update-preview",
			  G_CALLBACK (selection_changed), od);
	
	gtk_dialog_add_buttons (GTK_DIALOG (widget),
				GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
				GTK_STOCK_OPEN, GTK_RESPONSE_OK,
				NULL);
	gtk_dialog_set_default_response (GTK_DIALOG (widget), GTK_RESPONSE_OK);

	filter = gtk_file_filter_new ();
	gtk_file_filter_set_name (filter, _("All Files"));
	gtk_file_filter_add_pattern (filter, "*");
	gtk_file_chooser_add_filter (GTK_FILE_CHOOSER (widget), filter);

	filter = gtk_file_filter_new ();
	gtk_file_filter_set_name (filter, _("Audio Files"));

	mime_types = get_mime_types ();
	for (; mime_types; mime_types = mime_types->next) {
		gtk_file_filter_add_mime_type (filter, mime_types->data);
	}

	gtk_file_chooser_add_filter (GTK_FILE_CHOOSER (widget), filter);
	gtk_file_chooser_set_filter (GTK_FILE_CHOOSER (widget), filter);
	return widget;
}

GtkWidget *
marlin_file_save_dialog_new (GtkWindow *parent_window)
{
	GtkWidget *widget;

	widget = g_object_new (GTK_TYPE_FILE_CHOOSER_DIALOG,
			       "action", GTK_FILE_CHOOSER_ACTION_SAVE,
			       "title", _("Save Sample As..."),
			       NULL);
	gtk_dialog_add_buttons (GTK_DIALOG (widget),
				GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
				GTK_STOCK_SAVE, GTK_RESPONSE_OK,
				NULL);
	gtk_dialog_set_default_response (GTK_DIALOG (widget), GTK_RESPONSE_OK);
	gtk_window_set_default_size (GTK_WINDOW (widget), 600, 400);

	return widget;
}
