/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 *  Authors: Iain Holmes <iain@prettypeople.org>
 *
 *  Copyright 2002 Iain Holmes
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of version 2 of the GNU General Public 
 *  License as published by the Free Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string.h>

#include <glib/gi18n.h>

#include <gst/gst.h>

#include <libgnomevfs/gnome-vfs-mime-utils.h>

#include <gconf/gconf-client.h>

#include <marlin-load-pipeline.h>
#include <marlin-marshal.h>
#include <marlin-sample.h>
#include <marlin-sample-element.h>
#include <marlin-mt.h>

enum {
	PROP_0,
	PROP_SAMPLE,
	PROP_FILENAME
};

struct _MarlinLoadPipelinePrivate {
	MarlinSample *sample;
	GstElement *src, *decoder, *oneton, *sink, *audioconvert;
	char *filename;

	guint notify_id;
};

static MarlinPipeline *parent_class = NULL;

static void
finalize (GObject *object)
{
	MarlinLoadPipeline *pipeline;

	pipeline = MARLIN_LOAD_PIPELINE (object);

	if (pipeline->priv == NULL) {
		return;
	}

	if (pipeline->priv->notify_id > 0) {
		g_signal_handler_disconnect (G_OBJECT (pipeline->priv->sample),
					     pipeline->priv->notify_id);
	}

	g_object_unref (G_OBJECT (pipeline->priv->sample));
	g_free (pipeline->priv->filename);

	g_free (pipeline->priv);
	
	pipeline->priv = NULL;

	G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
sample_notify (MarlinSample *sample,
	       GParamSpec *pspec,
	       MarlinLoadPipeline *pipeline)
{
	if (strcmp (g_param_spec_get_name (pspec), "total-frames") != 0) {
		return;
	}

	marlin_pipeline_progress_changed (MARLIN_PIPELINE (pipeline));
}

static void
set_property (GObject *object,
	      guint prop_id,
	      const GValue *value,
	      GParamSpec *pspec)
{
	MarlinLoadPipeline *pipeline;
	MarlinLoadPipelinePrivate *priv;
	char *filename, *mimetype;
	GstCaps *filtercaps;
	
	pipeline = MARLIN_LOAD_PIPELINE (object);
	priv = pipeline->priv;

	switch (prop_id) {
	case PROP_SAMPLE:
		if (priv->sample != NULL) {
			g_signal_handler_disconnect (G_OBJECT (priv->sample),
						     priv->notify_id);
			g_object_unref (G_OBJECT (priv->sample));
		}

		priv->sample = g_value_get_object (value);
		priv->notify_id = g_signal_connect (G_OBJECT (priv->sample),
						    "notify",
						    G_CALLBACK (sample_notify),
						    pipeline);
		g_object_ref (G_OBJECT (priv->sample));

		g_object_set (G_OBJECT (priv->sink),
			      "sample", priv->sample,
			      NULL);

		g_object_set (G_OBJECT (priv->sample),
			      "filename", priv->filename,
			      NULL);
		break;

	case PROP_FILENAME:
		filename = (char *) g_value_get_string (value);
		
		if (priv->sample != NULL) {
			g_object_set (G_OBJECT (priv->sample),
				      "filename", filename,
				      NULL);
		}

		g_object_set (G_OBJECT (priv->src),
			      "location", filename,
			      NULL);
		if (priv->filename != NULL) {
			g_free (priv->filename);
		}

		priv->filename = g_strdup (filename);

		if (priv->decoder != NULL) {
			marlin_pipeline_remove (MARLIN_PIPELINE (pipeline),
						priv->decoder);
			
		}

#define USE_SPIDER 
#ifndef USE_SPIDER
		mimetype = gnome_vfs_get_mime_type (priv->filename);
		if (mimetype == NULL) {
			char *err;

			err = g_strdup_printf (_("%s is an unknown mimetype."),
					       priv->filename);
			marlin_pipeline_set_error (MARLIN_PIPELINE (pipeline),
						   err);
			g_free (err);
			return;
		}

		priv->decoder = get_decoder_for_mime (mimetype, "decoder");
		if (priv->decoder == NULL) {
			char *err;

			err = g_strdup_printf (_("Could not load %s because there is no decoder for %s"),
					       priv->filename, mimetype);
			marlin_pipeline_set_error (MARLIN_PIPELINE (pipeline),
						   err);
			g_free (mimetype);
			g_free (err);
			return;
		}
		g_free (mimetype);

		marlin_pipeline_add (MARLIN_PIPELINE (pipeline), priv->decoder);
		gst_element_link (priv->src, priv->decoder);
		gst_element_link (priv->decoder, priv->audioconvert);

		filtercaps = gst_caps_new_simple ("audio/x-raw-float", NULL);
		gst_element_link_filtered (priv->audioconvert,
					   priv->oneton, filtercaps);
		gst_caps_free (filtercaps);
		
#else
		{
			GstElement *typefind;
			
			priv->decoder = gst_element_factory_make ("spider", "decoder");

			typefind = gst_element_factory_make ("typefind", "typefind");
			marlin_pipeline_add (MARLIN_PIPELINE (pipeline), typefind);
			marlin_pipeline_add (MARLIN_PIPELINE (pipeline), priv->decoder);
			gst_element_link (priv->src, typefind);
			gst_element_link (typefind, priv->decoder);
			gst_element_link (priv->decoder, priv->audioconvert);
			
			filtercaps = gst_caps_new_simple ("audio/x-raw-float",
							  NULL);
			gst_element_link_filtered (priv->audioconvert,
						   priv->oneton, filtercaps);
			gst_caps_free (filtercaps);
		}
#endif

		break;
		
	default:
		break;
	}
}

static void
get_property (GObject *object,
	      guint prop_id,
	      GValue *value,
	      GParamSpec *pspec)
{
	MarlinLoadPipeline *pipeline;
	MarlinLoadPipelinePrivate *priv;

	pipeline = MARLIN_LOAD_PIPELINE (object);
	priv = pipeline->priv;

	switch (prop_id) {
	case PROP_SAMPLE:
		g_value_set_object (value, priv->sample);
		break;

	case PROP_FILENAME:
		g_value_set_string (value, priv->filename);
		break;
		
	default:
		break;
	}
}

/* Fixme: Should this take a GError? */
/* Fixme: This comment should be in marlin-pipeline.c */
static int
get_progress (MarlinPipeline *pipeline)
{
	MarlinLoadPipeline *load_pipeline;
	GstPad *src;
	guint64 position, total;
	GstFormat format = GST_FORMAT_BYTES;
	gboolean ret;
	
	load_pipeline = MARLIN_LOAD_PIPELINE (pipeline);

	src = gst_element_get_pad (load_pipeline->priv->src, "src");
	ret = gst_pad_query (src, GST_QUERY_POSITION,
			     &format, &position);
	if (ret == FALSE) {
		return 0;
	}

	format = GST_FORMAT_BYTES;
	ret = gst_pad_query (src, GST_QUERY_TOTAL,
			     &format, &total);
	if (ret == FALSE) {
		return 0;
	}

	return (int) ((float)(position * 100) / (float) (total));
}

static void
class_init (GObjectClass *klass)
{
	MarlinPipelineClass *pclass;

	pclass = MARLIN_PIPELINE_CLASS (klass);
	
	klass->finalize = finalize;
	klass->set_property = set_property;
	klass->get_property = get_property;

	pclass->get_progress = get_progress;
	
	parent_class = g_type_class_peek_parent (klass);

	/* Properties */
	g_object_class_install_property (klass,
					 PROP_SAMPLE,
					 g_param_spec_object ("sample",
							      "Sample",
							      "The sample to be loaded",
							      MARLIN_SAMPLE_TYPE,
							      G_PARAM_READWRITE));

	g_object_class_install_property (klass,
					 PROP_FILENAME,
					 g_param_spec_string ("filename",
							      "Filename",
							      "",
							      "Untitled",
							      G_PARAM_READWRITE));
}

static void
oneton_new_pad (GstElement *oneton,
		GstPad *src,
		MarlinLoadPipeline *pipeline)
{
	GstPad *sink, *qsrc, *qsink;
	GstElement *q;
	const char *name;
	char *qname;

	/* Request a new pad */
	sink = gst_element_get_request_pad (pipeline->priv->sink, "sink%d");
	g_assert (sink != NULL);

	name = gst_pad_get_name (src);
	qname = g_strdup_printf ("queue.%s", name);
	q = gst_element_factory_make ("queue", qname);
	g_free (qname);

	marlin_pipeline_add (MARLIN_PIPELINE (pipeline), q);
	
	qsrc = gst_element_get_pad (q, "src");
	qsink = gst_element_get_pad (q, "sink");
	
	if (!gst_pad_link (src, qsink)) {
		g_assert_not_reached ();
	}
	
	if (!gst_pad_link (qsrc, sink)) {
		g_warning ("Could not connect the two pads");
		/* FIXME: Should set an error here */
	}

	/* The pipeline is already at PLAYING, so we need to
	   sync the state to get the queue up to speed */
	gst_element_sync_state_with_parent (q);
}

static void
load_eos_reached (GstElement *element,
		  MarlinLoadPipeline *pipeline)
{
	MarlinSampleElementSink *sink = MARLIN_SAMPLE_ELEMENT_SINK (element);
	MarlinOperation *operation;
	GError *error = NULL;
	gboolean ret;

	g_object_get (G_OBJECT (pipeline),
		      "operation", &operation,
		      NULL);

#if 0	
	ret = marlin_sample_ensure_mmapped (pipeline->priv->sample, operation, &error);
	if (ret == FALSE) {
		marlin_pipeline_set_error (MARLIN_PIPELINE (pipeline), error, "");
		return;
	}

	ret = marlin_sample_generate_peaks (pipeline->priv->sample, operation, &error);
	if (ret == FALSE) {
		marlin_pipeline_set_error (MARLIN_PIPELINE (pipeline), error, "");
		return;
	}
#endif

	marlin_pipeline_set_eos (MARLIN_PIPELINE (pipeline));
}
		  
static void
init (MarlinLoadPipeline *pipeline)
{
	GstBin *sink_bin;
	
	pipeline->priv = g_new (MarlinLoadPipelinePrivate, 1);

	pipeline->priv->src = gst_element_factory_make ("gnomevfssrc", "src");
	pipeline->priv->decoder = NULL;

	pipeline->priv->audioconvert = gst_element_factory_make ("audioconvert", "audioconvert");
	pipeline->priv->oneton = gst_element_factory_make ("deinterleave", "oneton");
	g_signal_connect (G_OBJECT (pipeline->priv->oneton), "new-pad",
			  G_CALLBACK (oneton_new_pad), pipeline);

	sink_bin = gst_thread_new ("load-sink-thread");
	pipeline->priv->sink = GST_ELEMENT (marlin_sample_element_sink_new (NULL));
	g_signal_connect (G_OBJECT (pipeline->priv->sink), "eos",
			  G_CALLBACK (load_eos_reached), pipeline);
	gst_bin_add (sink_bin, GST_ELEMENT (pipeline->priv->sink));
	
	/* Add everything to the pipeline */
	marlin_pipeline_add (MARLIN_PIPELINE (pipeline), pipeline->priv->src);
	marlin_pipeline_add (MARLIN_PIPELINE (pipeline), pipeline->priv->audioconvert);
	marlin_pipeline_add (MARLIN_PIPELINE (pipeline), pipeline->priv->oneton);
  	marlin_pipeline_add (MARLIN_PIPELINE (pipeline), GST_ELEMENT (sink_bin));

	pipeline->priv->sample = NULL;
	pipeline->priv->filename = NULL;
	pipeline->priv->notify_id = 0;
}

GType
marlin_load_pipeline_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		static GTypeInfo info = {
			sizeof (MarlinLoadPipelineClass), NULL, NULL,
			(GClassInitFunc) class_init, NULL, NULL,
			sizeof (MarlinLoadPipeline), 0, (GInstanceInitFunc) init
		};

		type = g_type_register_static (MARLIN_PIPELINE_TYPE,
					       "MarlinLoadPipeline",
					       &info, 0);
	}
	
	return type;
}

MarlinLoadPipeline *
marlin_load_pipeline_new (MarlinOperation *operation)
{
	return g_object_new (MARLIN_LOAD_PIPELINE_TYPE,
			     "operation", operation,
			     NULL);
}

MarlinLoadPipeline *
marlin_load_pipeline_new_from_sample (MarlinOperation *operation,
				      MarlinSample *sample)
{
	return g_object_new (MARLIN_LOAD_PIPELINE_TYPE,
			     "operation", operation,
			     "sample", sample,
			     NULL);
}
