/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 *  Authors: Iain Holmes <iain@prettypeople.org>
 *
 *  Copyright 2002 Iain Holmes
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of version 2 of the GNU General Public 
 *  License as published by the Free Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *
 */

#include <gst/gst.h>

#include <marlin-operation.h>

#undef LOCK_DEBUG 
/* #define LOCK_DEBUG 1 */
#include <marlin-read-write-lock.h>

#include <marlin-mt.h>
#include <marlin-marshal.h>
#include <marlin-msgport.h>

enum {
	STARTED,
	FINISHED,
	PROGRESS,
	CANCELLED,
	PAUSED,
	ERROR,
	LAST_SIGNAL
};

enum {
	PROP_0,
	PROP_DESCRIPTION
};

static MarlinObjectClass *parent_class = NULL;
static guint signals[LAST_SIGNAL];

struct _MarlinOperationPrivate {
	MarlinReadWriteLock *lock;
	int progress;

	/* FIXME: This needs to be a stack */
	char *description;
	gboolean in_op;
};

static void
finalize (GObject *object)
{
	MarlinOperation *operation;
	MarlinOperationPrivate *priv;
	
	operation = MARLIN_OPERATION (object);

	priv = operation->priv;
	if (operation->priv == NULL) {
		return;
	}		

	WRITE_LOCK (priv->lock);
	operation->priv = NULL;

	g_free (operation->priv->description);
	
	/* Destroy everything else here */
	WRITE_UNLOCK (priv->lock);

	marlin_read_write_lock_destroy (priv->lock);
	g_free (priv);
	
	G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
set_property (GObject *object,
	      guint prop_id,
	      const GValue *value,
	      GParamSpec *pspec)
{
	MarlinOperation *operation;

	operation = MARLIN_OPERATION (object);
	
	switch (prop_id) {
	case PROP_DESCRIPTION:
		marlin_read_write_lock_lock (operation->priv->lock,
					     MARLIN_READ_WRITE_LOCK_MODE_WRITE);
		if (operation->priv->description != NULL) {
			g_free (operation->priv->description);
		}
		
		operation->priv->description = g_strdup (g_value_get_string (value));
		marlin_read_write_lock_unlock (operation->priv->lock,
					       MARLIN_READ_WRITE_LOCK_MODE_WRITE);
  		marlin_object_notify (object, "description-R");
		break;

	default:
		break;
	}
}

static void
get_property (GObject *object,
	      guint prop_id,
	      GValue *value,
	      GParamSpec *pspec)
{
	MarlinOperation *operation;

	operation = MARLIN_OPERATION (object);

	switch (prop_id) {
	case PROP_DESCRIPTION:
		marlin_read_write_lock_lock (operation->priv->lock,
					     MARLIN_READ_WRITE_LOCK_MODE_READ);
		g_value_set_string (value, operation->priv->description);
		marlin_read_write_lock_unlock (operation->priv->lock,
					       MARLIN_READ_WRITE_LOCK_MODE_READ);
		break;

	default:
		break;
	}
}

static void
class_init (MarlinOperationClass *klass)
{
	GObjectClass *object_class;

	object_class = G_OBJECT_CLASS (klass);

	object_class->finalize = finalize;
	object_class->set_property = set_property;
	object_class->get_property = get_property;

	parent_class = g_type_class_peek_parent (klass);

	g_object_class_install_property (object_class,
					 PROP_DESCRIPTION,
					 g_param_spec_string ("description",
							      "", "",
							      "",
							      G_PARAM_READWRITE));
	signals[STARTED] = g_signal_new ("started",
					 G_TYPE_FROM_CLASS (klass),
					 G_SIGNAL_RUN_FIRST |
					 G_SIGNAL_NO_RECURSE,
					 G_STRUCT_OFFSET (MarlinOperationClass, started),
					 NULL, NULL,
					 g_cclosure_marshal_VOID__VOID,
					 G_TYPE_NONE, 0);
	signals[FINISHED] = g_signal_new ("finished",
					  G_TYPE_FROM_CLASS (klass),
					  G_SIGNAL_RUN_FIRST |
					  G_SIGNAL_NO_RECURSE,
					  G_STRUCT_OFFSET (MarlinOperationClass, finished),
					  NULL, NULL,
					  g_cclosure_marshal_VOID__VOID,
					  G_TYPE_NONE, 0);
	signals[PROGRESS] = g_signal_new ("progress",
					  G_TYPE_FROM_CLASS (klass),
					  G_SIGNAL_RUN_FIRST |
					  G_SIGNAL_NO_RECURSE,
					  G_STRUCT_OFFSET (MarlinOperationClass, progress),
					  NULL, NULL,
					  g_cclosure_marshal_VOID__INT,
					  G_TYPE_NONE, 1, G_TYPE_INT);
	signals[CANCELLED] = g_signal_new ("cancelled",
					   G_TYPE_FROM_CLASS (klass),
					   G_SIGNAL_RUN_FIRST |
					   G_SIGNAL_NO_RECURSE,
					   G_STRUCT_OFFSET (MarlinOperationClass, cancelled),
					   NULL, NULL,
					   g_cclosure_marshal_VOID__VOID,
					   G_TYPE_NONE, 0);
	signals[ERROR] = g_signal_new ("error",
				       G_TYPE_FROM_CLASS (klass),
				       G_SIGNAL_RUN_FIRST |
				       G_SIGNAL_NO_RECURSE,
				       G_STRUCT_OFFSET (MarlinOperationClass,
							error),
				       NULL, NULL,
				       marlin_marshal_VOID__BOXED_STRING,
				       G_TYPE_NONE, 2,
				       GST_TYPE_G_ERROR,
				       G_TYPE_STRING);
	signals[PAUSED] = g_signal_new ("paused",
					G_TYPE_FROM_CLASS (klass),
					G_SIGNAL_RUN_FIRST |
					G_SIGNAL_NO_RECURSE,
					G_STRUCT_OFFSET (MarlinOperationClass,
							 paused),
					NULL, NULL,
					marlin_marshal_VOID__BOOLEAN,
					G_TYPE_NONE, 1, G_TYPE_BOOLEAN);

}

static void
init (MarlinOperation *operation)
{
	operation->priv = g_new0 (MarlinOperationPrivate, 1);
	operation->priv->lock = marlin_read_write_lock_new ();
	operation->priv->in_op = FALSE;
}

GType
marlin_operation_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		GTypeInfo info = {
			sizeof (MarlinOperationClass), NULL, NULL,
			(GClassInitFunc) class_init, NULL, NULL,
			sizeof (MarlinOperation), 0, (GInstanceInitFunc) init
		};

		type = g_type_register_static (MARLIN_OBJECT_TYPE,
					       "MarlinOperation", &info, 0);
	}

	return type;
}

MarlinOperation *
marlin_operation_new (void)
{
	return g_object_new (MARLIN_OPERATION_TYPE, NULL);
}

struct _op_msg {
	MarlinMTMsg msg;

	MarlinOperation *operation;
	gboolean finished;
};

/* MT stuff */
static void
do_op (MarlinMTMsg *mm)
{
	struct _op_msg *op = (struct _op_msg *) mm;

	if (op->finished == TRUE) {
		g_signal_emit (G_OBJECT (op->operation), signals[FINISHED], 0);
	} else {
		g_signal_emit (G_OBJECT (op->operation), signals[STARTED], 0);
	}
}

static void
free_op (MarlinMTMsg *mm)
{
	struct _op_msg *op = (struct _op_msg *) mm;
	
	g_object_unref (G_OBJECT (op->operation));
}

static MarlinMTMsgOp op_funcs = {
	NULL,
	do_op,
	NULL,
	free_op
};

void
marlin_operation_do (MarlinOperation *operation,
		     gboolean completed)
{
	struct _op_msg *m;
	
	m = marlin_mt_msg_new (&op_funcs, NULL, sizeof (*m));
	m->operation = operation;
	g_object_ref (operation);
	
	m->finished = completed;

	marlin_msgport_put (marlin_gui_port, (MarlinMsg *) m);
}

void
marlin_operation_start (MarlinOperation *operation)
{
	marlin_operation_do (operation, FALSE);
}

void
marlin_operation_finish (MarlinOperation *operation)
{
	marlin_operation_do (operation, TRUE);
}

struct _op_progress_msg {
	MarlinMTMsg msg;

	MarlinOperation *operation;

	int percent;
};

/* Progress */
static void
do_set_op_progress (MarlinMTMsg *mm)
{
	struct _op_progress_msg *op = (struct _op_progress_msg *) mm;
	int old;

 	WRITE_LOCK (op->operation->priv->lock);
	old = op->operation->priv->progress;
	op->operation->priv->progress = op->percent;
 	WRITE_UNLOCK (op->operation->priv->lock);

	/* So we don't emit signals constantly... */
	/* WARNING: If this is called too often then it will deadlock */
	/* FIXME FIXME FIXME */
	if (old != op->percent) {
  		g_signal_emit (G_OBJECT (op->operation), signals[PROGRESS], 0, op->operation->priv->progress);
	}

	WRITE_LOCK (op->operation->priv->lock);
	op->operation->priv->in_op = FALSE;
	WRITE_UNLOCK (op->operation->priv->lock);
}

static void
do_free_op_progress (MarlinMTMsg *mm)
{
	struct _op_progress_msg *op = (struct _op_progress_msg *) mm;

	g_object_unref (G_OBJECT (op->operation));
}

static MarlinMTMsgOp op_progress = {
	NULL,
	do_set_op_progress,
	NULL,
	do_free_op_progress,
};

void
marlin_operation_progress (MarlinOperation *operation,
			   int progress)
{
	struct _op_progress_msg *m;

	READ_LOCK (operation->priv->lock);
	if (operation->priv->in_op) {
		READ_UNLOCK (operation->priv->lock);
		return;
	}

	READ_UNLOCK (operation->priv->lock);
	
	m = marlin_mt_msg_new (&op_progress, NULL, sizeof (*m));
	m->operation = operation;
	g_object_ref (operation);
	m->percent = progress;

	WRITE_LOCK (operation->priv->lock);
	operation->priv->in_op = TRUE;
	WRITE_UNLOCK (operation->priv->lock);
	
	marlin_msgport_put (marlin_gui_port, (MarlinMsg *) m);
}

struct _op_cancel_msg {
	MarlinMTMsg msg;

	MarlinOperation *operation;
};

/* Cancellation */
static void
do_op_cancel (MarlinMTMsg *mm)
{
	struct _op_cancel_msg *op = (struct _op_cancel_msg *) mm;

	g_signal_emit (G_OBJECT (op->operation), signals[CANCELLED], 0);
}

static void
free_op_cancel (MarlinMTMsg *mm)
{
	struct _op_cancel_msg *op = (struct _op_cancel_msg *) mm;

	g_object_unref (G_OBJECT (op->operation));
}

static MarlinMTMsgOp op_cancel = {
	NULL,
	do_op_cancel,
	NULL,
	free_op_cancel,
};

void
marlin_operation_cancel (MarlinOperation *operation)
{
	struct _op_cancel_msg *m;

	m = marlin_mt_msg_new (&op_cancel, NULL, sizeof (*m));
	m->operation = operation;
	g_object_ref (operation);

	marlin_msgport_put (marlin_gui_port, (MarlinMsg *) m);
}

struct _error_msg {
	MarlinMTMsg msg;

	MarlinOperation *operation;
	GError *error;
	char *message;
};

static void
do_error (MarlinMTMsg *mm)
{
	struct _error_msg *em = (struct _error_msg *) mm;

	g_signal_emit (em->operation, signals[ERROR], 0, em->error, em->message);
}

static void
free_error (MarlinMTMsg *mm)
{
	struct _error_msg *em = (struct _error_msg *) mm;
	
	g_free (em->message);
	g_error_free (em->error);
	g_object_unref (G_OBJECT (em->operation));
}

static MarlinMTMsgOp error_op = {
	NULL,
	do_error,
	NULL,
	free_error
};

void
marlin_operation_set_error (MarlinOperation *operation,
			    GError *error,
			    const char *debug)
{
	struct _error_msg *m;

	m = marlin_mt_msg_new (&error_op, NULL, sizeof (*m));
	m->operation = operation;
	m->error = g_error_copy (error);
	m->message = g_strdup (debug);

	g_object_ref (G_OBJECT (m->operation));
	
	marlin_msgport_put (marlin_gui_port, (MarlinMsg *) m);
}

struct _pause_msg {
	MarlinMTMsg msg;

	MarlinOperation *operation;
	gboolean paused;
};

static void
do_pause (MarlinMTMsg *mm)
{
	struct _pause_msg *em = (struct _pause_msg *) mm;

	g_signal_emit (em->operation, signals[PAUSED], 0, em->paused);
}

static void
free_pause (MarlinMTMsg *mm)
{
	struct _pause_msg *em = (struct _pause_msg *) mm;
	
	g_object_unref (G_OBJECT (em->operation));
}

static MarlinMTMsgOp pause_op = {
	NULL,
	do_pause,
	NULL,
	free_pause
};

void
marlin_operation_pause (MarlinOperation *operation,
			gboolean paused)
{
	struct _pause_msg *m;

	m = marlin_mt_msg_new (&pause_op, NULL, sizeof (*m));
	m->operation = operation;
	m->paused = paused;

	g_object_ref (G_OBJECT (m->operation));

	marlin_msgport_put (marlin_gui_port, (MarlinMsg *) m);
}
