/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 *  Authors: Iain Holmes <iain@prettypeople.org>
 *
 *  Copyright 2002 Iain Holmes
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of version 2 of the GNU General Public 
 *  License as published by the Free Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <gst/gst.h>

#include <gconf/gconf-client.h>

#include <marlin-pipeline.h>
#include <marlin-marshal.h>
#include <marlin-operation.h>
#include <marlin-sample.h>
#include <marlin-sample-element.h>
#include <marlin-mt.h>

enum {
	EOS,
	SHUTDOWN,
	POSITION_CHANGED,
	LAST_SIGNAL
};

static guint32 signals[LAST_SIGNAL];

enum {
	PROP_0,
	PROP_OPERATION,
};

struct _MarlinPipelinePrivate {
	GstElement *pipeline;

	MarlinOperation *operation;
	guint shutdown_id;
	
	gboolean started;
};

static GObjectClass *parent_class = NULL;

static void
finalize (GObject *object)
{
	MarlinPipeline *pipeline;

	pipeline = MARLIN_PIPELINE (object);

	if (pipeline->priv == NULL) {
		return;
	}

	/* Disconnect our shutdown emitter so
	   that when we set the state to NULL
	   we don't trigger another shutdown */
	if (pipeline->priv->shutdown_id > 0) {
		g_signal_handler_disconnect (pipeline->priv->pipeline,
					     pipeline->priv->shutdown_id);
	}

/*   	gst_element_set_state (pipeline->priv->pipeline, GST_STATE_NULL); */
	g_object_unref (G_OBJECT (pipeline->priv->pipeline));

	if (pipeline->priv->operation) {
		g_object_unref (G_OBJECT (pipeline->priv->operation));
	}
	
	g_free (pipeline->priv);
	
	pipeline->priv = NULL;

	parent_class->finalize (object);
}

static void
set_property (GObject *object,
	      guint prop_id,
	      const GValue *value,
	      GParamSpec *pspec)
{
	MarlinPipeline *pipeline = MARLIN_PIPELINE (object);
	
	switch (prop_id) {
	case PROP_OPERATION:
		if (pipeline->priv->operation != NULL) {
			/* Can only set an operation once */
			return;
		}

		pipeline->priv->operation = g_value_get_object (value);
		g_object_ref (G_OBJECT (pipeline->priv->operation));
		break;

	default:
		break;
	}
}

static void
get_property (GObject *object,
	      guint prop_id,
	      GValue *value,
	      GParamSpec *pspec)
{
	MarlinPipeline *pipeline = MARLIN_PIPELINE (object);

	switch (prop_id) {
	case PROP_OPERATION:
		g_value_set_object (value, pipeline->priv->operation);
		break;

	default:
		break;
	}
}

static int
get_progress (MarlinPipeline *pipeline)
{
	g_warning ("The default get_process function was called - FIXME");
	return 0;
}

static void
class_init (GObjectClass *klass)
{
	MarlinPipelineClass *pclass;

	pclass = MARLIN_PIPELINE_CLASS (klass);
	
	klass->finalize = finalize;
	klass->set_property = set_property;
	klass->get_property = get_property;

	pclass->get_progress = get_progress;
	
	parent_class = g_type_class_peek_parent (klass);

	g_object_class_install_property (klass,
					 PROP_OPERATION,
					 g_param_spec_object ("operation",
							      "Operation",
							      "The operation this pipeline performs",
							      MARLIN_OPERATION_TYPE,
							      G_PARAM_READWRITE));

	signals[SHUTDOWN] = g_signal_new ("shutdown",
					  G_TYPE_FROM_CLASS (klass),
					  G_SIGNAL_RUN_FIRST |
					  G_SIGNAL_NO_RECURSE,
					  G_STRUCT_OFFSET (MarlinPipelineClass, shutdown),
					  NULL, NULL,
					  g_cclosure_marshal_VOID__VOID,
					  G_TYPE_NONE, 0);
	signals[EOS] = g_signal_new ("eos",
				     G_TYPE_FROM_CLASS (klass),
				     G_SIGNAL_RUN_FIRST |
				     G_SIGNAL_NO_RECURSE,
				     G_STRUCT_OFFSET (MarlinPipelineClass,
						      eos),
				     NULL, NULL,
				     g_cclosure_marshal_VOID__VOID,
				     G_TYPE_NONE, 0);
	signals[POSITION_CHANGED] = g_signal_new ("position-changed",
						  G_TYPE_FROM_CLASS (klass),
						  G_SIGNAL_RUN_FIRST |
						  G_SIGNAL_NO_RECURSE,
						  G_STRUCT_OFFSET (MarlinPipelineClass,
								   position_changed),
						  NULL, NULL,
						  marlin_marshal_VOID__UINT64,
						  G_TYPE_NONE, 1,
						  G_TYPE_UINT64);
}

/* Shutdown multithread stuff */
struct _shutdown_msg {
	MarlinMTMsg msg;
	
	MarlinPipeline *pipeline;
};

static void
do_shutdown (MarlinMTMsg *mm)
{
	struct _shutdown_msg *eosm = (struct _shutdown_msg *) mm;

	g_signal_emit (G_OBJECT (eosm->pipeline), signals[SHUTDOWN], 0);
}

static MarlinMTMsgOp shutdown_op = {
	NULL,
	do_shutdown,
	NULL,
	NULL
};

static void
shutdown (GstElement *element,
	  MarlinPipeline *pipeline)
{
	struct _shutdown_msg *m;

	m = marlin_mt_msg_new (&shutdown_op, NULL, sizeof (*m));
	m->pipeline = pipeline;

	marlin_msgport_put (marlin_gui_port, (MarlinMsg *) m);
}

struct _eos_msg {
	MarlinMTMsg msg;
	
	MarlinPipeline *pipeline;
};

static void
do_eos (MarlinMTMsg *mm)
{
	struct _eos_msg *eosm = (struct _eos_msg *) mm;

	g_signal_emit (G_OBJECT (eosm->pipeline), signals[EOS], 0);
	if (eosm->pipeline->priv->operation != NULL) {
		marlin_operation_finish (eosm->pipeline->priv->operation);
		eosm->pipeline->priv->started = FALSE;
	}
}

static MarlinMTMsgOp eos_op = {
	NULL,
	do_eos,
	NULL,
	NULL
};

void
marlin_pipeline_set_eos (MarlinPipeline *pipeline)
{
	struct _eos_msg *m;

	m = marlin_mt_msg_new (&eos_op, NULL, sizeof (*m));
	m->pipeline = pipeline;

	marlin_msgport_put (marlin_gui_port, (MarlinMsg *) m);
}

struct _position_msg {
	MarlinMTMsg msg;

	MarlinPipeline *pipeline;
	guint64 position;
};

static void
do_position (MarlinMTMsg *mm)
{
	struct _position_msg *msg = (struct _position_msg *) mm;

	g_signal_emit (G_OBJECT (msg->pipeline), signals[POSITION_CHANGED],
		       0, msg->position);
}

static MarlinMTMsgOp position_op = {
	NULL,
	do_position,
	NULL,
	NULL
};

void
marlin_pipeline_position_changed (MarlinPipeline *pipeline,
				  guint64 position)
{
	struct _position_msg *m;

	m = marlin_mt_msg_new (&position_op, NULL, sizeof (*m));
	m->pipeline = pipeline;
	m->position = position;
	
	marlin_msgport_put (marlin_gui_port, (MarlinMsg *) m);
}

static void
pipeline_error (GstElement *element,
		GstElement *source,
		GError *error,
		char *debug,
		MarlinPipeline *pipeline)
{
	if (pipeline->priv->operation) {
		marlin_operation_set_error (pipeline->priv->operation, 
					    error, debug);
	}
}

static void
init (MarlinPipeline *pipeline)
{
	pipeline->priv = g_new (MarlinPipelinePrivate, 1);

	pipeline->priv->pipeline = gst_thread_new ("pipeline");
	pipeline->priv->shutdown_id = g_signal_connect (pipeline->priv->pipeline,
							"shutdown",
							G_CALLBACK (shutdown),
							pipeline);

	g_signal_connect (pipeline->priv->pipeline, "error",
			  G_CALLBACK (pipeline_error), pipeline);
	
	pipeline->priv->started = FALSE;
	pipeline->priv->operation = NULL;
}

GType
marlin_pipeline_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		static GTypeInfo info = {
			sizeof (MarlinPipelineClass), NULL, NULL,
			(GClassInitFunc) class_init, NULL, NULL,
			sizeof (MarlinPipeline), 0, (GInstanceInitFunc) init
		};

		type = g_type_register_static (G_TYPE_OBJECT,
					       "MarlinPipeline",
					       &info, 0);
	}

	return type;
}

void
marlin_pipeline_set_state (MarlinPipeline *pipeline,
			   GstElementState state)
{
	g_return_if_fail (IS_MARLIN_PIPELINE (pipeline));

	gst_element_set_state (pipeline->priv->pipeline, state);
	if (state == GST_STATE_PLAYING &&
	    pipeline->priv->started == FALSE &&
	    pipeline->priv->operation != NULL) {
		marlin_operation_start (pipeline->priv->operation);
		pipeline->priv->started = TRUE;
	}
}

GstElementState
marlin_pipeline_get_state (MarlinPipeline *pipeline)
{
	return gst_element_get_state (pipeline->priv->pipeline);
}
			   
void
marlin_pipeline_eos (MarlinPipeline *pipeline)
{
	g_return_if_fail (IS_MARLIN_PIPELINE (pipeline));

	gst_element_set_eos (pipeline->priv->pipeline);
}

void
marlin_pipeline_add (MarlinPipeline *pipeline,
		     GstElement *element)
{
	g_return_if_fail (IS_MARLIN_PIPELINE (pipeline));

	gst_bin_add (GST_BIN (pipeline->priv->pipeline), element);
}

void
marlin_pipeline_remove (MarlinPipeline *pipeline,
			GstElement *element)
{
	g_return_if_fail (IS_MARLIN_PIPELINE (pipeline));

	gst_bin_remove (GST_BIN (pipeline->priv->pipeline), element);
}

void
marlin_pipeline_progress_changed (MarlinPipeline *pipeline)
{
	MarlinPipelineClass *pclass;
	int percentage;

	pclass = MARLIN_PIPELINE_GET_CLASS (pipeline);
	percentage = pclass->get_progress (pipeline);

	marlin_operation_progress (pipeline->priv->operation, percentage);
}

GstElement *
marlin_pipeline_get_element_by_name (MarlinPipeline *pipeline,
				     const char *name)
{
	return gst_bin_get_by_name (GST_BIN (pipeline->priv->pipeline), name);
}

void
marlin_pipeline_set_iterate_func (MarlinPipeline *pipeline,
				  GCallback *func,
				  gpointer closure)
{
	g_signal_connect (G_OBJECT (pipeline->priv->pipeline), "iterate",
			  func, closure);
}

void
marlin_pipeline_set_error (MarlinPipeline *pipeline,
			   GError *error,
			   const char *debug_message)
{
	if (pipeline->priv->operation) {
		marlin_operation_set_error (pipeline->priv->operation,
					    error, debug_message);
	}
}
