/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 *  Authors: Iain Holmes <iain@prettypeople.org>
 *
 *  Copyright 2002 Iain Holmes
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of version 2 of the GNU General Public 
 *  License as published by the Free Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *
 */

#include <glib.h>

#include <glib/gi18n.h>

#include <gst/gst.h>

#include <marlin/marlin-sample-element.h>
#include <marlin/marlin-block.h>
#include <marlin/marlin-channel.h>

typedef struct _SinkInputChannel {
	float *data;
	guint32 frames_in_data;
	gboolean eos;
	guint64 total;
	
	GstPad *sinkpad;
	guint32 unlink_id;
	GstElement *element;

	MarlinBlock *block_list, *bl_end;
} SinkInputChannel;

GstPadTemplate *sink_template = NULL;

static GstElementClass *sink_parent_class = NULL;

enum {
	PROP_SINK_0,
	PROP_SINK_SAMPLE,
	PROP_SINK_INSERT_POSITION
};

static void
sink_finalize (GObject *object)
{
	MarlinSampleElementSink *sink;
	GList *p;

	sink = MARLIN_SAMPLE_ELEMENT_SINK (object);

	for (p = sink->input_channels; p; p = p->next) {
		SinkInputChannel *c = p->data;
		
		g_free (c->data);
		g_free (c);
	}
	g_list_free (sink->input_channels);
	sink->input_channels = NULL;

	if (sink->no_space_error != NULL) {
		g_error_free (sink->no_space_error);
	}

	G_OBJECT_CLASS (sink_parent_class)->finalize (object);
}

static void
sink_dispose (GObject *object)
{
	MarlinSampleElementSink *sink;
	GList *p;
	
	sink = MARLIN_SAMPLE_ELEMENT_SINK (object);
	
	if (sink->sample != NULL) {
		g_object_unref (G_OBJECT (sink->sample));
		sink->sample = NULL;
	}

	/* Loop though the channels, turning off the unlink signal */
	for (p = sink->input_channels; p; p = p->next) {
		SinkInputChannel *c = p->data;
		
		g_signal_handler_disconnect (c->sinkpad, c->unlink_id);
	}
	
 	G_OBJECT_CLASS (sink_parent_class)->dispose (object);
}

/* Store a complete MARLIN_BLOCK */
static void
store_block (MarlinSampleElementSink *sink,
	     SinkInputChannel *c,
	     int channel_number,
	     float *data,
	     guint number_of_frames)
{
	MarlinSample *sample = sink->sample;
	MarlinChannel *channel;
	MarlinBlock *block;
	GError *error = NULL;
	gboolean ret;
	
	sink->frame_count += number_of_frames;
	
	if (sink->no_space) {
		return;
	}
	
	channel = marlin_sample_get_channel (sample, channel_number);
	block = marlin_channel_create_block (channel);
	
	ret = marlin_block_set_data (block, data, number_of_frames, &error);
	
	if (ret == FALSE) {
		if (error &&
		    error->code == MARLIN_FILE_ERROR_NO_SPACE) {
			sink->no_space = TRUE;
			sink->no_space_error = g_error_copy (error);
			sink->no_space_frames = sink->frame_count;
			
		} else if (error) {
 			GST_ELEMENT_ERROR (GST_ELEMENT (sink), RESOURCE, WRITE, (_("Error writing data")), ("%s", error->message));
		}

		marlin_block_free (block);
		return;
	}

	/* Store the new block in the list for linking once the element has
	   finished */
	if (c->block_list == NULL) {
		c->block_list = block;
		c->bl_end = block;
	} else {
		marlin_block_append (c->bl_end, block);
		c->bl_end = block;
	}
}

static gboolean
all_pads_eos (GList *pads)
{
	for (;pads; pads = pads->next) {
		SinkInputChannel *c = pads->data;
		
		if (c->eos == FALSE) {
			return FALSE;
		}
	}
	
	return TRUE;
}

static int
get_percent_from_element (GstElement *element)
{
	gint64 total, position;
	gboolean ret;
	GstFormat format = GST_FORMAT_BYTES;
	
	ret = gst_element_query (element, GST_QUERY_POSITION,
				 &format, &position);
	if (ret == FALSE) {
		return 0;
	}

	format = GST_FORMAT_BYTES;
	ret = gst_element_query (element, GST_QUERY_TOTAL,
				 &format, &total);
	if (ret == FALSE) {
		return 0;
	}

	return (int) ((float)(position * 100) / (float) (total));
}

static void
sink_parse_tag (MarlinSampleElementSink *sink,
		GstEvent *event)
{
	GstTagList *tags, *copy;

	tags = gst_event_tag_get_list (event);
	copy = gst_tag_list_copy (tags);

	g_object_set (G_OBJECT (sink->sample),
		      "taglist", copy,
		      NULL);
}

static void
marlin_sample_element_sink_loop (GstElement *element)
{
	MarlinSampleElementSink *sink;
	GstEvent *event = NULL;
	GList *inputs;
	SinkInputChannel *channel;
	guint8 *raw_in;
	guint32 got_bytes;
	int i;
	gboolean got_eos = FALSE;
	
	sink = MARLIN_SAMPLE_ELEMENT_SINK (element);

	while (got_eos == FALSE) {
		inputs = sink->input_channels;
		i = 0;

		if (inputs == NULL) {
			return;
		}
		
		while (inputs) {
			GstData *data;
			GstBuffer *buf;
			guint32 frames;
			float *in;

		start:
			channel = inputs->data;
			
			if (!GST_PAD_IS_USABLE (channel->sinkpad)
			    || channel->eos) {
				inputs = inputs->next;
				continue;
			}
			
			/* Get data from the pad of each input channel.
			   We need to check for events before passing on the
			   data to the sample. */
			do {
				data = gst_pad_pull (channel->sinkpad);

				if (GST_IS_EVENT (data)) {
					switch (GST_EVENT_TYPE (data)) {
					case GST_EVENT_EOS:
						channel->eos = TRUE;
						goto end;
						
					case GST_EVENT_TAG:
						sink_parse_tag (sink, GST_EVENT (data));
						break;
						
					default:
						gst_pad_event_default (channel->sinkpad, GST_EVENT (data));
						break;
					}
					data = NULL;
				}
			} while (data == NULL);

			buf = GST_BUFFER (data);

			/* Here's where the data gets copied. */
			got_bytes = GST_BUFFER_SIZE (buf);

			/* This increments every buffer */
			frames = got_bytes / sizeof (float);
/* 			frames = sink->buffer_frames; */
			in = (float *) GST_BUFFER_DATA (buf);

			if (channel->frames_in_data + frames >= MARLIN_BLOCK_SIZE) {
				guint32 needed, remain;
				int percent;
				
				needed = MARLIN_BLOCK_SIZE - channel->frames_in_data;
				remain = frames - needed;
				
				memcpy (channel->data + channel->frames_in_data,
					in, needed * sizeof (float));

				percent = get_percent_from_element (element);
/* 				marlin_sample_op_progress (sink->sample, percent, NULL); */
				
				store_block (sink, channel, i, channel->data, MARLIN_BLOCK_SIZE);

				g_object_set (G_OBJECT (sink->sample),
					      "total-frames", (guint64) channel->total,
					      NULL);
				/* Clear the block */
				memset (channel->data, 0,
					MARLIN_BLOCK_SIZE * sizeof (float));
				/* Move the rest of the data across */
				if (remain > 0) {
					memmove (channel->data,
						 in + needed,
						 remain * sizeof (float));
				}
				
				channel->frames_in_data = remain;
			} else {
				memcpy (channel->data + channel->frames_in_data,
					in, got_bytes);

				channel->frames_in_data += frames;
			}
						
			channel->total += frames;
			gst_buffer_unref (buf);
			
			inputs = inputs->next;
			i++;
		}

		return;
	}

 end:

	for (i = 0, inputs = sink->input_channels; inputs; inputs = inputs->next, i++) {
		SinkInputChannel *channel = inputs->data;
		
		if (channel->frames_in_data > 0) {
			store_block (sink, channel, i, channel->data, 
				     channel->frames_in_data);

			g_object_set (G_OBJECT (sink->sample),
				      "total_frames", (guint64) channel->total,
				      NULL);
/* 			marlin_sample_op_progress (sink->sample, 100, NULL); */
		}

	}

	if (sink->no_space) {
		guint64 bytes_needed, bytes_have;
		char *need, *have, *debug;

		bytes_needed = (sink->frame_count * sizeof (float)) + ((sink->frame_count / 128) * (sizeof (float) * 4));
		bytes_have = sink->no_space_frames * sizeof (float);

		need = gnome_vfs_format_file_size_for_display (bytes_needed);
		have = gnome_vfs_format_file_size_for_display (bytes_have);

		GST_ELEMENT_ERROR (GST_ELEMENT (sink), RESOURCE, WRITE, 
				   ("Marlin ran out of temporary space.\n%s of free space is needed on '/tmp' but you only have %s", need, have), 
				   ("%s", sink->no_space_error->message));
		g_free (need);
		g_free (have);

		return;
	} else {
		for (i = 0, inputs = sink->input_channels; inputs; inputs = inputs->next, i++) {
			SinkInputChannel *c = inputs->data;
			MarlinChannel *channel;
			
			channel = marlin_sample_get_channel (sink->sample, i);
			channel->first = c->block_list;
			channel->last = c->bl_end;

			channel->frames = marlin_block_recalculate_ranges (channel->first);
		}

		gst_element_set_eos (element);
	}
}

static GstPadConnectReturn
marlin_sample_element_sink_link (GstPad *pad,
			         const GstCaps *caps)
{
	MarlinSampleElementSink *sink;
	const GValue *value;
	GstStructure *structure;
	guint rate;
	
	sink = MARLIN_SAMPLE_ELEMENT_SINK (GST_PAD_PARENT (pad));

	structure = gst_caps_get_structure (caps, 0);
	value = gst_structure_get_value (structure, "rate");
	if (value == NULL) {
		return GST_PAD_CONNECT_DELAYED;
	}
	
	if (G_VALUE_TYPE (value) == G_TYPE_INT) {
		sink->rate = g_value_get_int (value);
	}
	g_object_get (G_OBJECT (sink->sample),
		      "sample_rate", &rate,
		      NULL);

	value = gst_structure_get_value (structure, "buffer-frames");
	if (value == NULL) {
		return GST_PAD_CONNECT_DELAYED;
	}
	
	if (G_VALUE_TYPE (value) == G_TYPE_INT) {
		sink->buffer_frames = g_value_get_int (value);
	}

	if (rate == sink->rate) {
		return GST_PAD_CONNECT_OK;
	}
	
	g_object_set (G_OBJECT (sink->sample),
		      "sample_rate", sink->rate,
		      NULL);

	return GST_PAD_CONNECT_OK;
}

static void
pad_unlinked (GstPad *pad,
	      GstPad *peer,
	      SinkInputChannel *channel)
{
	MarlinSampleElementSink *sink;

	sink = MARLIN_SAMPLE_ELEMENT_SINK (channel->element);

	sink->input_channels = g_list_remove (sink->input_channels, channel);
	gst_element_remove_pad (channel->element, pad);
	
	g_free (channel->data);

	g_free (channel);
	sink->numsinkpads--;
}

static GstPad *
request_new_pad (GstElement *element,
		 GstPadTemplate *templ,
		 const char *unused)
{
	char *name;
	MarlinSampleElementSink *sink;
	SinkInputChannel *input_channel;
	GstPad *pad;
	
	sink = MARLIN_SAMPLE_ELEMENT_SINK (element);

	if  (templ->direction != GST_PAD_SINK) {
		return NULL;
	}
	
	input_channel = g_new0 (SinkInputChannel, 1);
	
	name = g_strdup_printf ("sink%d", sink->numsinkpads);
	input_channel->sinkpad = gst_pad_new_from_template (templ, name);
	g_free (name);

	input_channel->unlink_id = g_signal_connect (G_OBJECT (input_channel->sinkpad),
						     "unlinked",
						     G_CALLBACK (pad_unlinked),
						     input_channel);
	
	input_channel->data = g_new (float, MARLIN_BLOCK_SIZE);
	input_channel->element = element;
	input_channel->frames_in_data = 0;
	input_channel->eos = FALSE;
	input_channel->total = 0;

	gst_element_add_pad (element, input_channel->sinkpad);
	gst_pad_set_connect_function (input_channel->sinkpad,
				      marlin_sample_element_sink_link);
	
	sink->input_channels = g_list_append (sink->input_channels, input_channel);
	sink->numsinkpads++;

	g_object_set (G_OBJECT (sink->sample),
		      "channels", sink->numsinkpads,
		      NULL);
	return input_channel->sinkpad;
}

static GstElementStateReturn
marlin_sample_element_sink_change_state (GstElement *element)
{
	MarlinSampleElementSink *sink = MARLIN_SAMPLE_ELEMENT_SINK (element);
	GList *p;
	
	switch (GST_STATE_TRANSITION (element)) {
		
	case GST_STATE_READY_TO_NULL:
		break;
		
	case GST_STATE_NULL_TO_READY:
		break;

	case GST_STATE_READY_TO_PAUSED:
		break;
		
	case GST_STATE_PAUSED_TO_READY:
		for (p = sink->input_channels; p; p = p->next) {
			SinkInputChannel *channel = p->data;

			memset (channel->data, 0, MARLIN_BLOCK_SIZE * sizeof (float));
			channel->total = 0;
			channel->frames_in_data = 0;
			channel->eos = FALSE;
		}

		break;

	default:
		break;
	}

	if (GST_ELEMENT_CLASS (sink_parent_class)->change_state) {
		return GST_ELEMENT_CLASS (sink_parent_class)->change_state (element);
	}

	return GST_STATE_SUCCESS;
}

static void
sink_set_property (GObject *object,
		   guint prop_id,
		   const GValue *value,
		   GParamSpec *pspec)
{
	MarlinSampleElementSink *sink;
	MarlinSample *sample;
	GList *p;
	
	sink = MARLIN_SAMPLE_ELEMENT_SINK (object);

	switch (prop_id) {
	case PROP_SINK_SAMPLE:
		sample = g_value_get_object (value);

		if (sink->sample != NULL) {
			g_object_unref (G_OBJECT (sink->sample));
		}
		
		sink->sample = sample;

		if (sample == NULL) {
			return;
		}

		g_object_ref (sample);

		g_object_set (G_OBJECT (sample),
			      "channels", sink->numsinkpads == 0 ? 2 : sink->numsinkpads,
			      "sample-rate", sink->rate == 0 ? 44100 : sink->rate,
			      NULL);

		/* Reset the channels */
		for (p = sink->input_channels; p; p = p->next) {
			SinkInputChannel *c = p->data;

			memset (c->data, 0, MARLIN_BLOCK_SIZE * sizeof (float));
			c->frames_in_data = 0;
			c->eos = FALSE;
			c->total = 0;
		}

		break;

	case PROP_SINK_INSERT_POSITION:
		break;

	default:
		break;
	}
}

static void
sink_get_property (GObject *object,
		   guint prop_id,
		   GValue *value,
		   GParamSpec *pspec)
{
}

static GstPadTemplate *
sink_factory (void)
{
	return gst_pad_template_new (
		"sink%d",
		GST_PAD_SINK,
		GST_PAD_REQUEST,
		gst_caps_new_simple ("audio/x-raw-float",
			"rate", GST_TYPE_INT_RANGE, 4000, 96000,
			"channels", G_TYPE_INT, 1,
			NULL));
}

static void
sink_class_init (MarlinSampleElementSinkClass *klass)
{
	GObjectClass *object_class;
	GstElementClass *element_class;
	
	object_class = G_OBJECT_CLASS (klass);
	element_class = GST_ELEMENT_CLASS (klass);
	
	element_class->request_new_pad = request_new_pad;

	object_class->finalize = sink_finalize;
	object_class->dispose = sink_dispose;
	object_class->set_property = sink_set_property;
	object_class->get_property = sink_get_property;

	element_class->change_state = marlin_sample_element_sink_change_state;
	if (sink_template == NULL) {
		sink_template = sink_factory ();
	}

	gst_element_class_add_pad_template (element_class, sink_template);

	sink_parent_class = g_type_class_ref (GST_TYPE_ELEMENT);

	/* Properties */
	g_object_class_install_property (object_class,
					 PROP_SINK_SAMPLE,
					 g_param_spec_object ("sample",
							      "Sample",
							      "The sample to be the destination",
							      MARLIN_SAMPLE_TYPE,
							      G_PARAM_READWRITE));
	g_object_class_install_property (object_class,
					 PROP_SINK_INSERT_POSITION,
					 g_param_spec_uint64 ("insert_position",
							      "", "",
							      0, G_MAXUINT64,
							      0, 
							      G_PARAM_READWRITE));
}

static void
sink_init (MarlinSampleElementSink *sink)
{
	gst_element_set_loop_function (GST_ELEMENT (sink),
				       marlin_sample_element_sink_loop);
	sink->numsinkpads = 0;
	sink->rate = 0;
	sink->input_channels = NULL;
	sink->no_space = FALSE;
	sink->no_space_error = NULL;
	sink->frame_count = 0;
}

GType
marlin_sample_element_sink_get_type (void)
{
	static GType type = 0;
	
	if (type == 0) {
		static GTypeInfo info = {
			sizeof (MarlinSampleElementSinkClass), NULL, NULL,
			(GClassInitFunc) sink_class_init, NULL, NULL,
			sizeof (MarlinSampleElementSink), 0,
			(GInstanceInitFunc) sink_init
		};
		
		type = g_type_register_static (GST_TYPE_ELEMENT, "MarlinSampleElementSink", &info, 0);
	}
	
	return type;
}

MarlinSampleElementSink *
marlin_sample_element_sink_new (MarlinSample *sample)
{
	MarlinSampleElementSink *sink;

	if (sample != NULL) {
		sink = g_object_new (MARLIN_SAMPLE_ELEMENT_SINK_TYPE,
				     "sample", sample,
				     NULL);
	} else {
		sink = g_object_new (MARLIN_SAMPLE_ELEMENT_SINK_TYPE, NULL);
	}
	
	gst_element_set_name (GST_ELEMENT (sink), "marlin-sample-sink");
	
	return sink;
}

/* ---------------------Src element--------------------*/
enum {
	PROP_SRC_0,
	PROP_SRC_SAMPLE,
	PROP_SRC_POSITION,
	PROP_SRC_SEND_NEW_MEDIA,
	PROP_SRC_LOOP
};

GstPadTemplate *src_template = NULL;
static GstElementClass *src_parent_class = NULL;

typedef struct _SrcOutputChannel {
	GstPad *srcpad;
	
	guint64 offset;
	guint32 offset_in_block;
	guint64 total_out;

	MarlinChannelPosition chan_pos;
	MarlinChannel *channel;
	MarlinBlock *current_block;

	gboolean seek_happened;
	gboolean need_flush;
	gboolean need_new_media;
} SrcOutputChannel;

static const GstEventMask *marlin_sample_element_src_get_event_mask (GstPad *pad);
static const GstFormat *marlin_sample_element_src_get_formats (GstPad *pad);
static const GstQueryType *marlin_sample_element_src_get_query_types (GstPad *pad);

static void
src_dispose (GObject *object)
{
	MarlinSampleElementSrc *src;
	GList *p;
	
	src = MARLIN_SAMPLE_ELEMENT_SRC (object);

	/* If we're not READY then the sample will still be locked */
	gst_element_set_state (GST_ELEMENT (src), GST_STATE_READY);

	if (src->sample != NULL) {
		g_signal_handler_disconnect (G_OBJECT (src->sample),
					     src->notify_id);
		g_object_unref (G_OBJECT (src->sample));
	}

	/* Destroy all the pad data */
	for (p = src->srcpads; p; p = p->next) {
		SrcOutputChannel *c = p->data;

		g_free (c);
	}
	g_list_free (src->srcpads);

	G_OBJECT_CLASS (src_parent_class)->dispose (object);
}
	
static SrcOutputChannel *
get_channel_for_pad (MarlinSampleElementSrc *src,
		     GstPad *pad)
{
	GList *p;

	for (p = src->srcpads; p; p = p->next) {
		SrcOutputChannel *channel = p->data;

		if (channel->srcpad == pad) {
			return channel;
		}
	}

	return NULL;
}

static MarlinMarker *
get_next_marker (MarlinSampleElementSrc *src,
		 guint64 frame)
{
	static GList *marker = NULL;

	if (marker == NULL) {
		marker = src->markers;
	}

	if (frame < ((MarlinMarker *) marker->data)->position) {
		return marker->data;
	}

	for (; marker; marker = marker->next) {
		MarlinMarker *m = marker->data;

		if (m->position > frame) {
			return m;
		}
	}

	return NULL;
}

static GstData *
marlin_sample_element_src_get (GstPad *pad)
{
	SrcOutputChannel *channel;
	guint32 readsize, left_in_block;
	guint64 readend;
	float *out;
	MarlinSampleElementSrc *src;
	GstBuffer *buf = NULL;
	GList *outputs;
	guint32 percent;
	float *frame_data;
	guint64 next_marker;
	
	src = MARLIN_SAMPLE_ELEMENT_SRC (gst_pad_get_parent (pad));

	if (src->sample == NULL) {
		return NULL;
	}

	channel = get_channel_for_pad (src, pad);
	if (channel == NULL) {
		g_warning ("Channel for pad %s == NULL", gst_pad_get_name (pad));
		return NULL;
	}

	if (channel->need_new_media) {
		GstEvent *event;

		event = gst_event_new_discontinuous (TRUE, GST_FORMAT_BYTES,
						     channel->offset * sizeof (float));
		GST_EVENT_DISCONT_NEW_MEDIA (event) = TRUE;

		channel->need_new_media = FALSE;
		src->sent_first = TRUE;
		return GST_DATA (event);
	}

	/* Check if we need to send metadata tags */
	if (src->need_taglist) {
		GstEvent *event;
		GstTagList *taglist;

		g_object_get (G_OBJECT (src->sample),
			      "taglist", &taglist,
			      NULL);

		src->need_taglist = FALSE;

		if (taglist != NULL) {
			event = gst_event_new_tag (gst_tag_list_copy (taglist));
			return GST_DATA (event);
		}
	}
	
	/* Check for seek */
	if (channel->seek_happened) {
		GstEvent *event;

		channel->seek_happened = FALSE;

		/* This should not be in bytes... */
		event = gst_event_new_discontinuous (FALSE, GST_FORMAT_BYTES,
						     channel->offset * sizeof (float));
		GST_EVENT_DISCONT_NEW_MEDIA (event) = FALSE;
		channel->need_flush = FALSE;
		return GST_DATA (event);
	}
	
	/* Check for flush */
	if (channel->need_flush) {
		channel->need_flush = FALSE;
		
		return GST_DATA (gst_event_new_flush ());
	}
	
	/* Check for EOF */
	if (channel->offset == src->end) {
		if (src->loop == FALSE) {
			gst_element_set_eos (GST_ELEMENT (src));
			return GST_DATA (gst_event_new (GST_EVENT_EOS));
		} else {
			/* Reset everything back to the start */
			channel->channel = marlin_sample_get_channel (src->sample, channel->chan_pos);
			g_assert (channel->channel != NULL);

			channel->current_block = marlin_channel_get_block_for_frame (channel->channel, src->start);
			
			channel->offset = src->start;
			channel->offset_in_block = channel->offset - channel->current_block->start;
		}

	}

	if (channel->current_block == NULL) {
		channel->current_block = marlin_channel_get_block_for_frame (channel->channel, channel->offset);
		channel->offset_in_block = channel->offset - channel->current_block->start;
	}

	/* In frames */
	readsize = src->block_size; /* In frames */
	readend = channel->offset + readsize;
	
	if (src->send_new_media) {
		MarlinMarker *m = get_next_marker (src, channel->offset);
		
		if (m) {
			readsize = MIN (src->block_size, 
					m->position - channel->offset);
			readend = channel->offset + readsize;
			if (readend == m->position) {
				channel->need_new_media = TRUE;
			}
		}
	}

	if (readend > src->end) {
		readsize = src->end - channel->offset;
		readend = (guint64) (channel->offset + readsize);
		channel->need_new_media = FALSE;
		src->new_media_marker = NULL;
	}

	/* Create a buffer for readsize */
	buf = gst_buffer_new ();
	GST_BUFFER_DATA (buf) = (gint8 *) g_new (float, readsize);
 	GST_BUFFER_TIMESTAMP (buf) = (GstClockTime) (channel->offset * GST_SECOND) / src->sample_rate;
	
	out = (float *) GST_BUFFER_DATA (buf);
	channel->total_out += readsize;
	
	left_in_block = channel->current_block->num_frames - channel->offset_in_block;

	WRITE_LOCK (channel->channel->lock);

	frame_data = marlin_block_get_frame_data (channel->current_block, NULL);
	memcpy (out, frame_data + channel->offset_in_block,
		sizeof (float) * MIN (readsize, left_in_block));

	if (left_in_block < readsize) {
#if 0
		guint remaining = readsize - left_in_block;

		channel->current_block = marlin_block_next (channel->current_block);
		g_assert (channel->current_block != NULL);

		frame_data = marlin_block_get_frame_data (channel->current_block, NULL);
		memcpy (out + left_in_block, frame_data, 
			sizeof (float) * remaining);
#endif

		readsize = left_in_block;
	}
	
	GST_BUFFER_SIZE (buf) = readsize * sizeof (float);

	channel->offset += readsize;
	if (channel->offset < src->end &&
	    channel->offset > channel->current_block->end){
		channel->current_block = marlin_block_next (channel->current_block);
		g_assert (channel->current_block != NULL);
	}
	channel->offset_in_block = channel->offset - channel->current_block->start;

	WRITE_UNLOCK (channel->channel->lock);
	return GST_DATA (buf);
}

static gboolean
marlin_sample_element_src_event (GstPad *pad,
				 GstEvent *event)
{
	SrcOutputChannel *channel;
	MarlinSampleElementSrc *src;
	guint64 offset;
	guint64 frame_offset;
	int format;
	
	src = MARLIN_SAMPLE_ELEMENT_SRC (gst_pad_get_parent (pad));

	if (src->sample == NULL) {
		goto error;
	}
	
	channel = get_channel_for_pad (src, pad);
	if (channel == NULL) {
		goto error;
	}

	switch (GST_EVENT_TYPE (event)) {
	case GST_EVENT_SEEK:
		format = GST_EVENT_SEEK_FORMAT (event);

		offset = GST_EVENT_SEEK_OFFSET (event);
		frame_offset = (guint64) (offset);

		switch (GST_EVENT_SEEK_METHOD (event)) {
		case GST_SEEK_METHOD_SET:
			if (frame_offset > src->num_frames) {
				goto error;
			}

			channel->offset = frame_offset;
			channel->current_block = NULL;
			break;

		case GST_SEEK_METHOD_CUR:
			if (frame_offset + channel->offset > src->num_frames) {
				goto error;
			}

			channel->offset += frame_offset;
			channel->current_block = NULL;
			break;

		case GST_SEEK_METHOD_END:
			if (ABS (frame_offset) > src->num_frames) {
				goto error;
			}

			channel->offset = src->num_frames - ABS (frame_offset);
			channel->current_block = NULL;
			break;

		default:
			goto error;
			break;
		}

		channel->seek_happened;
		channel->need_flush = GST_EVENT_SEEK_FLAGS (event) & GST_SEEK_FLAG_FLUSH;
		break;

	case GST_EVENT_SEEK_SEGMENT:
		format = GST_EVENT_SEEK_FORMAT (event);
#if 0
		if (format != GST_FORMAT_BYTES
		    || format != GST_FORMAT_UNITS) {
			g_print ("Invalid format %d\n", format);
			goto error;
		}
#endif
		switch (format) {
		case GST_FORMAT_BYTES:
			src->start = GST_EVENT_SEEK_OFFSET (event) / sizeof (float);
			src->end = GST_EVENT_SEEK_ENDOFFSET (event) / sizeof (float);
			break;

		case GST_FORMAT_DEFAULT:
			src->start = GST_EVENT_SEEK_OFFSET (event);
			src->end = GST_EVENT_SEEK_ENDOFFSET (event);
			break;

		default:
			g_print ("Unknown seek format\n");
			goto error;
		}

		if (src->end > src->num_frames) {
			src->end = src->num_frames;
		}
		
		src->segment_loop = GST_EVENT_SEEK_FLAGS (event) & GST_SEEK_FLAG_SEGMENT_LOOP;
		
		channel->offset = src->start;
		channel->current_block = NULL;
		break;
		
	case GST_EVENT_SIZE:
		if (GST_EVENT_SIZE_FORMAT (event) != GST_FORMAT_BYTES) {
			goto error;
		}
		
		/*src->block_size = GST_EVENT_SIZE_VALUE (event) / sizeof (float);*/
		break;

	case GST_EVENT_FLUSH:
		channel->need_flush = TRUE;
		break;

	case GST_EVENT_DISCONTINUOUS:
		return TRUE;
		break;

	case GST_EVENT_EOS:
		return TRUE;
		break;
		
	default:
		goto error;
		break;
	}

	return TRUE;
	
 error:
	return FALSE;
}

static GstElementStateReturn
marlin_sample_element_src_change_state (GstElement *element)
{
	MarlinSampleElementSrc *src = MARLIN_SAMPLE_ELEMENT_SRC (element);
	GstCaps *srccaps;
	GList *p;

	switch (GST_STATE_TRANSITION (element)) {
		
	case GST_STATE_READY_TO_NULL:
		break;
		
	case GST_STATE_NULL_TO_READY:
		break;

	case GST_STATE_READY_TO_PAUSED:
		break;
		
	case GST_STATE_PAUSED_TO_READY:
		for (p = src->srcpads; p; p = p->next) {
			SrcOutputChannel *channel = p->data;

			channel->total_out = 0;
			channel->offset = 0;
			channel->offset_in_block = 0;
		}
		
		break;

	case GST_STATE_PLAYING_TO_PAUSED:
		for (p = src->srcpads; p; p = p->next) {
			SrcOutputChannel *channel = p->data;

			channel->channel = NULL;
			channel->current_block = NULL;
		}

		marlin_sample_read_unlock (src->sample);
		break;

	case GST_STATE_PAUSED_TO_PLAYING:
		marlin_sample_read_lock (src->sample);

		for (p = src->srcpads; p; p = p->next) {
			SrcOutputChannel *channel = p->data;

			channel->channel = marlin_sample_get_channel (src->sample, channel->chan_pos);
			g_assert (channel->channel != NULL);

			channel->current_block = marlin_channel_get_block_for_frame (channel->channel, 0);
		}
		break;

	default:
		break;
	}

	if (GST_ELEMENT_CLASS (src_parent_class)->change_state) {
		return GST_ELEMENT_CLASS (src_parent_class)->change_state (element);
	}

	return GST_STATE_SUCCESS;
}
		
static gboolean
marlin_sample_element_src_query (GstPad *pad,
				 GstQueryType type,
				 GstFormat *format,
				 gint64 *value)
{
	MarlinSampleElementSrc *src = MARLIN_SAMPLE_ELEMENT_SRC (gst_pad_get_parent (pad));
	guint64 pos;

	switch (type) {
	case GST_QUERY_TOTAL:
		if (*format == GST_FORMAT_BYTES) {
			*value = (gint64) (src->num_frames * sizeof (float));
		} else {
			*value = (gint64) src->num_frames;
		}
		break;

	case GST_QUERY_POSITION:
		pos = ((SrcOutputChannel *) src->srcpads->data)->offset;
		if (*format == GST_FORMAT_BYTES) {
			*value = (gint64) (pos * sizeof (float));
		} else {
			*value = (gint64) pos;
		}
		break;

	case GST_QUERY_START:
		break;

	case GST_QUERY_SEGMENT_END:
		break;
		
	default:
		return FALSE;
		break;
	}

	return TRUE;
}

static GstCaps *
explicit_getcaps (GstPad *pad)
{
	MarlinSampleElementSrc *src;

	src = MARLIN_SAMPLE_ELEMENT_SRC (GST_PAD_PARENT (pad));

	return gst_caps_new_simple ("audio/x-raw-float",
				    "rate", G_TYPE_INT, src->sample_rate,
				    "channels", G_TYPE_INT, 1,
				    "endianness", G_TYPE_INT, G_BYTE_ORDER,
				    "width", G_TYPE_INT, 32,
				    "buffer-frames", G_TYPE_INT, src->block_size,
				    NULL);
}

static GstPadLinkReturn 
explicit_link (GstPad *pad,
	       const GstCaps *caps)
{
	return GST_PAD_LINK_OK;
}

static SrcOutputChannel *
create_channel (MarlinSampleElementSrc *src,
		const char *name,
		int channel_number)
{
	SrcOutputChannel *channel;
	
	channel = g_new0 (SrcOutputChannel, 1);
	
	channel->srcpad = gst_pad_new_from_template (src_template, name);
	gst_pad_set_getcaps_function (channel->srcpad, explicit_getcaps);
	gst_pad_set_link_function (channel->srcpad, explicit_link);

	gst_pad_set_get_function (channel->srcpad,
				  marlin_sample_element_src_get);
	gst_pad_set_event_function (channel->srcpad,
				    marlin_sample_element_src_event);
	gst_pad_set_event_mask_function (channel->srcpad,
					 marlin_sample_element_src_get_event_mask);
	gst_pad_set_query_function (channel->srcpad,
				    marlin_sample_element_src_query);
	gst_pad_set_query_type_function (channel->srcpad,
					 marlin_sample_element_src_get_query_types);
	gst_pad_set_formats_function (channel->srcpad,
				      marlin_sample_element_src_get_formats);
	
	channel->chan_pos = channel_number;
	
	/* These will get set up in the state_change function */
	channel->channel = NULL;
	channel->current_block = NULL;

	gst_element_add_pad (GST_ELEMENT (src), channel->srcpad);

	return channel;
}

static void
create_srcpads (MarlinSampleElementSrc *src)
{
	int i;

	for (i = 0; i < src->numsrcpads; i++) {
		SrcOutputChannel *channel;
		char *name;

		name = g_strdup_printf ("src%d", i);
		channel = create_channel (src, name, i);
		g_free (name);
		
		src->srcpads = g_list_append (src->srcpads, channel);
	}
}

static void
remove_srcpads (MarlinSampleElementSrc *src)
{
	GList *p;

	/* This should probably be locked */
	for (p = src->srcpads; p; p = p->next) {
		GstPad *peer;
		SrcOutputChannel *channel = p->data;

		peer = gst_pad_get_peer (channel->srcpad);
		if (peer) {
			gst_pad_disconnect (channel->srcpad, peer);
		}
		gst_element_remove_pad (GST_ELEMENT (src), channel->srcpad);
		g_free (channel);
	}

	g_list_free (src->srcpads);
	src->srcpads = NULL;
	src->numsrcpads = 0;
}

static void
sample_notify (GObject *object,
	       GParamSpec *pspec,
	       MarlinSampleElementSrc *src)
{
	const char *param_name;

	param_name = g_param_spec_get_name (pspec);
	if (strcmp (param_name, "total-frames") == 0) {
		g_object_get (object,
			      "total_frames", &src->num_frames,
			      NULL);
		return;
	}

	if (strcmp (param_name, "sample-rate") == 0) {
		guint new_rate;
		GstElementState old_state;
		GList *p;
		
		g_object_get (object,
			      "sample_rate", &new_rate,
			      NULL);

		if (new_rate == src->sample_rate) {
			return;
		}

		old_state = gst_element_get_state (GST_ELEMENT (src));
		if (old_state == GST_STATE_PLAYING) {
			g_warning ("Cannot change rate while pipeline is playing");
			return;
		}

		src->sample_rate = new_rate;
		/* Loop though the pads, disconnecting them,
		   changing the caps and then reconnecting them */
		for (p = src->srcpads; p; p = p->next) {
			GstCaps *srccaps;
			SrcOutputChannel *channel = p->data;

			srccaps = gst_pad_get_caps (channel->srcpad);
			gst_caps_set_simple (srccaps, "rate",
				      G_TYPE_INT, src->sample_rate, NULL);
			
			gst_pad_try_set_caps (channel->srcpad, srccaps);
			gst_caps_free (srccaps);
		}

		return;
	}

	if (strcmp (param_name, "channels") == 0) {
		guint new_channels;
		int dc, start, end;

		g_object_get (object,
			      "channels", &new_channels,
			      NULL);

		dc = src->numsrcpads - new_channels;
		if (dc == 0) {
			/* No change */
			return;
		} else if (dc < 0) {
			int i;
			
			/* Add more channels */
			start = src->numsrcpads;
			src->numsrcpads = new_channels;
			for (i = start; i < src->numsrcpads; i++) {
				SrcOutputChannel *channel;
				char *name;
				
				name = g_strdup_printf ("src%d", i);
				channel = create_channel (src, name, i);
				g_free (name);
				
				src->srcpads = g_list_append (src->srcpads,
							      channel);
			}

			return;
		} else if (dc > 0) {
			GList *p;

			if (dc > src->numsrcpads) {
				g_warning ("Not enough channels to remove %d\nWe only have %d", dc, src->numsrcpads);
				return;
			}
			
			/* Remove channels */
			for (p = g_list_last (src->srcpads);
			     (dc > 0) && p; dc--) {
				GstPad *peer;
				GList *p_copy = p;
				SrcOutputChannel *channel = p->data;
				
				peer = gst_pad_get_peer (channel->srcpad);
				if (peer) {
					gst_pad_disconnect (channel->srcpad, peer);
				}
				gst_element_remove_pad (GST_ELEMENT (src), channel->srcpad);
				g_free (channel);
				
				/* Get the next end...*/
				p = p->prev;
				
				/* Remove the link */
				src->srcpads = g_list_remove_link (src->srcpads, p_copy);
				g_list_free (p_copy);
			}
			
			src->numsrcpads = new_channels;

			return;
		}

		return;
	}
}

static void
set_property (GObject *object,
	      guint prop_id,
	      const GValue *value,
	      GParamSpec *pspec)
{
	MarlinSampleElementSrc *src;
	MarlinSample *sample;
	
	src = MARLIN_SAMPLE_ELEMENT_SRC (object);

	switch (prop_id) {
	case PROP_SRC_SAMPLE:

		sample = g_value_get_object (value);

		if (src->sample != NULL) {
			remove_srcpads (src);
			g_signal_handler_disconnect (G_OBJECT (src->sample),
						     src->notify_id);
			g_object_unref (G_OBJECT (src->sample));
		}
		src->sample = sample;
		src->notify_id = g_signal_connect (G_OBJECT (sample),
						   "notify",
						   G_CALLBACK (sample_notify),
						   src);
		g_object_ref (G_OBJECT (sample));
		
		g_object_get (G_OBJECT (sample),
			      "channels", &src->numsrcpads,
			      "sample_rate", &src->sample_rate,
			      "total_frames", &src->num_frames,
			      "markers", &src->model,
			      NULL);

		/* FIXME: Need to watch model for changes to markers */

		g_object_get (G_OBJECT (src->model),
			      "markers", &src->markers,
			      NULL);

		src->start = 0;
		src->end = src->num_frames;
		create_srcpads (src);
		break;

	case PROP_SRC_SEND_NEW_MEDIA:
		src->send_new_media = g_value_get_boolean (value);
		src->sent_first = FALSE;
		break;

	case PROP_SRC_LOOP:
		src->loop = g_value_get_boolean (value);
		break;

	default:
		break;
	}
}

static void
get_property (GObject *object,
	      guint prop_id,
	      GValue *value,
	      GParamSpec *pspec)
{
}

static GstPadTemplate *
src_factory (void)
{
	return gst_pad_template_new (
		"src%d",
		GST_PAD_SRC,
		GST_PAD_SOMETIMES,
		gst_caps_new_simple ("audio/x-raw-float",
			"rate", GST_TYPE_INT_RANGE, 4000, 96000,
			"channels", G_TYPE_INT, 1,
			"buffer-frames", GST_TYPE_INT_RANGE, 1, G_MAXINT,
			"endianness", G_TYPE_INT, G_BYTE_ORDER,
			"width", G_TYPE_INT, 32,
			NULL));
}

static void
src_class_init (MarlinSampleElementSrcClass *klass)
{
	GObjectClass *object_class;
	GstElementClass *element_class;

	object_class = G_OBJECT_CLASS (klass);
	element_class = GST_ELEMENT_CLASS (klass);

	if (src_template == NULL) {
		src_template = src_factory ();
	}

	gst_element_class_add_pad_template (element_class, src_template);
	src_parent_class = g_type_class_ref (GST_TYPE_ELEMENT);

	object_class->dispose = src_dispose;
	object_class->set_property = set_property;
	object_class->get_property = get_property;

	element_class->change_state = marlin_sample_element_src_change_state;

	g_object_class_install_property (object_class,
					 PROP_SRC_SAMPLE,
					 g_param_spec_object ("sample",
							      "Sample",
							      "The sample to be the source",
							      MARLIN_SAMPLE_TYPE,
							      G_PARAM_READWRITE));

	g_object_class_install_property (object_class,
					 PROP_SRC_POSITION,
					 g_param_spec_uint64 ("position",
							      "Position",
							      "The position in the stream",
							      0, G_MAXUINT64, 0,
							      G_PARAM_READABLE));
	g_object_class_install_property (object_class,
					 PROP_SRC_SEND_NEW_MEDIA,
					 g_param_spec_boolean ("send-new-media",
							       NULL, NULL,
							       FALSE,
							       G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_SRC_LOOP,
					 g_param_spec_boolean ("loop",
							       NULL, NULL,
							       FALSE,
							       G_PARAM_WRITABLE));
}

static void
src_init (MarlinSampleElementSrc *src)
{
	src->sample = NULL;

	src->srcpads = NULL;
	src->block_size = 1024; /* Number of frames output each block */
	src->numsrcpads = 0;
	src->sample_rate = 4000;
	src->num_frames = 0;
	src->start = 0;
	src->end = 0;
	src->segment_loop = FALSE;
	src->need_taglist = TRUE; /* We need to send our tags first */
	src->send_new_media = FALSE;
	src->need_new_media = FALSE;
	src->sent_first = TRUE;
	src->loop = FALSE;
}

GType
marlin_sample_element_src_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		static GTypeInfo info = {
			sizeof (MarlinSampleElementSrcClass), NULL, NULL,
			(GClassInitFunc) src_class_init, NULL, NULL,
			sizeof (MarlinSampleElementSrc), 0,
			(GInstanceInitFunc) src_init
		};

		type = g_type_register_static (GST_TYPE_ELEMENT, "MarlinSampleElementSrc", &info, 0);
	}

	return type;
}
		
MarlinSampleElementSrc *
marlin_sample_element_src_new (MarlinSample *sample)
{
	MarlinSampleElementSrc *src;

	if (sample == NULL) {
		src = g_object_new (MARLIN_SAMPLE_ELEMENT_SRC_TYPE, NULL);
	} else {
		src = g_object_new (MARLIN_SAMPLE_ELEMENT_SRC_TYPE,
				    "sample", sample,
				    NULL);
	}

	gst_element_set_name (GST_ELEMENT (src), "marlin-sample-src");
	return src;
}

static SrcOutputChannel *
get_src_channel (MarlinSampleElementSrc *src,
		 guint channel)
{
	GList *p;
	int i;

	for (p = src->srcpads; p && i < channel; p = p->next) {
		;
	}

	if (p == NULL) {
		return NULL;
	} else {
		return p->data;
	}
}

gboolean
marlin_sample_element_src_seek (MarlinSampleElementSrc *src,
				GstSeekType type,
				gint64 offset)
{
	int i;
	GList *p;

	for (i = 0, p = src->srcpads; p; i++, p = p->next) {
		SrcOutputChannel *src_channel = p->data;
		GstEvent *event;

		event = gst_event_new_seek (type, (gint64) (offset) * sizeof (float));
		
		if (gst_pad_send_event (src_channel->srcpad, event) == FALSE) {
			g_warning ("Seek failed channel %d\n", i);
			return FALSE;
		}
	}

	return TRUE;
}

gboolean
marlin_sample_element_src_seek_range (MarlinSampleElementSrc *src,
				      GstSeekType type,
				      gint64 start,
				      gint64 finish)
{
	int i;
	GList *p;

	for (i = 0, p = src->srcpads; p; i++, p = p->next) {
		SrcOutputChannel *src_channel = p->data;
		GstEvent *event;

		event = gst_event_new_segment_seek (type, start, finish);
		if (gst_pad_send_event (src_channel->srcpad, event) == FALSE) {
			g_warning ("Seek range failed channel %d\n", i);
			return FALSE;
		}
	}

	return TRUE;
}

guint64
marlin_sample_element_src_get_position (MarlinSampleElementSrc *src,
					guint channel)
{
	SrcOutputChannel *src_channel;

	src_channel = get_src_channel (src, channel);
	if (src_channel == NULL) {
		return 0;
	} else {
		return src_channel->offset;
	}
}

/* This is here because it really confuses XEmacs' code indentation thing */
GST_PAD_EVENT_MASK_FUNCTION (marlin_sample_element_src_get_event_mask,
{ GST_EVENT_SEEK, GST_SEEK_METHOD_CUR |
	  GST_SEEK_METHOD_SET |
	  GST_SEEK_METHOD_END |
	  GST_SEEK_FLAG_FLUSH },
{ GST_EVENT_SEEK_SEGMENT, 0}, 
{ GST_EVENT_FLUSH, 0 },
{ GST_EVENT_SIZE, 0 }
)

GST_PAD_QUERY_TYPE_FUNCTION (marlin_sample_element_src_get_query_types,
	GST_QUERY_TOTAL,
	GST_QUERY_POSITION,
	GST_QUERY_START,
	GST_QUERY_SEGMENT_END
	)

GST_PAD_FORMATS_FUNCTION (marlin_sample_element_src_get_formats,
	GST_FORMAT_BYTES,
	GST_FORMAT_DEFAULT
	)
