/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 *  Authors: Iain Holmes <iain@prettypeople.org>
 *
 *  Copyright 2002 Iain Holmes
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of version 2 of the GNU General Public 
 *  License as published by the Free Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <gst/gst.h>

#include <gconf/gconf-client.h>

#include <marlin-save-pipeline.h>
#include <marlin-marshal.h>
#include <marlin-sample.h>
#include <marlin-sample-element.h>
#include <marlin-mt.h>
#include <marlin/marlin-gst-extras.h>

enum {
	PROP_0,
	PROP_SAMPLE,
	PROP_FILENAME,
	PROP_MIMETYPE,
	PROP_ENCODER
};

struct _MarlinSavePipelinePrivate {
	MarlinSample *sample;

	GstElement *src, *interleave, *audioconvert, *sink, *encoder;
	char *filename, *mimetype;
};

static MarlinPipeline *parent_class = NULL;

#define DEFAULT_MIMETYPE "audio/x-wav"

static void
finalize (GObject *object)
{
	MarlinSavePipeline *pipeline;

	pipeline = MARLIN_SAVE_PIPELINE (object);

	if (pipeline->priv == NULL) {
		return;
	}

	g_object_unref (G_OBJECT (pipeline->priv->sample));
	g_free (pipeline->priv->filename);
	g_free (pipeline->priv->mimetype);
	
	g_free (pipeline->priv);
	
	pipeline->priv = NULL;

	G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
ipad_unlinked (GstPad *pad,
	       GstPad *peer,
	       MarlinSavePipeline *pipeline)
{
#if 0
	/* 27-08-04: This causes problems because the only time its ever called
	   is during the shutdown phase of the pipeline, and it causes
	   pad to get destroyed too early */
	GstElement *parent;

	parent = GST_PAD_PARENT (pad);
	gst_element_remove_pad (parent, pad);
#endif
}

static void
sample_src_new_pad (GstElement *src,
		    GstPad *pad,
		    MarlinSavePipeline *pipeline)
{
	GstPad *isink;

	isink = gst_element_get_request_pad (pipeline->priv->interleave,
					     "sink%d");
	g_assert (isink != NULL);

	g_signal_connect (G_OBJECT (isink), "unlinked",
			  G_CALLBACK (ipad_unlinked), pipeline);

	if (!gst_pad_link (pad, isink)) {
		g_warning ("Caps nego in save-pipeline-new_pad failed");
		return;
	}
}

static void
connect_pipeline (MarlinSavePipeline *pipeline)
{
	guint channels, i;

	g_object_get (G_OBJECT (pipeline->priv->sample),
		      "channels", &channels,
		      NULL);

	for (i = 0; i < channels; i++) {
		GstPad *srcpad, *isink;
		char *name;

		name = g_strdup_printf ("src%d", i);
		srcpad = gst_element_get_pad (pipeline->priv->src, name);
		g_assert (srcpad != NULL);
		g_free (name);

		isink = gst_element_get_request_pad (pipeline->priv->interleave,
						     "sink%d");
		g_assert (isink != NULL);
		g_signal_connect (G_OBJECT (isink), "unlinked",
				  G_CALLBACK (ipad_unlinked), pipeline);

		if (!gst_pad_link (srcpad, isink)) {
			g_warning ("Caps nego failed in save pipeline");
			return;
		}
	}
}

#if 0
/* 27-08-04: I'm not sure this is ever needed. We always create a new pipeline
   for each sample, so it never needs to be disconnected */
static void
disconnect_pipeline (MarlinSavePipeline *pipeline)
{
	gst_element_unlink (GST_ELEMENT (pipeline->priv->src),
			    pipeline->priv->interleave);
}
#endif

static void
set_property (GObject *object,
	      guint prop_id,
	      const GValue *value,
	      GParamSpec *pspec)
{
	MarlinSavePipeline *pipeline;
	MarlinSavePipelinePrivate *priv;
	char *filename, *new_mimetype;
	
	pipeline = MARLIN_SAVE_PIPELINE (object);
	priv = pipeline->priv;

	switch (prop_id) {
	case PROP_SAMPLE:
		if (priv->sample != NULL) {
			/* Sample is connected */
			/*disconnect_pipeline (pipeline);*/
			g_object_unref (priv->sample);
		}

		priv->sample = g_value_get_object (value);
		if (priv->sample == NULL) {
			return;
		}

		g_object_ref (G_OBJECT (priv->sample));

		/* Set the sample on the src */
		g_object_set (G_OBJECT (priv->src),
			      "sample", priv->sample,
			      NULL);

/* 		connect_pipeline (pipeline); */
		break;

	case PROP_FILENAME:
		if (priv->filename != NULL) {
			g_free (priv->filename);
		}

		priv->filename = g_strdup (g_value_get_string (value));
		g_object_set (G_OBJECT (priv->sink),
			      "location", priv->filename,
			      NULL);
		break;

	case PROP_MIMETYPE: {
		GstElement *encoder;
		gboolean ret;

		new_mimetype = (char *) g_value_get_string (value);
		if (strcmp (new_mimetype, priv->mimetype) == 0) {
			/* Same mimetype - return */
			return;
		}

		if (priv->mimetype != NULL) {
			g_free (priv->mimetype);
		}

		priv->mimetype = g_strdup (new_mimetype);
		encoder = get_encoder_for_mime (priv->mimetype);
		if (encoder == NULL) {
			g_warning ("No encoder for %s\n", priv->mimetype);
			return;
		}

		if (priv->encoder != NULL) {
			gst_element_disconnect (priv->audioconvert, priv->encoder);
			gst_element_disconnect (priv->encoder, priv->sink);
			marlin_pipeline_remove (MARLIN_PIPELINE (pipeline),
						priv->encoder);
		}
		priv->encoder = encoder;
		marlin_pipeline_add (MARLIN_PIPELINE (pipeline), priv->encoder);

		ret = gst_element_connect (priv->audioconvert, priv->encoder);
		g_assert (ret);
		ret = gst_element_connect (priv->encoder, priv->sink);
		g_assert (ret);

		break;
	}
		
	case PROP_ENCODER: {
		gboolean ret;

		if (priv->encoder != NULL) {
			gst_element_disconnect (priv->audioconvert, priv->encoder);
			gst_element_disconnect (priv->encoder, priv->sink);
			marlin_pipeline_remove (MARLIN_PIPELINE (pipeline),
						priv->encoder);
		}
		priv->encoder = g_value_get_object (value);

		marlin_pipeline_add (MARLIN_PIPELINE (pipeline), priv->encoder);
		ret = gst_element_connect (priv->audioconvert, priv->encoder);
		g_assert (ret);

		ret = gst_element_connect (priv->encoder, priv->sink);
		g_assert (ret);

		break;
	}
				    
	default:
		break;
	}
}

static void
get_property (GObject *object,
	      guint prop_id,
	      GValue *value,
	      GParamSpec *pspec)
{
	MarlinSavePipeline *pipeline;
	MarlinSavePipelinePrivate *priv;

	pipeline = MARLIN_SAVE_PIPELINE (object);
	priv = pipeline->priv;

	switch (prop_id) {
	case PROP_SAMPLE:
		g_value_set_object (value, priv->sample);
		break;

	case PROP_FILENAME:
		g_value_set_string (value, priv->filename);
		break;

	case PROP_MIMETYPE:
		g_value_set_string (value, priv->mimetype);
		break;
		
	default:
		break;
	}
}

static int
get_progress (MarlinPipeline *pipeline)
{
	MarlinSavePipeline *save_pipeline;
	GstPad *src;
	guint64 position, total;
	GstFormat format = GST_FORMAT_BYTES;
	gboolean ret;
	
	save_pipeline = MARLIN_SAVE_PIPELINE (pipeline);
	src = gst_element_get_pad (save_pipeline->priv->src, "src0");

	ret = gst_pad_query (src, GST_QUERY_POSITION,
			     &format, &position);
	if (ret == FALSE) {
		return 0;
	}

	format = GST_FORMAT_BYTES;
	ret = gst_pad_query (src, GST_QUERY_TOTAL,
			     &format, &total);
	if (ret == FALSE) {
		return 0;
	}

	return (int) ((float)(position * 100) / (float) (total));
}
	
static void
class_init (GObjectClass *klass)
{
	MarlinPipelineClass *pclass;

	pclass = MARLIN_PIPELINE_CLASS (klass);
	
	klass->finalize = finalize;
	klass->set_property = set_property;
	klass->get_property = get_property;

	pclass->get_progress = get_progress;
	
	parent_class = g_type_class_peek_parent (klass);

	/* Properties */
	g_object_class_install_property (klass,
					 PROP_SAMPLE,
					 g_param_spec_object ("sample",
							      "Sample",
							      "The sample to be loaded",
							      MARLIN_SAMPLE_TYPE,
							      G_PARAM_READWRITE));

	g_object_class_install_property (klass,
					 PROP_FILENAME,
					 g_param_spec_string ("filename",
							      "Filename",
							      "",
							      "Untitled",
							      G_PARAM_READWRITE));

	g_object_class_install_property (klass,
					 PROP_MIMETYPE,
					 g_param_spec_string ("mimetype",
							      "Mimetype",
							      "Save file as this mimetype",
							      DEFAULT_MIMETYPE,
							      G_PARAM_READWRITE));
	g_object_class_install_property (klass,
					 PROP_ENCODER,
					 g_param_spec_object ("encoder",
							      "", "",
							      GST_TYPE_ELEMENT,
							      G_PARAM_WRITABLE));
}

static void
save_eos_reached (GstElement *element,
		  MarlinSavePipeline *pipeline)
{
	marlin_pipeline_set_eos (MARLIN_PIPELINE (pipeline));
}

static void
save_handoff (GstElement *element,
	      MarlinSavePipeline *pipeline)
{
	marlin_pipeline_progress_changed (MARLIN_PIPELINE (pipeline));
}

static void
init (MarlinSavePipeline *pipeline)
{
	GstElementFactory *encoder_factory;
	GstCaps *filtercaps;
	gboolean ret;

	pipeline->priv = g_new (MarlinSavePipelinePrivate, 1);

	pipeline->priv->src = GST_ELEMENT (marlin_sample_element_src_new (NULL));
	gst_element_set_name (GST_ELEMENT (pipeline->priv->src), "save-src");
	g_signal_connect (G_OBJECT (pipeline->priv->src), "new-pad",
			  G_CALLBACK (sample_src_new_pad), pipeline);
	
	pipeline->priv->interleave = gst_element_factory_make ("interleave", "save-interleave");
	pipeline->priv->audioconvert = gst_element_factory_make ("audioconvert", "save-convert");
	
	pipeline->priv->sink = gst_element_factory_make ("filesink", "save-sink");
	g_signal_connect (G_OBJECT (pipeline->priv->sink), "eos",
			  G_CALLBACK (save_eos_reached), pipeline);
	
	/* Not really sure what the handoff signal is meant for
	   but it gets emitted after every iteration */
	g_signal_connect (G_OBJECT (pipeline->priv->sink), "handoff",
			  G_CALLBACK (save_handoff), pipeline);
	
	pipeline->priv->encoder = get_encoder_for_mime (DEFAULT_MIMETYPE);

	marlin_pipeline_add (MARLIN_PIPELINE (pipeline), pipeline->priv->interleave);
	marlin_pipeline_add (MARLIN_PIPELINE (pipeline), pipeline->priv->audioconvert);
	marlin_pipeline_add (MARLIN_PIPELINE (pipeline), pipeline->priv->encoder);
	marlin_pipeline_add (MARLIN_PIPELINE (pipeline), pipeline->priv->sink);
	marlin_pipeline_add (MARLIN_PIPELINE (pipeline), pipeline->priv->src);

	/* Connect some bits together */
	filtercaps = gst_caps_new_simple ("audio/x-raw-float", NULL);
	ret = gst_element_link_filtered (pipeline->priv->interleave, pipeline->priv->audioconvert, filtercaps);
	gst_caps_free (filtercaps);
	g_assert (ret);

	ret = gst_element_link (pipeline->priv->audioconvert, pipeline->priv->encoder);
	g_assert (ret);

	ret = gst_element_link (pipeline->priv->encoder, pipeline->priv->sink);
	g_assert (ret);
	
	pipeline->priv->sample = NULL;
	pipeline->priv->filename = NULL;
	pipeline->priv->mimetype = g_strdup (DEFAULT_MIMETYPE);
}

GType
marlin_save_pipeline_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		static GTypeInfo info = {
			sizeof (MarlinSavePipelineClass), NULL, NULL,
			(GClassInitFunc) class_init, NULL, NULL,
			sizeof (MarlinSavePipeline), 0, (GInstanceInitFunc) init
		};

		type = g_type_register_static (MARLIN_PIPELINE_TYPE,
					       "MarlinSavePipeline",
					       &info, 0);
	}
	
	return type;
}

MarlinSavePipeline *
marlin_save_pipeline_new (MarlinOperation *operation,
			  MarlinSample *sample)
{
	return g_object_new (MARLIN_SAVE_PIPELINE_TYPE,
			     "operation", operation,
			     "sample", sample,
			     NULL);
}
