/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 *  Authors: Iain Holmes <iain@prettypeople.org>
 *
 *  Copyright 2002-2005 Iain Holmes
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of version 2 of the GNU General Public
 *  License as published by the Free Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <glib/gi18n.h>

#include <gtk/gtkcheckbutton.h>
#include <gtk/gtkdialog.h>
#include <gtk/gtkstock.h>
#include <gtk/gtkvbox.h>

#include <marlin/marlin-cross-fader.h>
#include <marlin/marlin-program.h>

#include "marlin-crossfade-dialog.h"

struct _xfade_dialog_data {
	GtkWindow *window;
	GtkWidget *xfade;

	MarlinSample *sample;
};


static void
xfade_response_cb (GtkDialog *dialog,
		   guint response_id,
		   struct _xfade_dialog_data *xfd)
{
	MarlinSampleFade *src_fade, *dest_fade;
	MarlinProgram *program = marlin_program_get_default ();
	MarlinSample *src, *dest;
	guint64 start_pos, length;
	MarlinUndoManager *undo;
	MarlinUndoContext *ctxt;
	MarlinRange range;
	
	switch (response_id) {
	case GTK_RESPONSE_OK:
		src = MARLIN_SAMPLE (marlin_program_get_clipboard (program));
		dest = xfd->sample;

		g_object_get (G_OBJECT (xfd->xfade),
			      "src_fade", &src_fade,
			      "dest_fade", &dest_fade,
			      "start_frame", &(range.start),
			      NULL);

		g_object_get (G_OBJECT (dest),
			      "undo-manager", &undo,
			      NULL);
		
		g_object_get (G_OBJECT (src),
			      "total_frames", &length,
			      NULL);

		ctxt = marlin_undo_manager_context_begin (undo, _("Crossfade"));
		g_print ("dest: %llu %llu %f %f\n", dest_fade->fade_start,
			 dest_fade->fade_end, dest_fade->in_level,
			 dest_fade->out_level);
		marlin_sample_fade (dest, dest_fade, NULL, ctxt, NULL);
		marlin_sample_fade (src, src_fade, NULL, ctxt, NULL);

		/* Check the clip value */
		range.finish = range.start + length;
		range.coverage = MARLIN_COVERAGE_BOTH;
		marlin_sample_mix (dest, src, 0.0, 0.0, &range, FALSE, NULL, ctxt, NULL);
		marlin_undo_manager_context_end (undo, ctxt);

		break;

	case GTK_RESPONSE_HELP:
		marlin_display_help ("marlin-xfade-dialog");
		return;

	case GTK_RESPONSE_CANCEL:
	default:
		break;
	}

	gtk_widget_destroy (GTK_WIDGET (dialog));
	g_free (xfd);
}

static void
set_snap_to_grid (GtkToggleButton *tb,
		  struct _xfade_dialog_data *xfd)
{
	g_object_set (G_OBJECT (xfd->xfade),
		      "snap_to_grid", gtk_toggle_button_get_active (tb),
		      NULL);
}

GtkWidget *
marlin_crossfade_dialog_new (MarlinSample *sample,
			     GtkWindow *window,
			     guint64 start_pos)
{
	struct _xfade_dialog_data *xfd;
	MarlinProgram *program = marlin_program_get_default ();
	MarlinSample *clipboard;
	GtkWidget *dialog, *vbox, *hbox, *snap;
	char *name, *title;
	
	clipboard = MARLIN_SAMPLE (marlin_program_get_clipboard (program));
	
	xfd = g_new (struct _xfade_dialog_data, 1);
	xfd->window = window;
	xfd->sample = sample;

	g_object_get (G_OBJECT (sample),
		      "name", &name,
		      NULL);

	dialog = gtk_dialog_new ();
	gtk_dialog_set_has_separator (GTK_DIALOG (dialog), FALSE);
	gtk_window_set_resizable (GTK_WINDOW (dialog), FALSE);
	gtk_window_set_transient_for (GTK_WINDOW (dialog), GTK_WINDOW (window));
	title = g_strdup_printf (_("Crossfade Sample: %s"), name);
	g_free (name);

	gtk_window_set_title (GTK_DIALOG (dialog), title);
	g_free (title);

	vbox = gtk_vbox_new (FALSE, 12);
	gtk_container_set_border_width (GTK_CONTAINER (vbox), 12);
	gtk_widget_show (vbox);
	gtk_container_add (GTK_CONTAINER (GTK_DIALOG (dialog)->vbox), vbox);

	xfd->xfade = g_object_new (MARLIN_CROSS_FADER_TYPE,
				   "start_frame", start_pos,
				   "dest", sample,
				   "src", clipboard,
				   NULL);
	gtk_widget_show (xfd->xfade);
	gtk_box_pack_start (GTK_BOX (vbox), xfd->xfade, FALSE, FALSE, 0);

	snap = gtk_check_button_new_with_mnemonic (_("_Snap to grid"));
	gtk_box_pack_start (GTK_BOX (vbox), snap, FALSE, FALSE, 0);
	gtk_widget_show (snap);

	g_signal_connect (snap, "toggled",
			  G_CALLBACK (set_snap_to_grid), xfd);

	gtk_dialog_add_button (GTK_DIALOG (dialog),
			       GTK_STOCK_CANCEL,
			       GTK_RESPONSE_CANCEL);
	gtk_dialog_add_button (GTK_DIALOG (dialog),
			       _("Crossfade Samples"),
			       GTK_RESPONSE_OK);
	gtk_dialog_add_button (GTK_DIALOG (dialog),
			       GTK_STOCK_HELP,
			       GTK_RESPONSE_HELP);
	gtk_dialog_set_default_response (GTK_DIALOG (dialog), GTK_RESPONSE_OK);

	g_signal_connect (dialog, "response",
			  G_CALLBACK (xfade_response_cb), xfd);
	
	gtk_widget_show (dialog);

	return dialog;
}
