/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 *  Authors: Iain Holmes <iain@prettypeople.org>
 *
 *  Copyright 2002 Iain Holmes
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of version 2 of the GNU General Public 
 *  License as published by the Free Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <gnome.h>

#include <gconf/gconf-client.h>

#include <gst/gst.h>

#ifdef HAVE_MEDIA_PROFILES
#include <profiles/gnome-media-profiles.h>
#endif

#include <marlin/marlin-sample.h>
#include <marlin/marlin-program.h>

#include <ephy/ephy-toolbars-model.h>

#include "main.h"
#include "marlin-window.h"
#include "marlin-recent.h"
#include "marlin-session.h"
#include "marlin-stock.h"

#include "bacon/bacon-message-connection.h"

static EphyToolbarsModel *tb_model = NULL;
static GList *marlin_windows = NULL; /* List of MarlinWindow */
static BaconMessageConnection *conn;

static struct poptOption marlin_options[] = {
	{ NULL, '\0', POPT_ARG_INCLUDE_TABLE, NULL, 0, "GStreamer", NULL },
	POPT_TABLEEND
};

static gboolean
marlin_shutdown (gpointer data)
{
	marlin_program_destroy_default ();
	gtk_main_quit ();

	return FALSE;
}

static void
window_destroyed (GtkWidget *widget,
		  gpointer data)
{
	marlin_windows = g_list_remove (marlin_windows, widget);

	/* No windows left */
	if (marlin_windows == NULL) {
		g_idle_add ((GSourceFunc) marlin_shutdown, NULL);
	}
}

static gboolean
window_deleted (GtkWidget *window,
		GdkEventAny *event,
		gpointer data)
{
	return marlin_window_can_close (MARLIN_WINDOW (window));
}

GList *
marlin_get_window_list (void)
{
	return marlin_windows;
}

void
marlin_foreach_window (MarlinForeachFunction func,
		       gpointer closure)
{
	GList *p;

	for (p = marlin_windows; p; p = p->next) {
		func (p->data, closure);
	}
}

void
marlin_quit (void)
{
	GList *p;

	for (p = marlin_windows; p;) {
		MarlinWindow *window = p->data;

		/* p is set here instead of in the loop
		   because by the time we get back to the loop
		   p will be invalid. */
		p = p->next;
		marlin_window_close (window);
	}
}

static char *
make_window_name (void)
{
	static int sample_count = 1;
	char *name;
	
	if (sample_count != 1) {
		name = g_strdup_printf ("Untitled-%d", sample_count);
	} else {
		name = g_strdup ("Untitled");
	}

	sample_count++;
	return name;
}

GtkWidget *
marlin_open_window_with_sample (MarlinSample *sample)
{
	MarlinWindow *window;
	char *name;

	name = make_window_name ();

	g_object_set (G_OBJECT (sample),
		      "filename", name,
		      NULL);
	g_free (name);

	window = marlin_window_new (sample);
	
	g_object_unref (sample);
	
	g_signal_connect (G_OBJECT (window), "delete-event",
			  G_CALLBACK (window_deleted), NULL);
	g_signal_connect (G_OBJECT (window), "destroy",
			  G_CALLBACK (window_destroyed), NULL);

	marlin_windows = g_list_append (marlin_windows, window);
	gtk_widget_show (GTK_WIDGET (window));

	return GTK_WIDGET (window);
}

GtkWidget *
marlin_open_window (const char *filename,
		    const char *uuid)
{
	GConfClient *client = gconf_client_get_default ();
	MarlinWindow *window;
	MarlinSample *sample;
	int rate, channels;
	
	/* Create a sample to be used in the window */
	sample = marlin_sample_new ();

	/* Set the defaults */
	rate = gconf_client_get_int (client, "/apps/marlin/system-state/new-sample-rate", NULL);
	channels = gconf_client_get_int (client, "/apps/marlin/system-state/new-sample-channels", NULL);
	g_object_unref (G_OBJECT (client));

	/* If there's something wrong (like the user hasn't installed
	   marlin correctly, then these can be dumb values */
	if (channels == 0) {
		channels = 2;
	}

	if (rate == 0) {
		rate = 48000;
	}

	g_object_set (G_OBJECT (sample),
		      "sample-rate", rate,
		      "channels", channels,
		      NULL);

	if (filename == NULL) {
		char *name = make_window_name ();
		
		g_object_set (G_OBJECT (sample),
			      "filename", name,
			      NULL);

		g_free (name);
	}

	if (uuid != NULL) {
		window = marlin_window_new_from_state (sample, uuid);
	} else {
		window = marlin_window_new (sample);
	}
	g_object_unref (sample);
	
	g_signal_connect (G_OBJECT (window), "delete-event",
			  G_CALLBACK (window_deleted), NULL);
	g_signal_connect (G_OBJECT (window), "destroy",
			  G_CALLBACK (window_destroyed), NULL);

	if (filename != NULL) {
		EggRecentModel *model;
		EggRecentItem *item;
		char *uri;
		
		marlin_window_load_file (window, filename, NULL);

		/* Add the filename to the model */
		model = marlin_recent_get_model ();
		uri = g_filename_to_utf8 (filename, -1, NULL, NULL, NULL);

		item = egg_recent_item_new_from_uri (uri);
		egg_recent_item_add_group (item, "Marlin");
		egg_recent_model_add_full (model, item);
		egg_recent_item_unref (item);
		g_free (uri);
	}
	
	marlin_windows = g_list_prepend (marlin_windows, window);
	gtk_widget_show (GTK_WIDGET (window));

	return (GtkWidget *) window;
}

static void
update_toolbar_remove_flag (EphyToolbarsModel *model,
			    gpointer data)
{
	int i, n_toolbars;
	int not_removable = EGG_TB_MODEL_ICONS_ONLY;

	n_toolbars = egg_toolbars_model_n_toolbars (EGG_TOOLBARS_MODEL (model));
	if (n_toolbars <= 1) {
		not_removable &= EGG_TB_MODEL_NOT_REMOVABLE;
	}
	
	for (i = 0; i < n_toolbars - 1; i++) {
		egg_toolbars_model_set_flags (EGG_TOOLBARS_MODEL (model), not_removable, i);
	}
}

EphyToolbarsModel *
marlin_get_toolbars_model (void)
{
	if (tb_model == NULL) {
		tb_model = ephy_toolbars_model_new ();
		
		ephy_toolbars_model_load (tb_model);
		
		g_signal_connect (G_OBJECT (tb_model), "toolbar_added",
				  G_CALLBACK (update_toolbar_remove_flag),
				  NULL);
		g_signal_connect (G_OBJECT (tb_model), "toolbar_removed",
				  G_CALLBACK (update_toolbar_remove_flag),
				  NULL);
	}

	return tb_model;
}

/* From Totem */
static char *
make_full_path (const char *path)
{
	char *retval, *curdir, *curdir_withslash, *escaped;
	
	g_return_val_if_fail (path != NULL, NULL);

	if (strstr (path, "://") != NULL) {
		return g_strdup (path);
	}
	
	if (path[0] == '/') {
		escaped = gnome_vfs_escape_path_string (path);

		retval = g_strdup_printf ("%s", escaped);
		g_free (escaped);
		return retval;
	}

	curdir = g_get_current_dir ();
	escaped = gnome_vfs_escape_path_string (curdir);
	curdir_withslash = g_strdup_printf ("%s%s",
					    escaped, G_DIR_SEPARATOR_S);
	g_free (escaped);
	g_free (curdir);

	escaped = gnome_vfs_escape_path_string (path);
	retval = gnome_vfs_uri_make_full_from_relative (curdir_withslash, escaped);
	g_free (curdir_withslash);
	g_free (escaped);

	return retval;
}

static void
process_command_line (GnomeProgram *program,
		      gboolean use_remote)
{
	poptContext pctx;
	GValue value = {0,};
	char **args = NULL;

	/* Get the args from popt */
	g_value_init (&value, G_TYPE_POINTER);
	g_object_get_property (G_OBJECT (program),
			       GNOME_PARAM_POPT_CONTEXT, &value);
	pctx = g_value_get_pointer (&value);
	g_value_unset (&value);
	
	args = (char **) poptGetArgs (pctx);
	if (args == NULL) {
		if (use_remote) {
			bacon_message_connection_send (conn, "marlin:OpenWindow");
		} else {
			marlin_open_window (NULL, NULL);
		}
	} else {
		int i;
		/* Open a new window for each file */
		for (i = 0; args[i]; i++) {
			if (use_remote) {
				char *line, *fullpath;

				fullpath = make_full_path (args[i]);
				line = g_strdup_printf ("marlin:OpenFile %s",
							fullpath);
				g_free (fullpath);
				bacon_message_connection_send (conn, line);
				g_free (line);
			} else {
				char *fullpath;

				fullpath = make_full_path (args[i]);
				marlin_open_window (fullpath, NULL);
				g_free (fullpath);
			}
		}
	}
	poptFreeContext (pctx);
}

static void
message_received_cb (const char *msg,
		     gpointer data)
{
	char *command_str, *url;

	if (strlen (msg) < 7 || strncmp (msg, "marlin:", 7) != 0) {
		g_warning ("Malformed message: %s", msg);
		return;
	}

	msg += 7; /* Skip past "marlin:" */

	if (strncmp (msg, "OpenWindow", 10) == 0) {
		marlin_open_window (NULL, NULL);
	} else if (strncmp (msg, "OpenFile", 8) == 0) {
		marlin_open_window (msg + 9, NULL);
	}
}

int
main (int argc,
      char **argv)
{
	GnomeProgram *program;
#ifdef HAVE_MEDIA_PROFILES
	GConfClient *client;
#endif

	/* Init gettext */
	bindtextdomain (GETTEXT_PACKAGE, GNOMELOCALEDIR);
	bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
	textdomain (GETTEXT_PACKAGE);

	marlin_ensure_dir_exists (marlin_dot_dir ());
	
	/* Init GNOME and GStreamer */
	marlin_options[0].arg = (void *) gst_init_get_popt_table ();
	program = gnome_program_init ("marlin", VERSION,
				      LIBGNOMEUI_MODULE,
				      argc, argv,
				      GNOME_PARAM_POPT_TABLE, marlin_options,
				      GNOME_PARAM_HUMAN_READABLE_NAME,
				      _("Marlin Sample Editor"),
				      GNOME_PARAM_APP_DATADIR, DATADIR,
				      NULL);

	g_set_application_name (_("Marlin Sample Editor"));
	marlin_stock_icons_register ();

	/* Do this here so we can throw up a dialog */
	if (marlin_check_tmp_dir () == FALSE) {
		exit (1);
	}

#ifdef HAVE_MEDIA_PROFILES
	/* Init media profiles */
	client = gconf_client_get_default ();
	gnome_media_profiles_init (client);
	g_object_unref (G_OBJECT (client));
#endif

	/* IPC stuff */
	conn = bacon_message_connection_new ("marlin");
	if (bacon_message_connection_get_is_server (conn) == FALSE) {
		process_command_line (program, TRUE);
		gdk_notify_startup_complete ();
		exit (0);
	}

	/* Init Marlin */
	marlin_mt_initialise ();

	marlin_recent_initialise ();
	marlin_session_initialise (argv[0]);

	if (marlin_session_is_restored ()) {
		marlin_session_load ();
	} else {
		process_command_line (program, FALSE);
	}

	if (bacon_message_connection_get_is_server (conn)) {
		bacon_message_connection_set_callback (conn, (BaconMessageReceivedFunc) message_received_cb, NULL);
	}

	gtk_main ();
}
