/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 *  Authors: Iain Holmes <iain@prettypeople.org>
 *
 *  Copyright 2002-2005 Iain Holmes
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of version 2 of the GNU General Public 
 *  License as published by the Free Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *
 */

/* FIXME: Once we have a plugin set up, this is a prime candidate */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <glib/gi18n.h>

#include <gtk/gtk.h>

#include <gst/gst.h>

#include <marlin/marlin-sample.h>
#include <marlin/marlin-operation.h>
#include <marlin/marlin-sample-element.h>
#include <marlin/marlin-pipeline.h>
#include <marlin/marlin-position-spinner.h>
#include <marlin/marlin-utils.h>
#include <marlin/marlin-x-utils.h>

/* A pipeline class for generating sine source */
#define SINE_PIPELINE_TYPE (sine_pipeline_get_type ())
#define SINE_PIPELINE(obj) (G_TYPE_CHECK_INSTANCE_CAST ((obj), SINE_PIPELINE_TYPE, SinePipeline))
#define SINE_PIPELINE_CLASS(klass) (G_TYPE_CHECK_CLASS_CAST ((klass), SINE_PIPELINE_TYPE, SinePipelineClass))

typedef struct _SinePipeline SinePipeline;
typedef struct _SinePipelineClass SinePipelineClass;

struct _SinePipeline {
	MarlinPipeline pipeline;

	MarlinSample *sample;

	GstElement *sine[2], *queue[2], *sink;

	guint64 required, done;
};

struct _SinePipelineClass {
	MarlinPipelineClass pipeline_class;
};

static GObjectClass *parent_class = NULL;

enum {
	PROP_0,
	PROP_SAMPLE,
	PROP_LENGTH,
	PROP_FREQ,
};

static GType sine_pipeline_get_type (void);

static void
finalize (GObject *object)
{
	SinePipeline *pipeline;

	pipeline = SINE_PIPELINE (object);

	parent_class->finalize (object);
}

static void
disconnect_pipeline (SinePipeline *sine)
{
	guint channels, i;

	g_object_get (G_OBJECT (sine->sample),
		      "channels", &channels,
		      NULL);

	for (i = 0; i < channels; i++) {
		gst_element_unlink (sine->sine[i], sine->queue[i]);
		gst_element_unlink (sine->queue[i], sine->sink);

		marlin_pipeline_remove (MARLIN_PIPELINE (sine), sine->sine[i]);
		marlin_pipeline_remove (MARLIN_PIPELINE (sine), sine->queue[i]);
	}
}

static void
connect_pipeline (SinePipeline *sine)
{
	guint channels, i;

	g_object_get (G_OBJECT (sine->sample),
		      "channels", &channels,
		      NULL);

	for (i = 0; i < channels; i++) {
		char name[6];
		GstPad *sink, *qsrc, *qsink, *src;
		char *qname;

		g_snprintf (name, 6, "sine%d", i);
		sine->sine[i] = gst_element_factory_make ("sinesrc", name);
		marlin_pipeline_add (MARLIN_PIPELINE (sine), sine->sine[i]);
		src = gst_element_get_pad (sine->sine[i], "src");
		g_assert (src != NULL);

		qname = g_strdup_printf ("queue.%s", name);
		sine->queue[i] = gst_element_factory_make ("queue", qname);
		g_free (qname);
		marlin_pipeline_add (MARLIN_PIPELINE (sine), sine->queue[i]);

		qsrc = gst_element_get_pad (sine->queue[i], "src");
		qsink = gst_element_get_pad (sine->queue[i], "sink");

		if (!gst_pad_link (src, qsink)) {
			g_assert_not_reached ();
		}

		sink = gst_element_get_request_pad (sine->sink, "sink%d");
		if (!gst_pad_link (qsrc, sink)) {
			g_assert_not_reached ();
		}
	}
}

static void
set_property (GObject *object,
	      guint prop_id,
	      const GValue *value,
	      GParamSpec *pspec)
{
	SinePipeline *sine = SINE_PIPELINE (object);

	switch (prop_id) {
	case PROP_SAMPLE:
		if (sine->sample != NULL) {
			disconnect_pipeline (sine);
			g_object_unref (G_OBJECT (sine->sample));
		}

		sine->sample = g_value_get_object (value);
		g_object_ref (G_OBJECT (sine->sample));

		connect_pipeline (sine);
		
		break;

	case PROP_LENGTH:
		sine->required = g_value_get_uint64 (value);
		sine->done = 0; /* Reset value */
		break;

	default:
		break;
	}
}

static void
get_property (GObject *object,
	      guint prop_id,
	      GValue *value,
	      GParamSpec *pspec)
{
	
}

static int
get_progress (MarlinPipeline *pipeline)
{
}

static void
class_init (SinePipelineClass *klass)
{
	GObjectClass *oclass;
	MarlinPipelineClass *pclass;

	oclass = G_OBJECT_CLASS (klass);
	pclass = MARLIN_PIPELINE_CLASS (klass);

	oclass->finalize = finalize;
	oclass->set_property = set_property;
	oclass->get_property = get_property;

	pclass->get_progress = get_progress;

	parent_class = g_type_class_peek_parent (klass);

	/* Properties */
	g_object_class_install_property (oclass,
					 PROP_SAMPLE,
					 g_param_spec_object ("sample",
							      "", "",
							      MARLIN_SAMPLE_TYPE,
							      G_PARAM_READABLE));
	g_object_class_install_property (oclass,
					 PROP_LENGTH,
					 g_param_spec_uint64 ("frames",
							      "", "",
							      0, G_MAXUINT64,
							      0, 
							      G_PARAM_READABLE));
	g_object_class_install_property (oclass,
					 PROP_FREQ,
					 g_param_spec_uint ("frequency",
							    "", "",
							    0, 20000,
							    440,
							    G_PARAM_READABLE));
}

static void
init (SinePipeline *pipeline)
{
	GstElement *sink_bin;

	sink_bin = gst_thread_new ("sine-sink-thread");
	pipeline->sink = GST_ELEMENT (marlin_sample_element_sink_new (NULL));
	gst_bin_add (GST_BIN (sink_bin), pipeline->sink);

	marlin_pipeline_add (MARLIN_PIPELINE (pipeline), sink_bin);
}

static GType
sine_pipeline_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		static GTypeInfo info = {
			sizeof (SinePipelineClass), NULL, NULL,
			(GClassInitFunc) class_init, NULL, NULL,
			sizeof (SinePipeline), 0, (GInstanceInitFunc) init
		};

		type = g_type_register_static (MARLIN_PIPELINE_TYPE,
					       "SinePipeline",
					       &info, 0);
	}

	return type;
}

/* *************** */

struct _GenerateTone {
	GtkWidget *dialog, *length, *freq;
	GtkAdjustment *adj;

	SinePipeline *pipeline;
	MarlinSample *sample;
};

static void
generate_response_cb (GtkDialog *dialog,
		      guint response_id,
		      struct _GenerateTone *gt)
{
	int rate, channels, freq;
	guint64 frames;

	switch (response_id) {
	case GTK_RESPONSE_OK:
		gt->sample = marlin_sample_new ();
		
		rate = marlin_gconf_get_int ("/apps/marlin/system-state/new-sample-rate");
		channels = marlin_gconf_get_int ("/apps/marlin/system-state/new-sample-channels");
		
		if (channels == 0) {
			channels = 2;
		}
		
		if (rate == 0) {
			rate = 48000;
		}
		
		g_object_set (G_OBJECT (gt->sample),
			      "sample-rate", rate,
			      "channels", channels,
			      NULL);

		frames = (guint64) GTK_SPIN_BUTTON (gt->length)->adjustment->value;
		freq = atoi (gtk_entry_get_text (GTK_ENTRY (gt->freq)));
		gt->pipeline = g_object_new (SINE_PIPELINE_TYPE,
					     "sample", gt->sample,
					     "length", frames,
					     "frequency", 440,
					     NULL);

		break;

	case GTK_RESPONSE_HELP:
		marlin_display_help ("marlin-generate-tone-dialog");
		return;

	case GTK_RESPONSE_CANCEL:
	default:
		break;
	}

	gtk_widget_destroy (dialog);
	g_free (gt);
}

GtkWidget *
marlin_generate_tone_new (GtkWindow *parent)
{
	struct _GenerateTone *gt;
	GtkWidget *vbox, *table, *label;
	int display, rate;

	gt = g_new (struct _GenerateTone, 1);
	
	gt->dialog = gtk_dialog_new ();
	gtk_dialog_set_has_separator (GTK_DIALOG (gt->dialog), FALSE);
	gtk_window_set_resizable (GTK_WINDOW (gt->dialog), FALSE);
	gtk_window_set_transient_for (GTK_WINDOW (gt->dialog), 
				      GTK_WINDOW (parent));
	gtk_window_set_title (GTK_DIALOG (gt->dialog), _("Generate Tone"));

	vbox = gtk_vbox_new (FALSE, 12);
	gtk_container_set_border_width (GTK_CONTAINER (vbox), 12);
	gtk_widget_show (vbox);
	gtk_container_add (GTK_CONTAINER (GTK_DIALOG (gt->dialog)->vbox), vbox);

	table = marlin_make_table (2, 3, FALSE);
	gtk_container_set_border_width (GTK_CONTAINER (table), 6);
	gtk_box_pack_start (GTK_BOX (vbox), table, TRUE, TRUE, 0);
	gtk_widget_show (table);

	label = marlin_make_title_label (_("Frequency:"));
	gtk_misc_set_alignment (GTK_MISC (label), 1.0, 0.5);
	PACK (table, label, 0, 0, GTK_FILL);
	gtk_widget_show (label);

	/* FIXME: I'd like a sort of piano keyboard type popup widget thingy
	   for this. Maybe after my exams */
	gt->freq = gtk_entry_new ();
	gtk_table_attach (GTK_TABLE (table), gt->freq,
			  1, 3, 0, 1,
			  GTK_FILL | GTK_EXPAND,
			  GTK_FILL,
			  0, 0);
	gtk_widget_show (gt->freq);

	label = marlin_make_title_label (_("Length:"));
	gtk_misc_set_alignment (GTK_MISC (label), 1.0, 0.5);
	PACK (table, label, 0, 1, GTK_FILL);
	gtk_widget_show (label);

	gt->length = marlin_position_spinner_new ();
	display = marlin_gconf_get_int ("/apps/marlin/system-state/position-display");
	rate = marlin_gconf_get_int ("/apps/marlin/system-state/new-sample-rate");
	g_object_set (G_OBJECT (gt->length),
 		      "max_frames", (guint64) (rate * 120),
		      "display_as", (MarlinDisplay) display,
		      "rate", rate,
		      NULL);
	
	gtk_entry_set_activates_default (GTK_ENTRY (gt->length), TRUE);
	PACK (table, gt->length, 1, 1, GTK_FILL | GTK_EXPAND);
	gtk_widget_show (gt->length);

	label = marlin_position_spinner_label (MARLIN_POSITION_SPINNER (gt->length));
	PACK (table, label, 2, 1, GTK_FILL);
	gtk_widget_show (label);

	gtk_dialog_add_button (GTK_DIALOG (gt->dialog),
			       GTK_STOCK_CANCEL,
			       GTK_RESPONSE_CANCEL);
	gtk_dialog_add_button (GTK_DIALOG (gt->dialog),
			       _("Generate Tone"),
			       GTK_RESPONSE_OK);
	gtk_dialog_add_button (GTK_DIALOG (gt->dialog),
			       GTK_STOCK_HELP,
			       GTK_RESPONSE_HELP);
	gtk_dialog_set_default_response (GTK_DIALOG (gt->dialog), GTK_RESPONSE_OK);

	g_signal_connect (gt->dialog, "response",
			  G_CALLBACK (generate_response_cb), gt);

	gtk_widget_show (gt->dialog);

	return gt->dialog;
}
