/*
 *  Copyright (C) 2003 Marco Pesenti Gritti
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  $Id: marlin-open-history-action.c,v 1.3 2003/11/29 01:47:52 iain Exp $
 */

#include <libgnomevfs/gnome-vfs-mime-utils.h>

#include <libegg/egg-recent-model.h>
#include <libegg/egg-recent-item.h>

#include <ephy/ephy-arrow-toolbutton.h>
#include <ephy/ephy-debug.h>

#include <gtk/gtkimage.h>
#include <gtk/gtkimagemenuitem.h>

#include "marlin-recent.h"
#include "marlin-window.h"
#include "marlin-open-history-action.h"

static void marlin_open_history_action_init       (MarlinOpenHistoryAction *action);
static void marlin_open_history_action_class_init (MarlinOpenHistoryActionClass *class);

static GObjectClass *parent_class = NULL;

#define MARLIN_OPEN_HISTORY_ACTION_GET_PRIVATE(object)(G_TYPE_INSTANCE_GET_PRIVATE ((object), MARLIN_OPEN_HISTORY_ACTION_TYPE, MarlinOpenHistoryActionPrivate))

struct MarlinOpenHistoryActionPrivate
{
	MarlinWindow *window;
};

enum
{
	PROP_0,
	PROP_WINDOW
};

GType
marlin_open_history_action_get_type (void)
{
	static GType type = 0;

	if (!type)
	{
		static const GTypeInfo type_info =
		{
			sizeof (MarlinOpenHistoryActionClass),
			(GBaseInitFunc) NULL,
			(GBaseFinalizeFunc) NULL,
			(GClassInitFunc) marlin_open_history_action_class_init,
			(GClassFinalizeFunc) NULL,
			NULL,
			sizeof (MarlinOpenHistoryAction),
			0, /* n_preallocs */
			(GInstanceInitFunc) marlin_open_history_action_init,
		};

		type = g_type_register_static (GTK_TYPE_ACTION,
					       "MarlinOpenHistoryAction",
					       &type_info, 0);
	}
	return type;
}

#define MAX_LENGTH 10

static void
activate_history_item (GtkWidget *menuitem,
		       EggRecentItem *item)
{
	char *uri;
	char *temp;
	
	uri = egg_recent_item_get_uri (item);
	
	temp = gnome_vfs_unescape_string (uri, "");
	
	if (temp != NULL) {
		char *local;
		
		local = gnome_vfs_get_local_path_from_uri (temp);
		
		marlin_open_window (local, NULL);
		
		g_free (local);
		g_free (temp);
	}

	g_free (uri);
}

static void
setup_history_menu (MarlinWindow *window,
		    GtkMenuShell *ms)
{
	int pos, count;
	int start, end;
	GtkWidget *item;
	EggRecentModel *model;
	GList *li, *p;

	model = marlin_recent_get_model ();
	li = egg_recent_model_get_list (model);

	for (p = li; p; p = p->next) {
		EggRecentItem *recent = p->data;
		char *title;

		title = egg_recent_item_get_uri_for_display (recent);

		item = gtk_menu_item_new_with_label (title);
		g_free (title);
		
		g_signal_connect (item, "activate",
				  G_CALLBACK (activate_history_item), recent);

		gtk_menu_shell_append (ms, item);
		gtk_widget_show (item);
	}
}

static void
menu_activated_cb (EphyArrowToolButton *w,
		   MarlinOpenHistoryAction *b)
{
	MarlinOpenHistoryActionPrivate *p = b->priv;
	GtkMenuShell *ms = ephy_arrow_toolbutton_get_menu (w);
	MarlinWindow *win = b->priv->window;
	GList *children;
	GList *li;

	LOG ("Show navigation menu")

	children = gtk_container_get_children (GTK_CONTAINER (ms));
	for (li = children; li; li = li->next) {
		gtk_container_remove (GTK_CONTAINER (ms), li->data);
	}
	g_list_free (children);

	setup_history_menu (win, ms);
}

static void
connect_proxy (GtkAction *action, GtkWidget *proxy)
{
	LOG ("Connect navigation action proxy")

	g_signal_connect (proxy, "menu-activated",
			  G_CALLBACK (menu_activated_cb), action);

	(* GTK_ACTION_CLASS (parent_class)->connect_proxy) (action, proxy);
}

static void
marlin_open_history_action_set_property (GObject *object,
					 guint prop_id,
					 const GValue *value,
					 GParamSpec *pspec)
{
	MarlinOpenHistoryAction *nav;

	nav = MARLIN_OPEN_HISTORY_ACTION (object);

	switch (prop_id) {
	case PROP_WINDOW:
		nav->priv->window = MARLIN_WINDOW (g_value_get_object (value));
		break;
	}
}

static void
marlin_open_history_action_get_property (GObject *object,
					 guint prop_id,
					 GValue *value,
					 GParamSpec *pspec)
{
	MarlinOpenHistoryAction *nav;

	nav = MARLIN_OPEN_HISTORY_ACTION (object);

	switch (prop_id) {
	case PROP_WINDOW:
		g_value_set_object (value, nav->priv->window);
		break;
	}
}

static void
marlin_open_history_action_class_init (MarlinOpenHistoryActionClass *class)
{
	GObjectClass *object_class = G_OBJECT_CLASS (class);
	GtkActionClass *action_class = GTK_ACTION_CLASS (class);
	
	object_class->set_property = marlin_open_history_action_set_property;
	object_class->get_property = marlin_open_history_action_get_property;
	
	parent_class = g_type_class_peek_parent (class);
	
	action_class->toolbar_item_type = EPHY_TYPE_ARROW_TOOLBUTTON;
	action_class->connect_proxy = connect_proxy;

	g_object_class_install_property (object_class,
                                         PROP_WINDOW,
                                         g_param_spec_object ("window",
                                                              "Window",
                                                              "The navigation window",
                                                              G_TYPE_OBJECT,
                                                              G_PARAM_READWRITE));

	g_type_class_add_private (object_class, sizeof(MarlinOpenHistoryActionPrivate));
}

static void
marlin_open_history_action_init (MarlinOpenHistoryAction *action)
{
        action->priv = MARLIN_OPEN_HISTORY_ACTION_GET_PRIVATE (action);
}


