/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 *  Authors: Iain Holmes <iain@prettypeople.org>
 *
 *  Copyright 2003 Iain Holmes
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of version 2 of the GNU General Public 
 *  License as published by the Free Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *
 */

#include <glib/gi18n.h>

#include <gtk/gtkimagemenuitem.h>

#include "marlin-recent-action.h"

struct _MarlinRecentActionPrivate {
	char *location;
	char *icon;
};

enum {
	PROP_0,
	PROP_LOCATION,
	PROP_ICON
};

enum {
	OPEN_RECENT,
	LAST_SIGNAL
};

static GObjectClass *parent_class = NULL;
static guint signals[LAST_SIGNAL] = {0};

static void
finalize (GObject *object)
{
	MarlinRecentAction *mra;

	mra = MARLIN_RECENT_ACTION (object);

	g_free (mra->priv->location);
	g_free (mra->priv->icon);

	g_free (mra->priv);

	parent_class->finalize (object);
}

static void
set_property (GObject *object,
	      guint prop_id,
	      const GValue *value,
	      GParamSpec *pspec)
{
	MarlinRecentAction *mra = MARLIN_RECENT_ACTION (object);
	
	switch (prop_id) {
	case PROP_LOCATION:
		g_free (mra->priv->location);
		mra->priv->location = g_strdup (g_value_get_string (value));
		break;

	case PROP_ICON:
		g_free (mra->priv->icon);
		mra->priv->icon = g_strdup (g_value_get_string (value));
		break;

	default:
		break;
	}
}

static void
get_property (GObject *object,
	      guint prop_id,
	      GValue *value,
	      GParamSpec *pspec)
{
	MarlinRecentAction *mra = MARLIN_RECENT_ACTION (object);

	switch (prop_id) {
	case PROP_LOCATION:
		g_value_set_string (value, mra->priv->location);
		break;

	case PROP_ICON:
		g_value_set_string (value, mra->priv->icon);
		break;

	default:
		break;
	}
}

static void
activate_cb (GtkWidget *widget,
	     const char *text,
	     GtkAction *action)
{
	MarlinRecentAction *raction = MARLIN_RECENT_ACTION (action);
	
	g_signal_emit (action, signals[OPEN_RECENT], 0,
		       raction->priv->location);
}
	
static void
menu_activate_cb (GtkWidget *widget,
		  GtkAction *action)
{
	activate_cb (widget, NULL, action);
}

static void
connect_proxy (GtkAction *action,
	       GtkWidget *proxy)
{
	(* GTK_ACTION_CLASS (parent_class)->connect_proxy) (action, proxy);

	if (GTK_IS_MENU_ITEM (proxy)) {
		g_signal_connect (proxy, "activate",
				  G_CALLBACK (menu_activate_cb), action);
	}
}

static void
class_init (MarlinRecentAction *klass)
{
	GtkActionClass *action_class;
	GObjectClass *object_class = G_OBJECT_CLASS (klass);

	parent_class = g_type_class_peek_parent (klass);
	action_class = GTK_ACTION_CLASS (klass);

	action_class->menu_item_type = GTK_TYPE_IMAGE_MENU_ITEM;
	action_class->connect_proxy = connect_proxy;

	object_class->finalize = finalize;
	object_class->set_property = set_property;
	object_class->get_property = get_property;

	signals[OPEN_RECENT] = g_signal_new ("open-recent",
					     G_OBJECT_CLASS_TYPE (object_class),
					     G_SIGNAL_RUN_FIRST,
					     G_STRUCT_OFFSET (MarlinRecentActionClass, open_recent),
					     NULL, NULL,
					     g_cclosure_marshal_VOID__STRING,
					     G_TYPE_NONE,
					     1,
					     G_TYPE_STRING);

	g_object_class_install_property (object_class,
					 PROP_LOCATION,
					 g_param_spec_string ("location",
							      "", "",
							      NULL,
							      G_PARAM_READWRITE));
	g_object_class_install_property (object_class,
					 PROP_ICON,
					 g_param_spec_string ("icon", "", "",
							      NULL,
							      G_PARAM_READWRITE));
}

static void
init (MarlinRecentAction *mra)
{
	mra->priv = g_new0 (MarlinRecentActionPrivate, 1);
}

GType
marlin_recent_action_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		static const GTypeInfo type_info = {
			sizeof (MarlinRecentActionClass), NULL, NULL,
			(GClassInitFunc) class_init, NULL, NULL,
			sizeof (MarlinRecentAction), 0, (GInstanceInitFunc) init };

		type = g_type_register_static (GTK_TYPE_ACTION,
					       "MarlinRecentAction",
					       &type_info, 0);
	}

	return type;
}

GtkAction *
marlin_recent_action_new (const char *name,
			  const char *location,
			  const char *label)
{
	GtkAction *action;

	action = g_object_new (MARLIN_TYPE_RECENT_ACTION,
			       "name", name,
			       "label", label,
			       "location", location,
			       NULL);

	return action;
}
