/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/**
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *   James Willcox <jwillcox@cs.indiana.edu>
 *   Iain Holmes <iain@prettypeople.org>
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <glib/gi18n.h>

#include <gtk/gtk.h>

#include <gconf/gconf-client.h>

#include "egg-recent-model.h"
#include "egg-recent-view.h"
#include "egg-recent-util.h"
#include "egg-recent-item.h"

#include "marlin-recent-action.h"
#include "marlin-recent-view.h"

struct _MarlinRecentView {
	GObject parent_instance;	/* We emit signals */

	GtkUIManager *merge;
	GtkActionGroup *action_group;
	guint ui_id;

	guint changed_cb_id;

	EggRecentModel *model;
/* 	GConfClient *client; */
};


struct _MarlinRecentViewMenuData {
	MarlinRecentView *view;
	EggRecentItem *item;
};

typedef struct _MarlinRecentViewMenuData MarlinRecentViewMenuData;

enum {
	ACTIVATE,
	LAST_SIGNAL
};

/* GObject properties */
enum {
	PROP_BOGUS,
	PROP_MENU_MERGE,
};

static guint marlin_recent_view_signals[LAST_SIGNAL] = { 0 };

static void
marlin_recent_view_clear (MarlinRecentView *view)
{
	GtkUIManager *merge = view->merge;
	
	if (view->ui_id > 0) {
		gtk_ui_manager_remove_ui (merge, view->ui_id);
		gtk_ui_manager_ensure_update (merge);
		view->ui_id = 0;
	}

	if (view->action_group != NULL) {
		gtk_ui_manager_remove_action_group (merge, view->action_group);
		g_object_unref (view->action_group);
		view->action_group = NULL;
	}
}

static void
open_recent_cb (GtkAction *action,
		const char *location,
		gpointer data)
{
	MarlinRecentViewMenuData *md = (MarlinRecentViewMenuData *) data;
	EggRecentItem *item;

	item = md->item;
	egg_recent_item_ref (item);

	g_signal_emit (G_OBJECT(md->view),
		       marlin_recent_view_signals[ACTIVATE], 0,
		       item);

	egg_recent_item_unref (item);
}

static void
action_destroy_data (gpointer data)
{
	MarlinRecentViewMenuData *md = data;

	egg_recent_item_unref (md->item);
	g_free (md);
}


static void
marlin_recent_view_set_list (MarlinRecentView *view,
			     GList *list)
{
	GString *xml;
	gint i, len;
	GtkUIManager *merge = GTK_UI_MANAGER (view->merge);
	EggRecentModel *model;
	GList *p;
	
	g_return_if_fail (view);

	model = egg_recent_view_get_model (EGG_RECENT_VIEW (view));

	marlin_recent_view_clear (view);

	xml = g_string_new (NULL);
	g_string_append (xml, "<ui><menubar><menu name=\"FileMenu\" action=\"File\">"
			 "<menu name=\"FileRecentMenu\" action=\"FileRecent\">"
			 "<placeholder name=\"FileRecentPlaceholder\">");

	view->action_group = gtk_action_group_new ("RecentActions");
	gtk_ui_manager_insert_action_group (merge, view->action_group, 0);

	len = g_list_length (list);
	for (i = 1, p = list; i <= len; ++i, p = p->next) {
		char *verb;
		GtkAction *action;
		EggRecentItem *item = (EggRecentItem *) p->data;
		MarlinRecentViewMenuData *md;
		char *utf8_uri, *tmp, *accel_path, *accel, *basename;

		verb = g_strdup_printf ("GoRecent%d", i);
		utf8_uri = egg_recent_item_get_uri_for_display (item);
		if (utf8_uri == NULL) {
			continue;
		}

		basename = g_path_get_basename (utf8_uri);
		if (i < 10) {
			tmp = g_strdup_printf ("_%d. %s", i, basename);
		} else {
			tmp = g_strdup_printf ("%d. %s", i, basename);
		}
		action = marlin_recent_action_new (verb, utf8_uri, tmp);
		
		g_free (utf8_uri);
		g_free (basename);
		g_free (tmp);
		
		gtk_action_group_add_action (view->action_group, action);
		g_object_unref (action);

		md = g_new (MarlinRecentViewMenuData, 1);
		md->view = view;
		md->item = item;

		egg_recent_item_ref (md->item);

		g_signal_connect (action, "open-recent",
				  G_CALLBACK (open_recent_cb), md);
		g_object_set_data_full (action, "MenuData", md, (GDestroyNotify) action_destroy_data);

		tmp = g_strdup_printf ("<menuitem name=\"%sMenu\" action=\"%s\"/>\n", verb, verb);
		g_string_append (xml, tmp);

		g_free (tmp);
		g_free (verb);
	}

	g_string_append (xml, "</placeholder></menu></menu></menubar></ui>");

	if (len > 0) {
		GError *error = NULL;

		view->ui_id = gtk_ui_manager_add_ui_from_string (merge, xml->str, -1, &error);
		gtk_ui_manager_ensure_update (merge);
	}

	g_string_free (xml, TRUE);
}

static void
rebuild_menu (MarlinRecentView *view)
{
	GList *list;

	list = egg_recent_model_get_list (view->model);

	if (list != NULL) {
		marlin_recent_view_set_list (view, list);
	} else {
		marlin_recent_view_clear (view);
	}
}

static void
model_changed_cb (EggRecentModel *model, GList *list, MarlinRecentView *view)
{
	if (list != NULL)
		marlin_recent_view_set_list (view, list);
	else
		marlin_recent_view_clear (view);
}


static EggRecentModel *
marlin_recent_view_get_model (EggRecentView *view_parent)
{
	MarlinRecentView *view;
	
	g_return_val_if_fail (view_parent, NULL);
	view = MARLIN_RECENT_VIEW (view_parent);
	
	return view->model;
}

static void
marlin_recent_view_set_model (EggRecentView *view_parent, EggRecentModel *model)
{
	MarlinRecentView *view;
	GList *list;

	g_return_if_fail (view_parent);
	view = MARLIN_RECENT_VIEW (view_parent);
	
	if (view->model)
		g_signal_handler_disconnect (G_OBJECT (view->model),
					     view->changed_cb_id);
	
	view->model = model;
	g_object_ref (view->model);
	view->changed_cb_id = g_signal_connect_object (G_OBJECT (model),
						       "changed",
						       G_CALLBACK (model_changed_cb),
						       view, 0);

	list = egg_recent_model_get_list (model);
	if (list != NULL) {
		marlin_recent_view_set_list (view, list);
	} else {
		marlin_recent_view_clear (view);
	}
}

static void
marlin_recent_view_set_property (GObject *object,
				 guint prop_id,
				 const GValue *value,
				 GParamSpec *pspec)
{
	MarlinRecentView *view = MARLIN_RECENT_VIEW (object);

	switch (prop_id) {
	case PROP_MENU_MERGE:
		view->merge = g_value_get_object (value);

		if (view->model != NULL) {
			rebuild_menu (view);
		}
		break;
	default:
		break;
	}
}

static void
marlin_recent_view_get_property (GObject *object,
				 guint prop_id,
				 GValue *value,
				 GParamSpec *pspec)
{
	MarlinRecentView *view = MARLIN_RECENT_VIEW (object);

	switch (prop_id) {
	case PROP_MENU_MERGE:
		g_value_set_object (value, view->merge);
		break;

	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
	}
}

static void
marlin_recent_view_finalize (GObject *object)
{
	MarlinRecentView *view = MARLIN_RECENT_VIEW (object);
	if (view->action_group != NULL) {
		gtk_ui_manager_remove_action_group (view->merge,
						    view->action_group);
		g_object_unref (view->action_group);
	}
}

static void
marlin_recent_view_class_init (MarlinRecentViewClass * klass)
{
	GObjectClass *object_class;

	
	object_class = G_OBJECT_CLASS (klass);

	object_class->set_property = marlin_recent_view_set_property;
	object_class->get_property = marlin_recent_view_get_property;
	object_class->finalize     = marlin_recent_view_finalize;

	marlin_recent_view_signals[ACTIVATE] = g_signal_new ("activate",
			G_OBJECT_CLASS_TYPE (object_class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (MarlinRecentViewClass, activate),
			NULL, NULL,
			g_cclosure_marshal_VOID__BOXED,
			G_TYPE_NONE, 1,
			EGG_TYPE_RECENT_ITEM);

	g_object_class_install_property (object_class,
					 PROP_MENU_MERGE,
					 g_param_spec_object ("menu_merge",
							      "", "",
							      GTK_TYPE_UI_MANAGER,
							      G_PARAM_READWRITE));
	klass->activate = NULL;
}

static void
egg_recent_view_init (EggRecentViewClass *iface)
{
	iface->do_get_model = marlin_recent_view_get_model;
	iface->do_set_model = marlin_recent_view_set_model;
}

static void
marlin_recent_view_init (MarlinRecentView *view)
{
/* 	view->uid = egg_recent_util_get_unique_id (); */

/* 	view->client = gconf_client_get_default (); */
}

/**
 * marlin_recent_view_new:
 * @appname: The name of your application.
 * @limit:  The maximum number of items allowed.
 *
 * This creates a new MarlinRecentView object.
 *
 * Returns: a MarlinRecentView object
 */
MarlinRecentView *
marlin_recent_view_new (GtkUIManager *merge)
{
	MarlinRecentView *view;

	view = MARLIN_RECENT_VIEW (g_object_new (marlin_recent_view_get_type (),
						 "menu_merge", merge,
						 NULL));

	return view;
}

/**
 * marlin_recent_view_get_type:
 * @:
 *
 * This returns a GType representing a MarlinRecentView object.
 *
 * Returns: a GType
 */
GType
marlin_recent_view_get_type (void)
{
	static GType marlin_recent_view_type = 0;

	if(!marlin_recent_view_type) {
		static const GTypeInfo marlin_recent_view_info = {
			sizeof (MarlinRecentViewClass),
			NULL, /* base init */
			NULL, /* base finalize */
			(GClassInitFunc)marlin_recent_view_class_init, /* class init */
			NULL, /* class finalize */
			NULL, /* class data */
			sizeof (MarlinRecentView),
			0,
			(GInstanceInitFunc) marlin_recent_view_init
		};

		static const GInterfaceInfo view_info =
		{
			(GInterfaceInitFunc) egg_recent_view_init,
			NULL,
			NULL
		};

		marlin_recent_view_type = g_type_register_static (G_TYPE_OBJECT,
							"MarlinRecentView",
							&marlin_recent_view_info, 0);
		g_type_add_interface_static (marlin_recent_view_type,
					     EGG_TYPE_RECENT_VIEW,
					     &view_info);
	}

	return marlin_recent_view_type;
}

