/*
 *  Copyright (C) 2000 Marco Pesenti Gritti
 *            (C) 2001, 2002 Jorn Baayen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <string.h>

#include <glib/gi18n.h>

#include <gtk/gtkuimanager.h>
#include <gtk/gtkstock.h>

#include <marlin/marlin-file-utils.h>

#include <ephy-debug.h>
#include <ephy-toolbars-model.h>

#include "marlin-toolbar.h"
#include "marlin-window.h"
#include "marlin-open-history-action.h"

static void toolbar_class_init (ToolbarClass *klass);
static void toolbar_init (Toolbar *t);
static void toolbar_finalize (GObject *object);
static void toolbar_set_window (Toolbar *t, MarlinWindow *window);
static void toolbar_set_property (GObject *object,
				  guint prop_id,
				  const GValue *value,
				  GParamSpec *pspec);
static void toolbar_get_property (GObject *object,
				  guint prop_id,
				  GValue *value,
				  GParamSpec *pspec);


enum
{
	PROP_0,
	PROP_MARLIN_WINDOW
};

static GObjectClass *parent_class = NULL;

struct ToolbarPrivate
{
	MarlinWindow *window;
 	GtkUIManager *ui_merge;
	GtkActionGroup *action_group;
	gboolean visibility;
};

GType
toolbar_get_type (void)
{
        static GType toolbar_type = 0;

        if (toolbar_type == 0) {
                static const GTypeInfo our_info = {
                        sizeof (ToolbarClass),
                        NULL, /* base_init */
                        NULL, /* base_finalize */
                        (GClassInitFunc) toolbar_class_init,
                        NULL,
                        NULL, /* class_data */
                        sizeof (Toolbar),
                        0, /* n_preallocs */
                        (GInstanceInitFunc) toolbar_init
                };
		
                toolbar_type = g_type_register_static (EGG_TYPE_EDITABLE_TOOLBAR,
						       "Toolbar",
						       &our_info, 0);
        }
	
        return toolbar_type;
	
}

#if 0
static GtkAction *
toolbar_get_action (EggEditableToolbar *etoolbar,
		    const char *type,
		    const char *name)
{
	Toolbar *t = TOOLBAR (etoolbar);
	GtkAction *action = NULL;
	
	action = EGG_EDITABLE_TOOLBAR_CLASS (parent_class)->get_action (etoolbar, type, name);
	
	return action;
}
#endif

static void
toolbar_class_init (ToolbarClass *klass)
{
        GObjectClass *object_class = G_OBJECT_CLASS (klass);

        parent_class = g_type_class_peek_parent (klass);

        object_class->finalize = toolbar_finalize;
	object_class->set_property = toolbar_set_property;
	object_class->get_property = toolbar_get_property;

/* 	eet_class->get_action = toolbar_get_action; */

	g_object_class_install_property (object_class,
                                         PROP_MARLIN_WINDOW,
                                         g_param_spec_object ("MarlinWindow",
							      "", "",
                                                              MARLIN_WINDOW_TYPE,
                                                              G_PARAM_READWRITE));
}

static void
toolbar_set_property (GObject *object,
                      guint prop_id,
                      const GValue *value,
                      GParamSpec *pspec)
{
        Toolbar *t = TOOLBAR (object);

        switch (prop_id) {
	case PROP_MARLIN_WINDOW:
		toolbar_set_window (t, g_value_get_object (value));
		break;
        }
}

static void
toolbar_get_property (GObject *object,
                      guint prop_id,
                      GValue *value,
                      GParamSpec *pspec)
{
        Toolbar *t = TOOLBAR (object);

        switch (prop_id) {
	case PROP_MARLIN_WINDOW:
		g_value_set_object (value, t->priv->window);
		break;
        }
}

static void
window_cmd_open (GtkAction *action,
		 MarlinWindow *window)
{
	marlin_window_open (window);
}

static void
toolbar_setup_actions (Toolbar *t)
{
	GtkAction *action;

	t->priv->action_group = gtk_action_group_new ("SpecialToolbarActions");

	action = g_object_new (MARLIN_OPEN_HISTORY_ACTION_TYPE,
			       "name", "OpenHistory",
			       "label", _("Open"),
			       "stock_id", GTK_STOCK_OPEN,
			       "tooltip", _("Open file..."),
			       "window", t->priv->window,
			       /*"is_important", TRUE, */
			       NULL);
	g_signal_connect (action, "activate",
			  G_CALLBACK (window_cmd_open), t->priv->window);
	gtk_action_group_add_action (t->priv->action_group, action);
	g_object_unref (action);
}

static void
action_request_cb (EggEditableToolbar *etoolbar,
		   char *action_name,
		   gpointer data)
{
}

static void
init_toolbars (Toolbar *t)
{
	EphyToolbarsModel *tb_model = marlin_get_toolbars_model ();
	
	g_object_set (G_OBJECT (t),
		      "ToolbarsModel", tb_model,
		      NULL);
}

static void
toolbar_set_window (Toolbar *t, MarlinWindow *window)
{
	g_return_if_fail (t->priv->window == NULL);

	t->priv->window = window;
 	t->priv->ui_merge = marlin_window_get_menu_merge (window);

	toolbar_setup_actions (t);
	gtk_ui_manager_insert_action_group (t->priv->ui_merge,
					    t->priv->action_group, 1);

	g_signal_connect (t, "action_request",
			  G_CALLBACK (action_request_cb), NULL);
	
	g_object_set (t, "MenuMerge", t->priv->ui_merge, NULL);

	init_toolbars (t);
}

static void
toolbar_init (Toolbar *t)
{
	static EphyToolbarsModel *group = NULL;

        t->priv = g_new0 (ToolbarPrivate, 1);

	t->priv->window = NULL;
	t->priv->visibility = TRUE;
}

static void
toolbar_finalize (GObject *object)
{
	Toolbar *t;

        g_return_if_fail (object != NULL);
        g_return_if_fail (IS_TOOLBAR (object));

	t = TOOLBAR (object);
        g_free (t->priv);
	
	G_OBJECT_CLASS (parent_class)->finalize (object);
}

Toolbar *
toolbar_new (MarlinWindow *window)
{
	Toolbar *t;

	t = TOOLBAR (g_object_new (TOOLBAR_TYPE,
				   "MarlinWindow", window,
				   NULL));
	
	return t;
}
